/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.6)

   class      : AbstractHelpWindow
   date       : 20.05.98
   description: A abstract baseclass for help window implementations.

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.Frame;
import java.util.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import Utils.ResourceUtil;

/* ******************************************************************
   top level class declaration
*********************************************************************/

public abstract class AbstractHelpWindow {

  // class variable declarations
  /** Key to the help context map prefix. */
  private static final String CONTEXT_PREFIX = "help.contextMap";

  /** Pointer holding the one and only instance of this class. */
  private static AbstractHelpWindow theOnlyOne = null;

  // instance variable declarations
  /** The _Main resource bundle. */
  protected final ResourceBundle resources;
  /** Properties class for desktop settings. */
  protected final Properties desktop;
  /** Hashtable holding the urls for the known contexts. */
  protected final Hashtable content;

  // constructors
  /**
   * Create a new AbstractHelpWindow object.  Don't forget the
   * super call!
   *
   * @param resources Resource bundle containing GUI labels, etc.
   * @param desktop Properties class for desktop settings.
   */
  public AbstractHelpWindow(ResourceBundle resources, Properties desktop) {
    int i;
    String str;

    this.resources = resources;
    this.desktop   = desktop;
    content        = new Hashtable();
    i              = 1;

    str = ResourceUtil.getResourceString(resources,CONTEXT_PREFIX + (i++));
    while (str != null) {
      String[] aStr = ResourceUtil.splitToWords(str);

      if (aStr.length < 2)
        System.err.println("AbstractHelpWindow: Illegal resource: " + str);
      else
        content.put(aStr[0],aStr[1]);
      // end if
      str = ResourceUtil.getResourceString(resources,CONTEXT_PREFIX + (i++));
    } // end while
  } // end constructor AbstractHelpWindow

  // class methods
  /**
   * Returns an instance of a derived class of AbstractTool.
   *
   * @param clazz The class name.
   * @param resources Resource bundle containing GUI labels, etc.
   * @param desktop Properties class for desktop settings.
   */
  public static AbstractHelpWindow getInstance(String clazz,
                                     ResourceBundle resources,
                                     Properties desktop) {
    Object[] objects;
    Class[] classes;
    Class c;
    Constructor constr;
    AbstractHelpWindow help = null;

    if (theOnlyOne != null) return theOnlyOne;
    if (clazz == null) return null;
    objects    = new Object[2];
    objects[0] = resources;
    objects[1] = desktop;
    classes    = new Class[2];
    classes[0] = java.util.ResourceBundle.class;
    classes[1] = java.util.Properties.class;
    try {
      c      = Class.forName(clazz);
      constr = c.getConstructor(classes);
      help   = (AbstractHelpWindow)constr.newInstance(objects);
    } catch (ClassNotFoundException excep) {
      excep.printStackTrace();
    } catch (IllegalAccessException excep) {
       excep.printStackTrace();
    } catch (InstantiationException excep) {
       excep.printStackTrace();
    } catch (NoSuchMethodException excep) {
       excep.printStackTrace();
    } catch (InvocationTargetException excep) {
       excep.printStackTrace();
    }
    theOnlyOne = help;
    return help;
  } // end getInstance

  /** Quit the help window(s), free all resources. */
  public static void quit() {
    if (theOnlyOne == null) return;
    theOnlyOne.quitInstance();
    theOnlyOne = null;
  } // end quit

  // instance methods
  /** Show help for a specified keyword. */
  public abstract void showHelpOnKeyword(String keyword);

  /** Shows the content of the specified document. */
  protected abstract void showContent(String relPath);

  /** Called if a help window instance is destroyed. */
  protected abstract void quitInstance();

  /** Show help for the context specified by key. */
  public void showContext(String key) {
    String str = getContext(key);

    if (str != null) showContent(str);
  } // end showContext

  /** Get the content for a context specified by key. */
  public String getContext(String key) {
    String str = (String)content.get(key);

    if (str == null)
      System.err.println("AbstractHelpWindow: topic \"" +
                         key + "\" not found");
    return str;
  } // end getContext

  /** Show help on how to use this help window. */
  public void showHelpOnHelp() {
    showContext("help-on-help");
  } // end showHelpOnHelp

  /** Show the help index. */
  public void showIndex() {
    showContext("index");
  } // end showIndex

  /** Displays a settings dialog. */
  public void settings(Frame parent) {
    Utils.Assert.assert(Utils.Assert.ENABLED &&
      false,"Method should not be called");
  } // end settings

} // end AbstractHelpWindow

/* ******************************************************************
   end of file
*********************************************************************/
