/* ******************************************************************

   class    : AbstractMacroprocessor
   date     : 02.09.98
   compiler : javac (jdk 1.1.6)

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.Component;
import java.awt.Label;
import java.awt.Frame;
import java.io.Reader;
import java.io.File;
import java.util.ResourceBundle;
import java.util.Properties;
import java.util.Observable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import Utils.ResourceUtil;
import Messages.Message;
import Messages.AddFileMessage;

/**
 * An abstract base class for plugins which implement a
 * macroprocessor for source code templates.
 * A derived class can have a method settings(java.awt.Frame)
 * which is automatically recognized by the environment. The
 * purpose of this method is to specify macroprocessor settings.
 *
 * @author puehringer edgar
 */
public abstract class AbstractMacroprocessor extends Observable {

  // instance variable declarations
  protected final ResourceBundle resources;
  protected final Properties project;

  // constructors
  /** Create a new AbstractMacroprocessor class. */
  public AbstractMacroprocessor(ResourceBundle resources, Properties project) {
    this.resources = resources;
    this.project   = project;
  } // end constructor AbstractMacroprocessor

  // class methods
  /**
   * Returns an instance of a derived class of AbstractMacroprocessor.
   *
   * @param clazz The class name.
   * @param resources Resource bundle containing GUI labels, etc.
   * @param project Project related properties.
   */
  public static AbstractMacroprocessor getInstance(String clazz,
                  ResourceBundle resources, Properties project) {
    Object[] objects;
    Class[] classes;
    Class c;
    Constructor constr;
    AbstractMacroprocessor preproc = null;

    if (clazz == null) return null;
    objects    = new Object[2];
    objects[0] = resources;
    objects[1] = project;
    classes    = new Class[2];
    classes[0] = java.util.ResourceBundle.class;
    classes[1] = java.util.Properties.class;
    try {
      c       = Class.forName(clazz);
      constr  = c.getConstructor(classes);
      preproc = (AbstractMacroprocessor)constr.newInstance(objects);
    } catch (ClassNotFoundException excep) {
      excep.printStackTrace();
    } catch (IllegalAccessException excep) {
       excep.printStackTrace();
    } catch (InstantiationException excep) {
       excep.printStackTrace();
    } catch (NoSuchMethodException excep) {
       excep.printStackTrace();
    } catch (InvocationTargetException excep) {
       excep.printStackTrace();
    }
    return preproc;
  } // end getInstance

  // methods
  /**
   * Returns a Reader to read a file template from.  If the creation
   * of the template was canceled, then return null.
   *
   * @param parent A frame that can be used as parent for dialogs.
   */
  public abstract Reader getTemplate(Frame parent);

  /**
   * Returns a Reader to read a file template from.  If the creation
   * of the template was canceled, then return null.
   *
   * @param parent A frame that can be used as parent for dialogs.
   * @param file A file to read the template from.
   */
  public abstract Reader getTemplate(Frame parent, File file);

  /**
   * Returns the expansion of a macro defined in the last call
   * to getTemplate or null if it hadn't been defined.
   *
   * @param name The name of the macro.
   */
  public abstract String getMacro(String name);

  /**
   * Returns true if the given character equals the escape character for
   * preprocessor directives.
   */
  public abstract boolean isPreprocEsc(char ch);

  /** Return copyright information. */
  public Component getInfo() {
    return new Label(this.getClass().getName());
  } // end getInfo

  /** Displays a settings dialog. */
  public void settings(Frame parent) {
    Utils.Assert.assert(Utils.Assert.ENABLED &&
      false,"Method should not be called");
  } // end settings

  /** Returns true, if the settings method is available. */
  public boolean hasSettings() {
    String[] params = new String[1];

    params[0] = "java.awt.Frame";
    return hasMethod("settings",params);
  } // end hasSettings

  /** Returns true, if the specified method is available. */
  private boolean hasMethod(String method, String[] paramClasses) {
    Class clazz    = this.getClass();
    Class[] params = new Class[paramClasses.length];
    boolean found  = true;
    Method m;

    try {
      for (int i=0; i<params.length; i++)
        params[i] = Class.forName(paramClasses[i]);
      // end for
      m = clazz.getDeclaredMethod(method,params);
    } catch (NoSuchMethodException excep) {
      found = false;
    } catch (SecurityException excep) {
      excep.printStackTrace();
      found = false;
    } catch (ClassNotFoundException excep) {
      excep.printStackTrace();
      found = false;
    }
    return found;
  } // end hasMethod

  /** Returns an array containing all extensions of template files. */
  protected String[] getExtensions() {
    String str;

    str = project.getProperty(Const.TPL_EXT);
    Utils.Assert.assert(Utils.Assert.ENABLED &&
      (str != null),"Extension for template files not found");
    return ResourceUtil.splitToWords(str);
  } // end getExtensions

  /**
   * Returns the project directory.  It's usefull to add this to
   * the library path (classpath one java platform).
   */
  protected String getProjectDirectory() {
    // This method can both read files and urls
    String str;

    str = project.getProperty(Const.PROJECT_DIR);
    Utils.Assert.assert(Utils.Assert.ENABLED &&
      (str != null),"Project directory not spezified");
    if (str.startsWith("file:/")) str = ResourceUtil.urlToFile(str);
    return str;
  } // end getProjectDirectory

  /** Returns the source code template directory. */
  protected String getTemplateDirectory() {
    String path = project.getProperty(Const.TEMPLATE_DIR);

    if (path != null) return ResourceUtil.urlToFile(path);
    return System.getProperty("user.dir");
  } // end getTemplateDirectory

  /**
   * Write the source code template directory on the project properties,
   * if it has been changed.
   */
  protected void setTemplateDirectory(String path) {
    path = ResourceUtil.fileToURL(path);
    if (path.equals(project.getProperty(Const.TEMPLATE_DIR))) return;
    project.put(Const.TEMPLATE_DIR,path);
    setProjectChanged();
  } // end setTemplateDirectory

  /** Marks the current project as changed. */
  protected void setProjectChanged() {
    project.put(Const.PROJECT_CHANGED,"" + true);
  } // end setProjectChanged

  /** Posts a message to all observers of this observable. */
  public void postMessage(Message msg) {
    setChanged();
    notifyObservers(msg);
  } // end postMessage

  /** Add a file to the current project's source file list. */
  public void addFileToProject(File file) {
    postMessage(new AddFileMessage(file));
  } // end addFileToProject

} // end AbstractMacroprocessor

/* ******************************************************************
   end of file
*********************************************************************/
