/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.2)

   class      : CodepadPanel
   date       : 23.02.98
   description: an editor panel to build up programmers editors

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.util.*;

import javax.swing.text.*;
import javax.swing.undo.*;
import javax.swing.event.*;
import javax.swing.*;

import Utils.ResourceUtil;
import Utils.FrameUtil;
import Streams.PrintDeviceWriter;
import Streams.TabReader;
import Streams.TabWriter;
import Dialogs.SearchDlg;
import Dialogs.FileFormatDlg;
import Dialogs.AboutDlg;
import CustCtrl.JOkTextField;

/* ******************************************************************
   top level class declaration
*********************************************************************/

/**
 * Editor panel using the JCodePane component.
 *
 * @author Puehringer Edgar
 */
class CodepadPanel extends JPanel implements UndoableEditListener,
  CaretListener, FrameContentManager, AbstractEditor,
  ActionListener, FocusListener, Runnable {

  // class variable declarations
  protected static ResourceBundle resources;
  private static Properties desktop;
  protected Hashtable coloringMap;

  /** UndoManager that we add edits to. */
  protected UndoManager undo = new UndoManager();

  /** Key to the menubar resource. */
  private static final String MENUBAR = "menubar";
  /** Key to the toolbar resource. */
  private static final String TOOLBAR = "toolbar";
  /** Key to the quick search text field resource. */
  private static final String QUICK_SEARCH = "qsearch";
  /** Key to the quick search tooltip resource. */
  private static final String QSEARCH_TOOLTIP = "qsearchTooltip";
  /** The history list maximum length. */
  private static final int HISTLIST_LENGTH = 10;
  /** Key to the status line resource */
  private static final String STATUS_LINE = "StatusLine";
  /** The help topic to get help for this panel. */
  private static final String HELP_TOPIC = "codepad";
  /** The help topic to get help for the jdk. */
  private static final String JDK_HELP_TOPIC = "api-index";

  /** Identifier of the print action */
  public static final String PRINT_ACTION = "print";
  /** Identifier of the undo action for toolbar buttons. */
  public static final String UNDO_BTN_ACTION = "undo-btn";
  /** Identifier of the find action */
  public static final String FIND_ACTION = "find";
  /** Identifier of the find again action */
  public static final String FINDAGAIN_ACTION = "findagain";
  /** Identifier of the replace action */
  public static final String REPLACE_ACTION = "replace";
  /** Identifier of the font action */
  public static final String FONT_ACTION = "font";
  /** Identifier of the class browser on/off action */
  public static final String BROWSER_ACTION = "browser";
  /** Identifier of the class browser expand action */
  public static final String EXPAND_ACTION = "expand-tree";
  /** Identifier of the go to line action */
  public static final String GOTOLINE_ACTION = "go-to-line";
  /** Identifier of the file format action */
  public static final String FORMAT_ACTION = "format";
  /** Identifier of the help content action */
  public static final String CONTENT_ACTION = "content";
  /** Identifier of the jdk help content action */
  public static final String JDK_CONTENT_ACTION = "jdk-content";
  /** Identifier of the help on keyword action */
  public static final String KEYWORD_ACTION = "keyword";
  /** Identifier of the about action */
  public static final String ABOUT_ACTION = "about";
  /** Identifier of the file history action */
  public static final String FILEHIST_ACTION = "file-history";

  /** Key which spezifies if the AltGr bug should be fixed. */
  private static final String ALTGR_BUGFIX_ON = "fixAltGrBug";
  /** Key which spezifies if the mnemonic bug should be fixed. */
  private static final String MNEMONIC_BUGFIX_ON = "fixMnemonicBug";
  /** Key to the font settings resource. */
  private static final String FONT_SETTINGS = "fontSettings";

  private static final String GOTOLINEDLG_TITLE = "GoToLineDlg.title";
  private static final String GOTOLINEDLG_TEXT  = "GoToLineDlg.text";
  private static final String SEARCHDLG_PREFIX  = "searchDialog.";
  private static final String FORMATDLG_PREFIX  = "fileFormatDlg.";
  private static final String ABOUTDLG_PREFIX   = "aboutDlg.";

  /** Key to the search not found message label. */
  private static final String NOTFOUNDMSG_LABEL = "searchNotFoundDlg.label";
  /** Key to the replace confirm dialog label. */
  private static final String REPCONFDLG_BUTTONS = "replaceConfDlg.buttons";
  /** Key to the replace confirm dialog label. */
  private static final String REPCONFDLG_LABEL = "replaceConfDlg.label";
  /** Key to the add file to project dialog label. */
  private static final String ADDFILEMSG_LABEL = "addFileToProjectDlg.label";
  /** Key to the syntax coloring map prefix. */
  private static final String COLORING_PREFIX = "coloringMap";
  /** Key to the default document name. */
  public final static String DEFNAME_PROPERTY = "propertiesDefName";

  // Constants for the run method to decide what to do.
  private final static int TOGGLE_BROWSER = 0;
  private final static int GOTO_LINE      = 1;

  // instance variable declarations
  private Frame frame;
  private WindowStateTracker wndState;
  private Action[] externalActions;
  private JPanel panel;
  private JScrollPane scroller;
  private EditorActionsSet editorActions;
  private Hashtable commands;
  private Hashtable menuItems;
  private JMenuLoader menuLoader;
  private JMenuBar menubar;
  private JToolBarLoader toolbarLoader;
  private JToolBar toolbar;
  private JLabel status;
  private String[] sourceExt;
  protected Properties project;
  protected JCodePane editor;
  protected String originalTitle;
  /** JSplitPane being shown to the user. */
  protected JSplitPane splitPane;
  /** The quick search text field. */
  protected JOkTextField tfQuickSearch;
  /** The file history list data. */
  protected Vector histListData;
  /** The search options. */
  protected class SearchOpt {
    public String text      = null;
    public String replace   = null;
    public boolean caseSens = false;
    public boolean wordOnly = false;
    public boolean quick    = false;
    public boolean focus    = false;
  }
  protected SearchOpt searchOpt = new SearchOpt();
  /** Tells the run method what to do. */
  private int runOption;
  /** Acts as parameter for the run method. */
  private boolean runBoolParam;
  /** Acts as parameter for the run method. */
  private int runIntParam;

  // constructors
  /** Create a new codepad editor panel. */
  public CodepadPanel(Frame frame, ResourceBundle res, Properties desk,
                      Properties project) {
    this(frame,res,desk,project,null,null);
  } // end constructor CodepadPanel

  /**
   * Create a new codepad editor panel.
   *
   * @param externalActions A list of additionally supported actions.
   */
  public CodepadPanel(Frame frame, ResourceBundle res, Properties desk,
                      Properties project, Action[] externalActions) {
    this(frame,res,desk,project,externalActions,null);
  } // end constructor CodepadPanel

  /**
   * Create a new codepad editor panel.
   *
   * @param externalActions A list of additionally supported actions.
   * @param file A file that will be loaded at creation of this panel.
   */
  public CodepadPanel(Frame frame, ResourceBundle res, Properties desk,
                      Properties project, Action[] externalActions,
                      File file) {
    super(true);
    this.frame    = frame;
    originalTitle = frame.getTitle();
    resources     = res;
    desktop       = desk;
    this.project  = project;
    wndState      = new WindowStateTracker();
    frame.addWindowListener(wndState);
    this.externalActions = externalActions;

    setLayout(new BorderLayout());

    // initialize the undo manager
    undo.setLimit(30);
    // create the embedded JTextComponent
    String str = desktop.getProperty(FONT_SETTINGS);
    Font font  = null;

    if (str != null) try {
      String[] aStr = ResourceUtil.splitToWords(str);

      if (aStr.length == 3) {
        font = new Font(aStr[0],Integer.parseInt(aStr[1]),
                        Integer.parseInt(aStr[2]));
      } // end if
    } catch (NumberFormatException excep) {
      excep.printStackTrace();
    }
    if (font == null) {
      System.err.println("CodepadPanel: resource not found: " + FONT_SETTINGS);
      font = new Font("Courier",Font.PLAIN,12);
    } // end if
    editor = createEditor();
    editor.setFont(font);
    // Add this as a listener for undoable edits
    editor.getDocument().addUndoableEditListener(this);
    // Add this as a listener for caret movement
    editor.addCaretListener(this);
    // The editor should request the focus on activationg the window
    wndState.setFocusRequester(editor);
    // install the command table
    editorActions = new EditorActionsSet(this);
    commands      = new Hashtable();
    Action[] actions = getActions();
    for (int i = 0; i < actions.length; i++) {
      Action a = actions[i];

      commands.put(a.getValue(Action.NAME), a);
    } // end for
    scroller = new JScrollPane();
    scroller.setBackground(SystemColor.control);

    JViewport port = scroller.getViewport();

    port.setBackground(SystemColor.window);
    port.add(editor);
    try {
      String vpFlag = resources.getString("ViewportBackingStore");
      Boolean bs    = new Boolean(vpFlag);
      port.setBackingStoreEnabled(bs.booleanValue());
    } catch (MissingResourceException mre) {
      // just use the viewport default
    }

    menuItems  = new Hashtable();
    menuLoader = new JMenuLoader(resources);
    menuLoader.setCommands(commands);
    menubar = menuLoader.createMenuBar(getMenuKey());
    add("North",menubar);
    panel = new JPanel();
    panel.setLayout(new BorderLayout());
    // initialize the quick search control
    tfQuickSearch = new JOkTextField(15) {

      public boolean isFocusTraversable() {
        return false;
      } // end isFocusTraversable

    };
    tfQuickSearch.setToolTipText(ResourceUtil.getResStringOrKey(resources,
                                   QSEARCH_TOOLTIP));
    tfQuickSearch.addActionListener(this);
    tfQuickSearch.addFocusListener(this);
    // initialize the file history list
    Object[] aObj = new Object[2];

    histListData = new Vector(HISTLIST_LENGTH);
    aObj[0] = "";
    aObj[1] = new Integer(0);
    histListData.addElement(aObj);
    toolbarLoader = new SmartToolBarLoader(resources) {

      /** Hook through which every toolbar item is created. */
      protected Component createTool(String key) {
        if (key.equals(QUICK_SEARCH)) return tfQuickSearch;
          else return super.createTool(key);
      } // end createTool

    };
    toolbarLoader.setCommands(commands);
    toolbar = toolbarLoader.createToolBar(getToolbarKey());
    panel.setBackground(SystemColor.window);
    panel.add("North",toolbar);
    panel.add("Center",scroller);
    add("Center",panel);
    status = new JLabel(ResourceUtil.getResStringOrKey(resources,
                          getStatuslineKey()));
    add("South",status);
    editor.requestFocus();
    // initalize the filename filter
    str = project.getProperty(Const.SOURCE_EXT);
    if (str != null) sourceExt = ResourceUtil.splitToWords(str);
      else System.err.println("CodepadPanel: " + Const.SOURCE_EXT +
                              " not in resource file");
    // if the source code template preprocessor isn't available,
    // then disable the menu item.
    if (ProjectPanel.getPreprocessor() == null) {
      Action a = (Action)commands.get(Const.ALTNEW_ACTION);

      a.setEnabled(false);
    } // end if
    // Build up syntax coloring map
    if (coloringMap == null) {
      int i      = 1;

      coloringMap = new Hashtable();
      str = ResourceUtil.getResourceString(resources,COLORING_PREFIX + (i++));
      while (str != null) {
        String[] aStr = ResourceUtil.splitToWords(str);

        if (aStr.length < 2)
          System.err.println("CodepadPanel: Illegal resource: " + str);
        else
          coloringMap.put(aStr[0],aStr[1]);
        // end if
        str = ResourceUtil.getResourceString(resources,COLORING_PREFIX + (i++));
      } // end while
    } // end if
    // if a file was spezified, then load it
    if (file != null) editorActions.loadFile(file);
  } // end constructor CodepadPanel

  // class methods
/* DEBUG ON */
  /** Only for debuging purposes */
  public static void main(String[] args) {
    // load the resource bundle
    ResourceBundle res;
    Properties desktop = new Properties();
    Properties project = new Properties();
    File file;

    try {
      res = ResourceBundle.getBundle("_Main",Locale.getDefault());
    } catch (MissingResourceException mre) {
      System.err.println("CodepadPanel: _Main.properties not found");
      System.exit(1);
      res = null; // to avoid not initialized error
    }
    // Load the desktop settings
    try {
      file = new File("Desktop.properties");
    } catch (NullPointerException npe) {
      System.err.println("CodepadPanel: Desktop.properties not found");
      System.exit(2);
      file = null; // to avoid not initialized error
    }
    if (file.exists()) try {
      FileInputStream fis    = new FileInputStream(file);
      BufferedInputStream in = new BufferedInputStream(fis);

      desktop.load(in);
      in.close();
      fis.close();
    } catch (IOException excep) {
      excep.printStackTrace();
    }

    JFrame frame     = new JFrame();
    Container cp     = frame.getContentPane();
    CodepadPanel codepad = new CodepadPanel(frame,res,desktop,project);

    frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    cp.setLayout(new BorderLayout());
    cp.add("Center",codepad);
    frame.pack();
    frame.setSize(500,600);
    frame.show();
  } // end main
/* DEBUG OFF */

  // instance methods
  /**
   * Messaged when the Caret position has changed.
   */
  public void caretUpdate(CaretEvent e) {
    setCaretLabel(editor.getCaretCharPos());
  } // end caretUpdate

  protected void setCaretLabel(Point point) {
    String str = (point.x+1)+":"+(point.y+1);

    if (editor.hasChanged()) str += " *";
    status.setText(str);
  } // end setCaretLabel

  /**
   * Messaged when the Document has created an edit, the edit is
   * added to <code>undo</code>, an instance of UndoManager.
   */
  public void undoableEditHappened(UndoableEditEvent e) {
    undo.addEdit(e.getEdit());
    undoAction.update();
    redoAction.update();
  } // end undoableEditHappened

  /**
   * Fetch the list of actions supported by this
   * editor.  It is implemented to return the list
   * of actions supported by the embedded JTextComponent
   * augmented with the actions defined locally.
   */
  public Action[] getActions() {
    if (externalActions == null)
      return TextAction.augmentList(editor.getActions(),
               TextAction.augmentList(editorActions.getActions(),
                                      defaultActions));
    else
      return TextAction.augmentList(editor.getActions(),
               TextAction.augmentList(editorActions.getActions(),
                 TextAction.augmentList(defaultActions,
                                        externalActions)));
    // end if
  } // end getActions

  /** Returns the key to the menu bar. */
  protected String getMenuKey() {
    return MENUBAR;
  } // end getMenuKey

  /** Returns the key to the tool bar. */
  protected String getToolbarKey() {
    return TOOLBAR;
  } // end getToolbarKey

  /** Returns the key to the status line text. */
  protected String getStatuslineKey() {
    return STATUS_LINE;
  } // end getStatuslineKey

  /** Returns the help topic for this panel. */
  protected String getHelpTopic() {
    return HELP_TOPIC;
  } // end getHelpTopic

  /** Returns the key to the about dialog resource prefix. */
  protected String getAboutDlgKey() {
    return ABOUTDLG_PREFIX;
  } // end getAboutDlgKey

  /** Create an empty document. */
  protected Document createDocument() {
    String str;

    str = ResourceUtil.getResourceString(resources,CodeEditorKit.COLORING_PROPERTY);
    return createDocument(str);
  } // end createDocument

  /** Create an empty document. */
  protected Document createDocument(String coloringClass) {
    Document doc = new PlainDocument();
    String str;

    str = project.getProperty(Const.TAB_SIZE,"8");
    try {
      doc.putProperty(CodeEditorKit.TABSIZE_PROPERTY,new Integer(str));
    } catch (NumberFormatException excep) {
      excep.printStackTrace();
    }
    if (coloringClass != null)
      doc.putProperty(CodeEditorKit.COLORING_PROPERTY,coloringClass);
    // end if
    return doc;
  } // end createDocument

  /** Get the coloring class name for a file name. */
  private final String getColoringName(String name) {
    String key;
    Enumeration enum = coloringMap.keys();

    while (enum.hasMoreElements()) {
        key = enum.nextElement().toString();
        if (name.endsWith(key)) return coloringMap.get(key).toString();
    } // end while
    return null;
  } // end getColoringName

  /**
   * Create an editor to represent the given document.
   */
  protected JCodePane createEditor() {
    JCodePane cp;
    String s     = ResourceUtil.getResourceString(resources,ALTGR_BUGFIX_ON);
    boolean b    = (s != null) && s.toLowerCase().equals("true");

    cp = new JCodePane(createDocument());
    cp.fixAltGrBug(b);
    s = ResourceUtil.getResourceString(resources,MNEMONIC_BUGFIX_ON);
    b = (s != null) && s.toLowerCase().equals("true");
    cp.fixMnemonicBug(b);
    return cp;
  } // end createEditor

  /**
   * Fetch the editor contained in this panel.
   */
  public JTextComponent getEditor() {
    return editor;
  } // end getEditor

  /**
   * Return the word around the caret.
   */
  public String getWordAtCaretPos() {
    if (searchOpt.focus && (tfQuickSearch.getText().length() > 0))
      return tfQuickSearch.getText().trim();
    else
      return editor.getWordAtCaretPos();
    // end if
  } // end getWordAtCaretPos

  /** Move the caret to the specified editor line. */
  public void goToLine(int lineNr) {
    runOption   = GOTO_LINE;
    runIntParam = lineNr;
    SwingUtilities.invokeLater(this);
  } // end goToLine

  /** Return the current editor file. */
  public File getCurrentFile() {
    return editorActions.getCurrentFile();
  } // end getCurrentFile

  /** Save the current editor file. */
  public boolean saveCurrentFile() {
    // conditional save
    if ((getCurrentFile() == null) ||
        isChanged()) return editorActions.saveFile();
    return true;
  } // end saveCurrentFile

  /** Returns true if the current editor file is member of the project. */
  public boolean isProjectMember() {
    return isProjectMember(getCurrentFile());
  } // end isProjectMember

  /** Returns true if file is member of the project. */
  private boolean isProjectMember(File file) {
    int numFiles  = Integer.parseInt(project.getProperty(Const.NUM_FILES));
    String cur,name;

    if (file == null) return false;
    cur = file.getAbsolutePath();
    for (int i=0; i<numFiles; i++) {
      name = project.getProperty(Const.FILE_PREFIX + i);
      if (name == null) return false;
      name = ResourceUtil.urlToFile(name);
      if (name.equals(cur)) return true;
    } // end for
    return false;
  } // end isProjectMember

  /** Returns the project directory. */
  protected String getProjectDirectory() {
    String str = project.getProperty(Const.PROJECT_DIR);

    Utils.Assert.assert(Utils.Assert.ENABLED && (str != null),"Project directory not spezified");
    return ResourceUtil.urlToFile(str);
  } // end getProjectDirectory

  /** Turn the class browser on and off. */
  protected void setClassBrowserVisible(boolean b) {
    tfQuickSearch.requestFocus();
    runOption    = TOGGLE_BROWSER;
    runBoolParam = b;
    SwingUtilities.invokeLater(this);
  } // end setClassBrowserVisible

  protected Action getAction(String cmd) {
    return (Action)commands.get(cmd);
  } // end getAction

  /** Check if editor can be dismissed without loss of data. */
  public boolean canClose() {
    return editorActions.canClose();
  } // end canClose

  /** Returns false, because this is no application main frame. */
  public boolean isMainFrame() {
    return false;
  } // end isMainFrame

  /**
   * Called from the frame manager if the list of frames has been
   * changed.
   */
  public void update(Observable o, Object arg) {
    Enumeration enum = (Enumeration)arg;
    int index        = menubar.getMenuCount() - 2;
    int i            = 0;
    JMenu menu;

    if (index < 0) {
      getToolkit().beep();
      return;
    } // end if
    menu = menubar.getMenu(index);
    menu.removeAll();
    if (!enum.hasMoreElements())
      enum = ProjectPanel.getFrameManager().getFrames();
    // end if
    while (enum.hasMoreElements()) {
      Frame f = (Frame)enum.nextElement();
      JMenuItem mi;

      if (f == frame) mi = new JCheckBoxMenuItem(f.getTitle(),true);
        else mi = new JMenuItem(f.getTitle());
      mi.setActionCommand("" + (i++));
      mi.addActionListener(this);
      menu.add(mi);
    } // end while
  } // end update

  // action implementations
  private UndoAction    undoAction      = new UndoAction();
  private UndoBtnAction undoBtnAction   = new UndoBtnAction();
  private RedoAction    redoAction      = new RedoAction();
  private BrowserAction browserAction   = new BrowserAction();
  private ExpandAction  expandAction    = new ExpandAction();
  /** Actions defined by the CodepadPanel class. */
  private Action[] defaultActions = {
    new PrintAction(),
    undoAction,
    undoBtnAction,
    redoAction,
    new FindAction(),
    new FindAgainAction(),
    new ReplaceAction(),
    new FontAction(),
    browserAction,
    expandAction,
    new GoToLineAction(),
    new FormatAction(),
    new ContentAction(),
    new ContentAction(JDK_CONTENT_ACTION,JDK_HELP_TOPIC),
    new KeywordAction(),
    new AboutAction(),
    new FileHistAction()
  };

  //--------- AbstractEditor ----------------------------------------
  /** Returns the hosting frame. */
  public Frame getFrame() {
    return frame;
  } // end getFrame

  /** Returns the resource bundle. */
  public ResourceBundle getResources() {
    return resources;
  } // end getResources

  /** Returns properties containing desktop settings. */
  public Properties getDesktopProperties() {
    return desktop;
  } // end getDesktopProperties

  /** Save the document to the spezified file. */
  public boolean save(File file) throws IOException {
    AbstractClassBrowser classBrowser = ProjectPanel.getClassBrowser();
    Writer out          = new FileWriter(file);
    BufferedWriter bout = new BufferedWriter(out);
    Document doc        = editor.getDocument();
    int line            = editor.getCaretCharPos().y;
    TabWriter tab;
    int ts = 8;
    String str;
    Boolean useTabs;
    int format;

    try {
      ts = Integer.parseInt(doc.getProperty(CodeEditorKit.TABSIZE_PROPERTY).toString());
    } catch (NumberFormatException excep) {
      excep.printStackTrace();
    }
    str     = project.getProperty(Const.USE_TABS,"false");
    useTabs = new Boolean(str);
    str     = project.getProperty(Const.FILE_FORMAT,"" + TabWriter.PLATFORM);
    try {
      format = Integer.parseInt(str);
    } catch (NumberFormatException excep) {
      excep.printStackTrace();
      format = TabWriter.PLATFORM;
    }
    tab = new TabWriter(bout,ts,useTabs.booleanValue(),format);
    editor.write(tab);
    tab.close();
    bout.close();
    out.close();
    ProjectPanel.getFrameManager().setTitle(frame,file.getName() +
                                            " - " + originalTitle);
    setCaretLabel(editor.getCaretCharPos());
    if (classBrowser != null) {
      classBrowser.update(frame);
      browserAction.setEnabled(true);
    } // end if
    // update file history list
    addHistListItem(file,line,true);
    // ask if file should be added to the project
    if (!isProjectMember(file) &&
        accept(new File(file.getParent()),file.getName()) &&
        (ProjectPanel.getPreprocessor() != null) &&
        !file.equals(editorActions.getCurrentFile())) {
      String text = '"' + file.getAbsolutePath() + '"' +
                    ResourceUtil.getResStringOrKey(resources,ADDFILEMSG_LABEL);
      int result  = JOptionPane.showConfirmDialog(frame,text,
                      originalTitle,JOptionPane.YES_NO_OPTION,
                      JOptionPane.QUESTION_MESSAGE);
      if (result == JOptionPane.YES_OPTION) {
        ProjectPanel.getPreprocessor().addFileToProject(file);
      } // end if
    } // end if
    // update the assert property
    String assertDir = project.getProperty(Const.CUR_ASSERT);

    str = (String)doc.getProperty(Const.CUR_ASSERT);
    if ((str == null) || str.equals(assertDir)) return true;
    project.put(Const.CUR_ASSERT,str);
    project.put(Const.PROJECT_CHANGED,"" + true);
    return true;
  } // end save

  /** Load the document from the spezified file. */
  public boolean load(File file) throws IOException {
    Document doc       = editor.getDocument();
    Reader in          = new FileReader(file);
    BufferedReader bin = new BufferedReader(in);
    int fmt            = ResourceUtil.getResourceInt(resources,Const.LOAD_FORMAT,
                                                     TabReader.NO_CHANGE);
    int line           = editor.getCaretCharPos().y;
    String cClass      = getColoringName(file.getName());
    TabReader tab;
    int ts = 8;
    Object[] aObj;

    if (doc != null) {
      doc.removeUndoableEditListener(this);
      undo.discardAllEdits();
    } // end if
    editor.setDocument(createDocument(cClass));
    ProjectPanel.getFrameManager().setTitle(frame,file.getName() +
                                            " - " + originalTitle);
    browserAction.setEnabled(true);
    doc = editor.getDocument();
    try {
      ts = Integer.parseInt(doc.getProperty(CodeEditorKit.TABSIZE_PROPERTY).toString());
    } catch (NumberFormatException excep) {
      excep.printStackTrace();
    }
    tab = new TabReader(bin,ts,fmt);
    editor.read(tab,"");
    tab.close();
    bin.close();
    in.close();
    doc.addUndoableEditListener(CodepadPanel.this);
    editor.setCaretPosition(0);
    ProjectPanel.getFrameManager().setTitle(frame,file.getName() +
                                            " - " + originalTitle);
    // update file history list
    addHistListItem(file,line,false);
    return true;
  } // end load

  /** Create a new empty document. */
  public boolean create(boolean alternate) {
    if (alternate) {
      AbstractMacroprocessor preproc = ProjectPanel.getPreprocessor();
      Reader reader                = preproc.getTemplate(frame);
      if (reader == null) return false;
        else return createFromTemplate(reader,preproc);
    } // end if

    Document oldDoc = editor.getDocument();
    int line        = editor.getCaretCharPos().y;

    if(oldDoc != null) {
      oldDoc.removeUndoableEditListener(CodepadPanel.this);
      undo.discardAllEdits();
    } // end if
    editor.setDocument(createDocument());
    editor.getDocument().addUndoableEditListener(CodepadPanel.this);
    ProjectPanel.getFrameManager().setTitle(frame,originalTitle);
    browserAction.setEnabled(false);
    revalidate();
    // update file history list
    addHistListItem(null,line,false);
    return true;
  } // end create

  private final boolean createFromTemplate(Reader in,
                                           AbstractMacroprocessor preproc) {
    Document doc       = editor.getDocument();
    BufferedReader bin = new BufferedReader(in);
    int fmt            = ResourceUtil.getResourceInt(resources,Const.LOAD_FORMAT,
                           TabReader.NO_CHANGE);
    int line           = editor.getCaretCharPos().y;
    TabReader tab;
    int ts = 8;
    String str;

    if (doc != null) {
      doc.removeUndoableEditListener(this);
      undo.discardAllEdits();
    } // end if
    doc = createDocument(getColoringName(preproc.getMacro("__EXT__")));
    // check if new document is an assertion class
    str = preproc.getMacro("__ASSERT__");
    if (str != null) {
      str = preproc.getMacro("__RELPATH__");
      if (str == null) str = "";
      doc.putProperty(Const.CUR_ASSERT,str);
    } // end if
    // build the default name of the new document
    str = preproc.getMacro("__NAME__");
    if ((str != null) && (preproc.getMacro("__EXT__") != null)) {
      str += preproc.getMacro("__EXT__");
      if (preproc.getMacro("__RELPATH__") != null)
        str = preproc.getMacro("__RELPATH__") + File.separator + str;
      // end if
      doc.putProperty(DEFNAME_PROPERTY,str);
    } // end if
    editor.setDocument(doc);
    ProjectPanel.getFrameManager().setTitle(frame,originalTitle);
    browserAction.setEnabled(false);
    try {
      ts = Integer.parseInt(doc.getProperty(CodeEditorKit.TABSIZE_PROPERTY).toString());
    } catch (NumberFormatException excep) {
      excep.printStackTrace();
    }
    tab = new TabReader(bin,ts,fmt);
    try {
      editor.read(tab,"");
    } catch (IOException excep) {
      excep.printStackTrace();
      editor.setDocument(createDocument());
      doc = editor.getDocument();
    } finally {
      try { tab.close(); } catch (IOException e) {}
      try { bin.close(); } catch (IOException e) {}
      try { in.close();  } catch (IOException e) {}
    }
    doc.addUndoableEditListener(CodepadPanel.this);
    editor.setCaretPosition(0);
    // update file history list
    addHistListItem(null,line,false);
    return true;
  } // end createFromTemplate

  /** Write an item to the file history list. */
  private void addHistListItem(File file, int oldLine, boolean replace) {
    String path   = file == null ? ""  : file.getAbsolutePath();
    int line      = editor.getCaretPosition();
    Object[] aObj = (Object[])histListData.elementAt(0);
    Object[] tmp;

    if (aObj[0].toString().equals(path) || aObj[0].toString().equals(""))
      replace = true;
    // end if
    if (replace) histListData.removeElementAt(0);
      else aObj[1] = new Integer(oldLine);
    aObj    = new Object[2];
    aObj[0] = path;
    aObj[1] = new Integer(line);
    histListData.insertElementAt(aObj,0);
    // Remove duplicate elems.
    for (int i=1; i<histListData.size(); i++) {
      tmp = (Object[])histListData.elementAt(i);
      if (tmp[0].toString().equals(path)) {
        histListData.removeElementAt(i);
        break;
      } // end if
    } // end for
    // Trim history list to the max. length.
    if (histListData.size() > HISTLIST_LENGTH)
      histListData.setSize(HISTLIST_LENGTH);
    // end if
  } // end addHistListItem

  /** Returns true, if the document has been changed. */
  public boolean isChanged() {
    return editor.hasChanged();
  } // end isChanged

  /** Return a default file to display in the save file dialog. */
  public File getDefaultFile() {
    File file = editorActions.getCurrentFile();
    Document doc;
    String str;

    if (file != null) return file;
    doc = getEditor().getDocument();
    str = (String)doc.getProperty(DEFNAME_PROPERTY);
    if (str != null)
      return new File(getProjectDirectory(),str);
    // end if
    if ((sourceExt == null) || (sourceExt.length < 1)) return null;
    return new File("*" + sourceExt[0]);
  } // end getDefaultFile

  public boolean accept(File dir, String name) {
    if (sourceExt == null) return true;
    for (int i=0; i<sourceExt.length; i++)
      if (name.endsWith(sourceExt[i])) return true;
    return false;
  } // end accept

  // --- ActionListener ---------------------------------------------
  /** Invoked when an action occurs. */
  public void actionPerformed(ActionEvent e) {
    boolean result;

    if (e.getSource() instanceof JMenuItem) { // window menu
      Object obj = e.getSource();

      if (obj instanceof JCheckBoxMenuItem) {
        ((JCheckBoxMenuItem)obj).setState(true);
      }
      else {
        int index        = Integer.parseInt(e.getActionCommand());
        int i            = 0;
        Enumeration enum = ProjectPanel.getFrameManager().getFrames();

        while (enum.hasMoreElements()) {
          Frame f = (Frame)enum.nextElement();

          if ((i++) == index) {
            f.toFront();
            f.requestFocus();
            return;
          } // end if
        } // end while
      } // end if
      return;
    } // end if
    searchOpt.text    = tfQuickSearch.getText();
    searchOpt.replace = "";
    searchOpt.quick   = true;
    if (searchOpt.text.length() > 0)
      tfQuickSearch.select(0,searchOpt.text.length());
    // end if
    result = editor.search(searchOpt.text,searchOpt.wordOnly,
                           searchOpt.caseSens);
    if (!result) {
      String str;

      str = ResourceUtil.getResStringOrKey(resources,NOTFOUNDMSG_LABEL);
      status.setText(str);
      if (searchOpt.focus) editor.requestFocus();
      getToolkit().beep();
    } // end if
  } // end actionPerformed

  // --- FocusListener ----------------------------------------------
  /** Invoked when a component gains the keyboard focus. */
  public void focusGained(FocusEvent e) {
    searchOpt.focus = true;
  } // end focusGained

  /** Invoked when a component loses the keyboard focus. */
  public void focusLost(FocusEvent e) {
    searchOpt.focus = false;
  } // end focusLost

  //---- Runnable --------------------------------------------------
  public void run() {
    switch (runOption) {
      case TOGGLE_BROWSER:
        AbstractClassBrowser classBrowser = ProjectPanel.getClassBrowser();
        boolean b = runBoolParam;

        if (b == (splitPane != null)) return;
        tfQuickSearch.requestFocus();
        if (b) {
          Component cbPanel = classBrowser.getViewPanel();
          Dimension d = scroller.getSize();

          panel.remove(scroller);
          d.width /= 10;
          scroller.setMinimumSize(d);
          if (cbPanel instanceof JComponent) {
            ((JComponent)cbPanel).setMinimumSize(d);
          } // end if
          splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,cbPanel,scroller);
          splitPane.setContinuousLayout(false);
          panel.add(splitPane,BorderLayout.CENTER);
          panel.validate();
          panel.repaint();
        }
        else {
          panel.remove(splitPane);
          splitPane.remove(scroller);
          splitPane = null;
          scroller.setMinimumSize(null);
          panel.add(scroller,BorderLayout.CENTER);
          panel.validate();
          panel.repaint();
        } // end if
        expandAction.setEnabled(b);
        editor.requestFocus();
        System.gc();
      break;
      default: // go to line
        if (!frame.isShowing() || !wndState.isActive()) {
          try {
            Thread.sleep(50);
          } catch (InterruptedException excep) {}
          SwingUtilities.invokeLater(this);
          return;
        } // end if
        int len = editor.getDocument().getLength();
        int pos;
        Rectangle rect;
        JViewport port = scroller.getViewport();
        Dimension d    = port.getExtentSize();
        int y, end;

        editor.goToLine(runIntParam);
        setCaretLabel(new Point(0,runIntParam));
        pos = editor.getCaretPosition();
/*        try {
          rect = editor.modelToView(pos);
          y    = rect.y;
          rect = editor.modelToView(len - 1);
          end  = rect.y;
          if ((y + d.height) > end) y = end - d.height;
          if (y < 0) y = 0;
          port.setViewPosition(new Point(0,y));
        } catch (BadLocationException excep) {
          getToolkit().beep();
        } */ /* not needed in java 1.2 */
    } // end switch
  } // end run

  // inner classes
  class UndoAction extends AbstractAction {

    // constructors
    public UndoAction() {
      super("Undo");
      setEnabled(false);
    } // end constructor UndoAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      try {
        undo.undo();
      } catch (CannotUndoException excep) {
        System.err.println("Unable to undo: "+excep);
        excep.printStackTrace();
      }
      update();
      redoAction.update();
    } // end actionPerformed

    protected void update() {
      if(undo.canUndo()) {
        setEnabled(true);
        putValue(Action.NAME,undo.getUndoPresentationName());
      }
      else {
        setEnabled(false);
        putValue(Action.NAME,"Undo");
      } // end if
      undoBtnAction.update();
    } // end update

  } // end UndoAction

  /**
   * This class redirects the actionPerformed call to UndoAction.
   * It was implemented to avoid changing the text of the toolbar
   * button.
   */
  class UndoBtnAction extends AbstractAction {

    // constructors
    public UndoBtnAction() {
      super(UNDO_BTN_ACTION);
      setEnabled(false);
    } // end constructor UndoBtnAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      undoAction.actionPerformed(e);
    } // end actionPerformed

    protected void update() {
      if(undo.canUndo()) {
        setEnabled(true);
      }
      else {
        setEnabled(false);
      } // end if
    } // end update

  } // end UndoBtnAction

  class RedoAction extends AbstractAction {

    // constructors
    public RedoAction() {
      super("Redo");
      setEnabled(false);
    } // end constructor RedoAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      try {
        undo.redo();
      } catch (CannotRedoException excep) {
        System.err.println("Unable to redo: "+excep);
        excep.printStackTrace();
      }
      update();
      undoAction.update();
    } // end actionPerformed

    protected void update() {
      if(undo.canRedo()) {
        setEnabled(true);
        putValue(Action.NAME, undo.getRedoPresentationName());
      }
      else {
        setEnabled(false);
        putValue(Action.NAME, "Redo");
      } // end if
    } // end update

  } // end RedoAction

  class PrintAction extends AbstractAction {

    // constructors
    PrintAction() {
      super(PRINT_ACTION);
    } // end constructor PrintAction

    PrintAction(String nm) {
      super(nm);
    } // end constructor PrintAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      Writer prn = new PrintDeviceWriter(editor.getFont());

      try {
        editor.write(prn);
      } catch (IOException excep) {
        excep.printStackTrace();
      } finally {
        try { prn.close(); } catch (IOException e1) {}
      }
    } // end actionPerformed

  } // end PrintAction

  /** Search a string in the current editor document. */
  class FindAction extends AbstractAction {

    // instance variable declarations
    protected String notFoundTitle;

    // constructors
    FindAction() {
      super(FIND_ACTION);
    } // end constructor FindAction

    FindAction(String nm) {
      super(nm);
    } // end constructor FindAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      SearchDlg dlg = getDialog();

      dlg.setSearchOptions(searchOpt.text,searchOpt.replace,
                           searchOpt.caseSens,searchOpt.wordOnly);
      dlg.setVisible(true);
      if (dlg.isCanceled()) return;
      notFoundTitle      = dlg.getTitle();
      searchOpt.text     = dlg.getSearchString();
      searchOpt.replace  = dlg.getReplaceString();
      searchOpt.caseSens = dlg.isCaseSensitive();
      searchOpt.wordOnly = dlg.isWordsOnlyMode();
      searchOpt.quick    = false;
      tfQuickSearch.setText(searchOpt.text);
      search(dlg.isToConfirm());
    } // end actionPerformed

    /** Perform a search specified by searchOpt. */
    protected void search(boolean confirm) {
      boolean found = editor.search(searchOpt.text,
                                      searchOpt.wordOnly,
                                      searchOpt.caseSens);
      if (found) {
        replaceString(searchOpt.replace,confirm);
      }
      else {
        if (searchOpt.quick) {
          String str;

          str = ResourceUtil.getResStringOrKey(resources,NOTFOUNDMSG_LABEL);
          status.setText(str);
          getToolkit().beep();
        }
        else {
          JOptionPane.showMessageDialog(frame,
            ResourceUtil.getResStringOrKey(resources,NOTFOUNDMSG_LABEL),
            notFoundTitle,JOptionPane.INFORMATION_MESSAGE);
        } // end if
      } // end if
    } // end search

    /** May be overwritten to create a replace operation. */
    protected SearchDlg getDialog() {
        return new SearchDlg(frame,SearchDlg.SEARCH,
                               resources,SEARCHDLG_PREFIX);
    } // end getDialog

    /** May be overwritten to create a replace operation. */
    protected void replaceString(String replace, boolean confirm) {
      // do nothing
    } // end replaceString

  } // end FindAction

  /** Repeat a previous performed find action. */
  class FindAgainAction extends FindAction {

    // constructors
    FindAgainAction() {
      super(FINDAGAIN_ACTION);
    } // end constructor FindAgainAction

    // methods
    public void actionPerformed(ActionEvent e) {
      if (searchOpt.focus) editor.requestFocus();
      if (searchOpt.text == null) super.actionPerformed(e);
        else search(false);
    } // end actionPerformed

  } // end FindAgainAction

  /** Search and replace a string in the current editor document. */
  class ReplaceAction extends FindAction {

    // constructors
    ReplaceAction() {
      super(REPLACE_ACTION);
    } // end constructor ReplaceAction

    // methods
    protected SearchDlg getDialog() {
      return new SearchDlg(frame,SearchDlg.REPLACE,
                             resources,SEARCHDLG_PREFIX);
    } // end getDialog

    protected void replaceString(String replace, boolean confirm) {
      int optNr = 0; // default: replace it
      boolean found;

      do {
        if (confirm) {
          String str;
          Object[] options;

          str = ResourceUtil.getResStringOrKey(resources,REPCONFDLG_BUTTONS);
          options = ResourceUtil.splitToWords(str);
          str = ResourceUtil.getResStringOrKey(resources,REPCONFDLG_LABEL);
          optNr = JOptionPane.showOptionDialog(frame,str,
                    notFoundTitle,JOptionPane.DEFAULT_OPTION,
                    JOptionPane.QUESTION_MESSAGE,null,options,options[0]);
        } // end if
        switch (optNr) {
          case 0:  // replace it
            editor.replaceSelection(replace);
          case 1:  // continue operation
            found = editor.search(searchOpt.text,
                                    searchOpt.wordOnly,
                                    searchOpt.caseSens);
          break;
          default: // suspend operation
            found = false;
        } // end switch
      } while (found);
    } // end replaceString

  } // end ReplaceAction

  class FontAction extends AbstractAction {

    // constructors
    FontAction() {
      super(FONT_ACTION);
    } // end constructor FontAction

    // methods
    public void actionPerformed(ActionEvent e) {
      FontDlg fontDlg = new FontDlg(frame,editor.getFont(),
                                    resources,"fontDialog.");

      fontDlg.setVisible(true);
      if (!fontDlg.isCanceled()) {
        Font font  = fontDlg.getSelectedFont();
        String str = font.getName() + " " + font.getStyle() + " " +
                     font.getSize();

        editor.setFont(font);
        desktop.put(FONT_SETTINGS,str);
      } // end if
    } // end actionPerformed

  } // end FontAction

  class BrowserAction extends AbstractAction {

    // instance variable declarations
    private JCheckBoxMenuItem cbmi;

    // constructors
    BrowserAction() {
      super(BROWSER_ACTION);
      super.setEnabled(false);
    } // end constructor BrowserAction

    // methods
    public void actionPerformed(ActionEvent e) {
      cbmi = (JCheckBoxMenuItem)e.getSource();
      setClassBrowserVisible(cbmi.isSelected());
    } // end actionPerformed

    public synchronized void setEnabled(boolean newValue) {
      AbstractClassBrowser classBrowser = ProjectPanel.getClassBrowser();
      if (newValue) {
        super.setEnabled(classBrowser != null);
      }
      else {
        super.setEnabled(false);
        if ((cbmi != null) && cbmi.isSelected()) {
          cbmi.setSelected(false);
          setClassBrowserVisible(false);
        } // end if
      } // end if
    } // end setEnabled

  } // end BrowserAction

  /**
   * Expand the subtrees of all classes that are implemented
   * in the current editor file.
   */
  class ExpandAction extends AbstractAction {

    // constructors
    ExpandAction() {
      super(EXPAND_ACTION);
      setEnabled(false);
    } // end constructor ExpandAction

    // methods
    public void actionPerformed(ActionEvent e) {
      if (splitPane == null) {
        getToolkit().beep();
      }
      else {
        ClassPane cp = (ClassPane)splitPane.getLeftComponent();

        cp.expandClassesOfFile(getCurrentFile());
      } // end if
    } // end actionPerformed

  } // end ExpandAction

  /** Move the caret to an editor line keyed in in a small dialog box. */
  class GoToLineAction extends AbstractAction {

    // constructors
    GoToLineAction() {
      super(GOTOLINE_ACTION);
    } // end constructor GoToLineAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      int i        = 0;
      boolean done = false;
      String str;

      do {
        str = JOptionPane.showInputDialog(frame,
                ResourceUtil.getResStringOrKey(resources,GOTOLINEDLG_TEXT),
                ResourceUtil.getResStringOrKey(resources,GOTOLINEDLG_TITLE),
                JOptionPane.PLAIN_MESSAGE);
        if (str != null) try {
          i    = Integer.valueOf(str).intValue();
          done = true;
        } catch (NumberFormatException excep) {}
      } while ((str != null) && !done);
      if (done) editor.goToLine(i - 1);
    } // end actionPerformed

  } // end GoToLineAction

  /** Select the file format and the tabulator size. */
  class FormatAction extends AbstractAction {

    // constructors
    FormatAction() {
      super(FORMAT_ACTION);
    } // end constructor FormatAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      Document doc = editor.getDocument();
      Object obj;
      String str;
      Integer tabSize;
      Boolean useTabs;
      Integer format;
      FileFormatDlg dlg;

      obj = doc.getProperty(CodeEditorKit.TABSIZE_PROPERTY);
      if (obj == null) obj = "8";
      try {
        tabSize = new Integer(obj.toString());
      } catch (NumberFormatException excep) {
        excep.printStackTrace();
        tabSize = new Integer(8);
      }
      str     = project.getProperty(Const.USE_TABS,"false");
      useTabs = new Boolean(str);
      str     = project.getProperty(Const.FILE_FORMAT,"" + TabWriter.PLATFORM);
      try {
        format = new Integer(str);
      } catch (NumberFormatException excep) {
        excep.printStackTrace();
        format = new Integer(TabWriter.PLATFORM);
      }
      dlg = new FileFormatDlg(frame,resources,FORMATDLG_PREFIX,
                              tabSize.intValue(),useTabs.booleanValue(),
                              format.intValue());
      dlg.show();
      if (!dlg.isCanceled()) {
        tabSize = new Integer(dlg.getTabSize());
        doc.putProperty(CodeEditorKit.TABSIZE_PROPERTY,tabSize);
        project.put(Const.TAB_SIZE,tabSize.toString());
        desktop.put(Const.TAB_SIZE,tabSize.toString());
        useTabs = new Boolean(dlg.getUseTabs());
        project.put(Const.USE_TABS,useTabs.toString());
        desktop.put(Const.USE_TABS,useTabs.toString());
        format = new Integer(dlg.getFileFormat());
        project.put(Const.FILE_FORMAT,format.toString());
        desktop.put(Const.FILE_FORMAT,format.toString());
        project.put(Const.PROJECT_CHANGED,"" + true);
      } // end if
    } // end actionPerformed

  } // end FormatAction

  /** Starts the online help. */
  class ContentAction extends AbstractAction {

    // instance variable declarations
    private String topic;

    // constructors
    ContentAction() {
      super(CONTENT_ACTION);
      topic = getHelpTopic();
    } // end constructor ContentAction

    ContentAction(String nm, String topic) {
      super(nm);
      this.topic = topic;
    } // end constructor ContentAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      String clazz = ResourceUtil.getResourceString(resources,Const.HELP_CLASS);
      AbstractHelpWindow wnd;

      wnd = AbstractHelpWindow.getInstance(clazz,resources,desktop);
      if (wnd == null) setEnabled(false);
        else wnd.showContext(topic);
    } // end actionPerformed

  } // end ContentAction

  /** Starts a keyword search on the online help. */
  class KeywordAction extends AbstractAction {

    // constructors
    KeywordAction() {
      super(KEYWORD_ACTION);
    } // end constructor KeywordAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      String str = ResourceUtil.getResourceString(resources,Const.HELP_CLASS);
      AbstractHelpWindow wnd;

      wnd = AbstractHelpWindow.getInstance(str,resources,desktop);
      if (wnd == null) {
        setEnabled(false);
      }
      else {
        str = getWordAtCaretPos();
        if (str != null) wnd.showHelpOnKeyword(str);
          else Toolkit.getDefaultToolkit().beep();
      } // end if
  } // end actionPerformed

  } // end KeywordAction

  /** Display a application information. */
  class AboutAction extends AbstractAction {

    // constructors
    AboutAction() {
      super(ABOUT_ACTION);
    } // end constructor AboutAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      AboutDlg dlg = new AboutDlg(frame,resources,getAboutDlgKey());

      dlg.show();
    } // end actionPerformed

  } // end AboutAction

  /** Display a file history popup menu. */
  class FileHistAction extends AbstractAction {

    // constructors
    FileHistAction() {
      super(FILEHIST_ACTION);
    } // end constructor FileHistAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      String cmd = e.getActionCommand();

      if (cmd.startsWith("_")) {
        int index = Integer.parseInt(cmd.substring(1));

        if (index < 1) return;

        Object[] aObj = (Object[])histListData.elementAt(index);
        File file     = new File(aObj[0].toString());
        int line      = ((Integer)aObj[1]).intValue();
        File cur;

        editorActions.loadFile(file);
        cur = editorActions.getCurrentFile();
        if ((cur != null) &&
            file.getAbsolutePath().equals(cur.getAbsolutePath())) {
          goToLine(line);
        } // end if
        return;
      } // end if

      Object obj = e.getSource();
      Component comp;
      Enumeration enum;
      JPopupMenu popup;
      int i;
      if (!(obj instanceof Component)) {
        Toolkit.getDefaultToolkit().beep();
        return;
      } // end if
      comp  = (Component)obj;
      enum  = histListData.elements();
      popup = new JPopupMenu();
      i     = 0;
      while (enum.hasMoreElements()) {
        Object[] aObj = (Object[])enum.nextElement();
        String path   = aObj[0].toString();
        String name   = path.equals("") ? " " : (new File(path)).getName();
        JMenuItem mi  = new JMenuItem(name);

        mi.setActionCommand("_" + (i++));
        mi.addActionListener(this);
        popup.add(mi);
      } // end while
      popup.show(comp,0,comp.getSize().height - 1);
    } // end actionPerformed

  } // end FileHistAction

} // end CodepadPanel

/* ******************************************************************
   end of file
*********************************************************************/
