/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.6)

   class      : DefaultClassBrowser
   date       : 08.07.98
   description: an class browser, that parses java source code and
                displays it as part of the codepad panel.

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.*;
import java.io.File;

import javax.swing.*;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.DefaultMutableTreeNode;

import SID.Meta.*;

import Utils.ResourceUtil;
import Utils.FrameUtil;

/* ******************************************************************
   top level class declaration
*********************************************************************/

public class DefaultClassBrowser extends AbstractClassBrowser
  implements Runnable,ActionListener {

  // class variable declarations
  /** Key to the scanning progress label resource */
  private static final String SCAN_LABEL = "browser.scanLabel";
  /** Key to the use flat hierarchy flag. */
  private static final String USE_FLAT = "browser.useFlatHierarchy";

  // instance variable declarations
  /** The current selected executable file. */
  protected Hashtable curFiles;
  /** The text for the progress message. */
  private String scanLabel;
  /** The Thread that performs the file scanning. */
  private Thread thread;
  /** A vector that holds the files that must be scanned. */
  private Vector toScan;
  /** The number of files to scan. */
  private int numFilesToScan = 0;
  /** Hierarchy containing the java classes. */
  protected Hierarchy hierarchy;
  /** Tree model representing the java hierarchy. */
  protected DefaultTreeModel model;

  // constructors
  public DefaultClassBrowser(ResourceBundle resources,
                             Properties project) {
    super(resources,project);
    curFiles  = new Hashtable();
    scanLabel = ResourceUtil.getResStringOrKey(this.resources,SCAN_LABEL);
    toScan    = new Vector();
    hierarchy = JavaHierarchy.newJavaHierarchy();
    model     = new DefaultTreeModel(getClassTree());
  } // end constructor DefaultClassBrowser

  // methods
  /** Returns a new instance of a class browsers view panel. */
  public Component getViewPanel() {
    ClassPane cp = new ClassPane(resources,model);

    cp.addActionListener(this);
    return cp;
  } // end getViewPanel

  /**
   * Updates the class browsers modell.
   *
   * @param frame A frame that can be used as parent for dialogs.
   * @param files The source files of the project.
   */
  public synchronized void update(Frame frame, File[] files) {
    Long l;
    long time,modif;
    int found = 0;

    // check if some project files are no longer part of the project
    for (int i=0; i<files.length; i++) {
      if (curFiles.containsKey(files[i].getAbsolutePath())) found++;
    } // end for
    if (curFiles.size() > found) {     // if files removed
      // scan ALL files and build a new table
      curFiles.clear();
      clearClassTree();
      for (int i=0; i<files.length; i++) {
        modif = files[i].lastModified();
        curFiles.put(files[i].getAbsolutePath(),new Long(modif));
        scanFile(files[i]);
      } // end for
    }
    else {                             // if no files removed
      // scan files that have been changed
      for (int i=0; i<files.length; i++) {
        l     = (Long)curFiles.get(files[i].getAbsolutePath());
        time  = (l == null) ? Long.MIN_VALUE : l.longValue();
        modif = files[i].lastModified();
        if (modif > time) {
          curFiles.put(files[i].getAbsolutePath(),new Long(modif));
          scanFile(files[i]);
        } // end if
      } // end for
    } // end if
  } // end update

  /** Puts a file into the file scanning queue. */
  protected synchronized void scanFile(File file) {
    String str = file.getAbsolutePath();

    if ((thread == null) || !thread.isAlive()) numFilesToScan = 0;
    if (!toScan.contains(str)) {
      toScan.addElement(str);
      numFilesToScan++;
    } // end if
    if (toScan.size() < 1) return;
    if ((thread == null) || !thread.isAlive()) {
      thread = new Thread(this);
      thread.start();
    } // end if
  } // end scanFile

  /** Clears the class tree of this class browser. */
  protected synchronized void clearClassTree() {
    toScan.removeAllElements();
    toScan.addElement("");
  } // end clearClassTree

  /**
   * Returns the number of files that must be scanned.  That's
   * NOT equal to the that are left in the toScan vector.
   */
  protected synchronized int getNumFilesToScan() {
    return numFilesToScan;
  } // getNumFilesToScan

  /** Removes a file from the list of files to scan. */
  protected synchronized String getFileToScan() {
    String str;

    if (toScan.size() < 1) return null;
    str = (String)toScan.firstElement();
    toScan.removeElementAt(0);
    return str;
  } // end getFileToScan

  /** Hook through which the class hierarchy root is created. */
  protected MutableTreeNode getClassTree() {
    DefaultMutableTreeNode root;
    boolean flat;

    flat = ResourceUtil.getResourceBool(resources,USE_FLAT);
    root = new DefaultMutableTreeNode(flat ? "packages" : "classes",true);
    if (flat) return hierarchy.getFlatClassTree(root,true);
      else return hierarchy.getInheritanceClassTree();
  } // end getClassTree

  /** Notifies all Observers from a file. */
  protected void notifyScan(String name, int cur) {
    String path = getProjectDirectory();

    if (name.startsWith(path)) {
      name = name.substring(path.length());
      if (name.startsWith(File.separator)) name = name.substring(1);
    } // end if
    notifyScan(scanLabel,new File(name),cur,getNumFilesToScan());
  } // end notifyScan

  //--------- Runnable ----------------------------------------------
  public void run() {
    try {
      String name = getFileToScan();
      int i       = 0;

      while (name != null) {
        // an empty string causes deletion of the class tree
        if (name.equals("")) {
          hierarchy  = JavaHierarchy.newJavaHierarchy();
        }
        else {
          notifyScan(name,i);
          hierarchy.scanFile(name);
        } // end if
        name = getFileToScan();
        i++;
      } // end while
      if (i > 0) {
        hierarchy.connectInheritanceRelations();

        MutableTreeNode tmp = getClassTree();
        MutableTreeNode tn  = (MutableTreeNode)model.getRoot();
        int num             = tn.getChildCount()-1;
        Enumeration enum    = tmp.children();

        for (int j=num; j>=0; j--) {
          tn.remove(j);
        } // end for
        num = 0;
        // nodes are removed from the source, if inserted into another tree
        while (tmp.getChildCount() > 0) {
          tn.insert((MutableTreeNode)tmp.getChildAt(0),num);
          num++;
        } // end while
        model.reload();
        notifyEndOfScan(true);
      } // end if
    } catch (Throwable t) {
      t.printStackTrace();
      notifyEndOfScan(true);
    }
  } // end run

  // --- ActionListener ---------------------------------------------
  /** Invoked when an action occurs. */
  public void actionPerformed(ActionEvent e) {
    String[] aStr = ResourceUtil.splitToWords(e.getActionCommand());
    Frame frame;
    File file;
    int line;

    if (aStr.length < 2) {
      Toolkit.getDefaultToolkit().beep();
      return;
    } // end if
    try {
      line = Integer.parseInt(aStr[0]);
    } catch (NumberFormatException excep) {
      line = 0;
    }
    file  = new File(aStr[1]);
    frame = FrameUtil.getParentFrame((Component)e.getSource());
    showLocation(frame,file,line);
  } // end actionPerformed

} // end DefaultClassBrowser

/* ******************************************************************
   end of file
*********************************************************************/
