/* ******************************************************************

   class    : DefaultCompiler
   date     : 27.07.98
   compiler : javac (jdk 1.1.6)

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.util.ResourceBundle;
import java.util.Properties;
import java.util.Observer;
import java.util.Observable;
import java.util.Vector;
import java.util.StringTokenizer;

import Utils.ResourceUtil;
import ErrorParser.Parser;
import ErrorParser.ErrorDesc;
import ErrorParser.ParserObservable;
import Messages.ErrorMessage;
import Messages.LocatableErrorMessage;
import CustCtrl.BaseDialog;
import SysExt.OutputRedirector;

/**
 * A java compiler plugin.
 *
 * @author puehringer edgar
 */
public class DefaultCompiler extends AbstractCompiler
  implements Runnable,Observer {

  // class variable declarations
  /** Key to the compiling progress label resource */
  private static final String COMPILE_LABEL = "compiler.compileLabel";

  // Keys to compiler settings
  public final static String COMP_COMMAND         = "compiler.command";
  public final static String COMP_ADDPARAM        = "compiler.addParam";
  public final static String COMP_USECLASSPATH    = "compiler.useClassPath";
  public final static String COMP_CLASSPATHPREFIX = "compiler.classPathPrefix";
  public final static String COMP_TRAILINGSPACE   = "compiler.trailingSpace";
  public final static String COMP_USEERRORSTREAM  = "compiler.useErrorStream";
  // Keys to the settings dialog
  public final static String SETTDLG_TITLE   = "compiler.settingsDlg.title";
  public final static String SETTDLG_BUTTONS = "compiler.settingsDlg.buttons";
  public final static String SETTDLG_PATH    = "compiler.settingsDlg.path";
  public final static String SETTDLG_ADDPARA = "compiler.settingsDlg.addPara";
  public final static String SETTDLG_USEPATH = "compiler.settingsDlg.useCPath";
  public final static String SETTDLG_PREFIX  = "compiler.settingsDlg.prefix";
  public final static String SETTDLG_SPACES  = "compiler.settingsDlg.spaces";

  // instance variable declarations
  private String compileLabel;
  private File[] curFiles;
  /** The modification time of the newest object file. */
  private long curModTime;
  private ParserObservable observable = new ParserObservable();
  /** The number of compiler errors. */
  private int errors;
  /** The number of compiler warnings. */
  private int warnings;

  // constructors
 /** Create a new DefaultCompiler class. */
  public DefaultCompiler(ResourceBundle resources,
                         Properties project) {
    super(resources,project);
    compileLabel = ResourceUtil.getResStringOrKey(this.resources,COMPILE_LABEL);
    observable.addObserver(this);
  } // end constructor DefaultCompiler

  // methods
  /** Compile an array of files. */
  public void compile(File[] files) {
    internalMake(files,true);
  } // end compile

  /**
   * Compile all files that have been changed and that depend from.
   *
   * @return An array containing the object files.
   */
  public File[] make(File[] files) {
    return internalMake(files,false);
  } // end make

  /**
   * Compile all project files.
   *
   * @return An array containing the object files.
   */
  public File[] build(File[] files) {
    return internalMake(files,true);
  } // end build

  private File[] internalMake(File[] files, boolean allFiles) {
    String str;
    String[] sourceExt = getExtensions();
    String[] objectExt = getObjectExtensions();
    Vector makeVector  = new Vector(files.length);
    Vector objVector   = new Vector(files.length);
    long lastMod       = 0L;
    File objFile;

    for (int i=0; i<files.length; i++) {
      str = files[i].getAbsolutePath();
      str = changeExtension(str,sourceExt,objectExt);
      if (str != null) {
        objFile = new File(str);
        // not existing source files are handed over to the compiler,
        // which will create a error message.
        if (allFiles ||
            !files[i].exists() ||
            !objFile.exists() ||
            (files[i].lastModified() > objFile.lastModified())) {
          makeVector.addElement(files[i]);
          if (objFile.lastModified() > lastMod)
            lastMod = objFile.lastModified();
          // end if
        } // end if
        if (files[i].exists()) objVector.addElement(objFile);
      } // end if
    } // end for
    files = new File[makeVector.size()];
    for (int i=0; i<files.length; i++) {
      files[i] = (File)makeVector.elementAt(i);
    } // end for
    // start the compiler thread
    curFiles   = files;
    curModTime = lastMod;
    (new Thread(this)).start();
    files = new File[objVector.size()];
    for (int i=0; i<files.length; i++) {
      files[i] = (File)objVector.elementAt(i);
    } // end for
    return files;
  } // end internalMake

  /** Displays a settings dialog. */
  public void settings(Frame parent) {
    SettingsDlg dlg = new SettingsDlg(parent);

    dlg.show();
    if (!dlg.isCanceled()) {
      dlg.storeResults();
      setProjectChanged();
    } // end if
  } // end settings

  /** Start a compiler process. */
  protected Process invokeCompiler(File file) {
    String path         = file.getAbsolutePath();
    String[] extensions = getExtensions();
    boolean isSource    = false;
    String[] cmdArray;
    Process proc;

    // the first half of the extensions refer to classes, the second half
    // refers to html files
    for (int i=0; i<(extensions.length / 2); i++) {
      if (path.endsWith(extensions[i])) {
        isSource = true;
        break;
      } // end if
    } // end for
    if (!isSource) return null;
    cmdArray = buildCmdLine(file);
    try {
      proc = Runtime.getRuntime().exec(cmdArray);
    } catch (IOException excep) {
      postMessage(new ErrorMessage(ErrorMessage.ERROR,
                    "Can't execute \"" + cmdArray[0] +
                    "\": " + excep));
      proc = null;
    }
    return proc;
  } // end invokeCompiler

  /** Return an array containing the command line. */
  protected String[] buildCmdLine(File file) {
    /* PENDING */
    String str;
    StringTokenizer tok;
    int i;
    String path;
    String[] result;

    str  = getSetting(COMP_COMMAND) + " ";
    str += (getSetting(COMP_ADDPARAM) + " ");
    if ((new Boolean(getSetting(COMP_USECLASSPATH))).booleanValue()) {
      str += getSetting(COMP_CLASSPATHPREFIX);
      if ((new Boolean(getSetting(COMP_TRAILINGSPACE))).booleanValue()) {
        str += " ";
      } // end if
      str += getProjectDirectory();
      str += File.pathSeparator;
      str += System.getProperty("java.class.path");
    } // end if
    path = file.getAbsolutePath();
    tok    = new StringTokenizer(str," \n\r\t",false);
    i      = tok.countTokens() + 1;
    result = new String[i];
    i      = 0;
    while (tok.hasMoreTokens()) result[i++] = tok.nextToken();
    result[i] = path;
    if (Utils.Assert.ENABLED) {
      str = "";
      for (i=0; i<result.length; i++) str += (result[i] + " ");
      System.out.println("cmdLine: " + str);
    } // end if
    return result;
  } // end buildCmdLine

  /** Clips a ending from str or return null if ending doesn't exist. */
  private final String changeExtension(String str, String[] oldExt,
                                       String[] newExt) {
    int len = Math.min(oldExt.length,newExt.length);

    for (int i=0; i<len; i++)
      if (str.endsWith(oldExt[i]))
        return str.substring(0,str.length() - oldExt[i].length()) + newExt[i];
    return null;
  } // end changeExtension

  /** Notifies all Observers from compiling a file. */
  protected void notifyCompile(File file, int cur) {
    String str  = file.getAbsolutePath();
    String path = getProjectDirectory();

    if (str.startsWith(path)) {
      str = str.substring(path.length());
      if (str.startsWith(File.separator)) str = str.substring(1);
    } // end if
    notifyCompile(compileLabel,new File(str),cur,curFiles.length);
  } // end notifyCompile

  //--------- Runnable ----------------------------------------------
  public void run() {
    long timeStamp     = System.currentTimeMillis();
    String[] sourceExt = getExtensions();
    String[] objectExt = getObjectExtensions();
    ErrorDesc err      = new ErrorDesc();
    int exitValue      = 0;
    String str;
    long l;

    errors   = 0;
    warnings = 0;
    notifyStartOfCompile();
    for (int i=0; i<curFiles.length; i++) {
      boolean success = true;
      int tmp         = 0;
      Process proc;

      exitValue = 0;
      notifyCompile(curFiles[i],i);
      str = changeExtension(curFiles[i].getAbsolutePath(),
                            sourceExt,objectExt);
      l   = str == null ? 0L : (new File(str)).lastModified();
      // compile only files, that haven't been compiled in a previous pass
      if (l > curModTime) proc = null;
        else proc = invokeCompiler(curFiles[i]);
      if (proc != null) {
        Parser.setObservable(observable);
        // The coco parser sometimes throws an out of memory error
        // on parsing unexpected input.
        try {
          Boolean b;

          b = new Boolean(getSetting(COMP_USEERRORSTREAM));
          new OutputRedirector(proc,!b.booleanValue());
          if (b.booleanValue()) Parser.parse(proc.getErrorStream());
            else Parser.parse(proc.getInputStream());
        } catch (OutOfMemoryError oome) {
          postMessage(new ErrorMessage(ErrorMessage.ERROR,
                        "Error parsing compiler output: " + oome));
        }
        try {
          tmp = proc.waitFor();
          if (tmp != 0) exitValue = tmp;
        } catch (InterruptedException excep) {
          excep.printStackTrace();
          success = false;
        }
        if (!success) proc.destroy();
      } // end if
      if (Utils.Assert.ENABLED)
        System.out.println("Exit value: " + tmp);
      // end if
    } // end for
    // report a error summary
    int type;
    String info;
    String min;

    timeStamp = (System.currentTimeMillis() - timeStamp) / 1000;
    if (errors > 0) type = ErrorMessage.ERROR;
      else if (warnings > 0) type = ErrorMessage.WARNING;
        else type = ErrorMessage.MESSAGE;
    str  = "Compile done; " + errors + " error(s), " +
           warnings + " warning(s).";
    min  = "" + (timeStamp % 60);
    if (min.length() < 2) min = "0" + min;
    info = "Compile statistics: " + curFiles.length +
          " file(s); time: " + (timeStamp / 60) + ":" + min;
    postMessage(new ErrorMessage(type,str,info));
    notifyEndOfCompile(exitValue == 0);
    // write the compiler version to the project properties
    str  = ResourceUtil.getResStringOrKey(resources,COMP_COMMAND);
    str += (" (" + System.getProperty("java.version") + ")");
    if (!str.equals(project.getProperty(Const.COMPILER_VERSION))) {
      project.put(Const.COMPILER_VERSION,str);
      setProjectChanged();
    } // end if
  } // end run

  //--------- Observer ----------------------------------------------
  public void update(Observable o, Object arg) {
    if (arg instanceof int[]) {
      int[] aInt = (int[])arg;

      Utils.Assert.assert(Utils.Assert.ENABLED &&
        (aInt.length == 2),"Wrong argument length");
      errors   += aInt[0];
      warnings += aInt[1];
    }
    else { // ErrorDesc
      ErrorDesc ed = (ErrorDesc)arg;
      ErrorMessage em;
      int type;

      if (ed.isError()) type = ErrorMessage.ERROR;
        else if (ed.isWarning()) type = ErrorMessage.WARNING;
          else type = ErrorMessage.MESSAGE;
      if (ed.hasFileAndLine()) {
        File path = new File(getProjectDirectory());

        em = new LocatableErrorMessage(type,ed.msg,path,ed.file,ed.line,
                                       ed.code);
      }
      else {
        em = new ErrorMessage(type,ed.msg);
      } // end if
      postMessage(em);
    } // end if
  } // end update

  private final String getSetting(String key) {
    String str = project.getProperty(key);

    if (str != null) return str;
    str = ResourceUtil.getResourceString(resources,key);
    if (str != null) return str;
    return "";
  } // end getSetting

  // inner classes
  /** A dialog that changes the settings of this class. */
  class SettingsDlg extends BaseDialog {

    // instance variable declarations
    protected TextField tfPath;
    protected TextField tfParam;
    protected Checkbox cbUseCP;
    protected TextField tfPrefix;
    protected Checkbox cbSpaces;

    // constructors
    /** Create a new SettingsDlg object. */
    public SettingsDlg(Frame parent) {
      super(parent,true);

      Panel panel = getContentPanel();
      Label label;
      String str;
      int i;

      setTitle(ResourceUtil.getResStringOrKey(resources,SETTDLG_TITLE));
      createButtons(ResourceUtil.getResStringOrKey(resources,SETTDLG_BUTTONS));

      GridBagLayout gridbag          = new GridBagLayout();
      GridBagConstraints constraints = new GridBagConstraints();
      panel.setLayout(gridbag);
      constraints.insets     = new Insets(10,6,10,6);
      constraints.anchor     = GridBagConstraints.WEST;
      constraints.weightx    = 1.0;
      constraints.weighty    = 1.0;
      constraints.gridwidth  = 1;
      constraints.gridheight = 1;
      label = new Label(getSetting(SETTDLG_PATH));
      gridbag.setConstraints(label,constraints);
      panel.add(label);
      constraints.insets     = new Insets(10,0,10,6);
      tfPath = new TextField(getSetting(COMP_COMMAND),15);
      gridbag.setConstraints(tfPath,constraints);
      panel.add(tfPath);
      constraints.gridwidth  = GridBagConstraints.RELATIVE;
      label = new Label(getSetting(SETTDLG_ADDPARA));
      gridbag.setConstraints(label,constraints);
      panel.add(label);
      constraints.gridwidth  = GridBagConstraints.REMAINDER;
      tfParam = new TextField(getSetting(COMP_ADDPARAM),15);
      gridbag.setConstraints(tfParam,constraints);
      panel.add(tfParam);
      constraints.insets     = new Insets(0,6,10,6);
      constraints.gridwidth  = 1;
      constraints.gridheight = GridBagConstraints.RELATIVE;
      label = new Label(getSetting(SETTDLG_PREFIX));
      gridbag.setConstraints(label,constraints);
      panel.add(label);
      constraints.insets     = new Insets(0,0,10,6);
      tfPrefix = new TextField(getSetting(COMP_CLASSPATHPREFIX),15);
      gridbag.setConstraints(tfPrefix,constraints);
      panel.add(tfPrefix);
      constraints.gridwidth  = GridBagConstraints.REMAINDER;
      cbSpaces = new Checkbox(getSetting(SETTDLG_SPACES),
                   (new Boolean(getSetting(COMP_TRAILINGSPACE))).booleanValue());
      gridbag.setConstraints(cbSpaces,constraints);
      panel.add(cbSpaces);
      constraints.insets     = new Insets(0,6,10,6);
      constraints.gridwidth  = 1;
      constraints.gridheight = GridBagConstraints.REMAINDER;
      cbUseCP = new Checkbox(getSetting(SETTDLG_USEPATH),
                  (new Boolean(getSetting(COMP_USECLASSPATH))).booleanValue());
      gridbag.setConstraints(cbUseCP,constraints);
      panel.add(cbUseCP);

      pack();
      center(parent);
      setResizable(false);
    } // end constructor SettingsDlg

    /** Write the dialog controls content to the project properties. */
    public void storeResults() {
      project.put(COMP_COMMAND,tfPath.getText());
      project.put(COMP_ADDPARAM,tfParam.getText());
      project.put(COMP_USECLASSPATH,"" + cbUseCP.getState());
      project.put(COMP_CLASSPATHPREFIX,tfPrefix.getText());
      project.put(COMP_TRAILINGSPACE,"" + cbSpaces.getState());
    } // end storeResults

  } // end SettingsDlg

} // end DefaultCompiler

/* ******************************************************************
   end of file
*********************************************************************/
