/* ******************************************************************

   class      : DefaultHelpWindow
   date       : 20.05.98
   compiler   : javac (jdk 1.1.6)

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;

import javax.swing.*;

import Utils.ResourceUtil;
import Utils.FrameUtil;
import CustCtrl.BaseDialog;
import CustCtrl.SeparatorCanvas;
import CustCtrl.ComboTextField;
import SysExt.OutputRedirector;

/* ******************************************************************
   top level class declaration
*********************************************************************/

/**
 * A help window implementation that uses html as help text.
 *
 * @author Puehringer Edgar
 */
public class DefaultHelpWindow extends AbstractHelpWindow
  implements ActionListener, Runnable {

  // class variable declarations
  /** Key to api refernence root path. */
  public static final String REF_PATH = "help.apiRefPath";
  /** Key to a flag, that indicates if the api reference is absolute. */
  public static final String REF_ABS = "help.apiRefAbsolute";
  /** Key to the help root path. */
  public static final String HELP_PATH = "help.helpPath";
  /** Key to filename of help index */
  public static final String INDEX_NAME = "help.helpIndexFile";
  /** Key to filename of help fast index */
  public static final String FAST_INDEX = "help.fastIndexFile";
  /** The default extension of help fast index files. */
  public static final String DEF_INDEX_EXT = ".whi";
  /** Key to number of max. displayed search results. */
  public static final String MAX_HITS = "help.maxHits";
  /** Key to the help window title */
  public static final String TITLE = "help.title";
  /** Key to the filename of the help window icon */
  public static final String ICON_IMAGE = "help.iconImage";
  /** Key to the help window ok button. */
  public static final String  BUTTON_OK = "help.labelOK";
  /** Key to the help window cancel button. */
  public static final String  BUTTON_CANCEL = "help.labelCancel";
  /** Key to the toolbar resource */
  public static final String TOOLBAR = "help.toolbar";
  /** Key to file not found messagebox text */
  public static final String FILE_NOT_FOUND = "help.fileNotFound";
  /** Key to keyword not found messagebox text */
  public static final String KEYW_NOT_FOUND = "help.keywordNotFound";
  /** Key to continue search messagebox text */
  public static final String CONTINUE_SEARCH = "help.continueSearch";
  /** Key to io exception messagebox text */
  public static final String IO_EXCEPTION = "help.ioException";

  /** Key to the selected applet viewer resource. */
  private final static String SELECTED_VIEWER = "exec.viewer";

  // Keys to a java enabled html browser settings
  public final static String BR_COMMAND         = "help.br.command";
  public final static String BR_ADDPARAM        = "help.br.addParam";

  // Keys to the settings dialog
  public final static String SETTDLG_TITLE    = "help.settingsDlg.title";
  public final static String SETTDLG_BUTTONS  = "help.settingsDlg.buttons";
  public final static String SETTDLG_VLABEL   = "help.settingsDlg.viewerLabel";
  public final static String SETTDLG_VCHOICE  = "help.settingsDlg.viewerChoice";
  public final static String SETTDLG_BRTITLE  = "help.settingsDlg.brTitle";
  public final static String SETTDLG_PATH     = "help.settingsDlg.path";
  public final static String SETTDLG_ADDPARA  = "help.settingsDlg.addPara";
  public final static String SETTDLG_HPTITLE  = "help.settingsDlg.helpPathTitle";
  public final static String SETTDLG_HELPPATH = "help.settingsDlg.helpPath";
  public final static String SETTDLG_REF_PATH = "help.settingsDlg.refPath";
  public final static String SETTDLG_REFINDEX = "help.settingsDlg.refIndex";
  public final static String SETTDLG_RCHOICE  = "help.settingsDlg.refIndexChoice";

  // instance variable declarations
  private JFrame frame = null;
  private JPanel panel;
  private CardLayout layout;
  private JPanel htmlPanel;
  private JToolBarLoader toolbarLoader;
  private JToolBar toolbar;
  private HtmlPanel html = null;
  private Hashtable commands;
  private String apiRefRoot;
  private String helpRoot;
  private int maxHits;
  private Hashtable ht;
  private JPanel listPanel = null;
  private DefaultListModel listModel;
  private final JList list = new JList();
  private MouseListener mouseListener;
  private JScrollPane listScroll;
  private JButton okButton;
  private JButton cancelButton;
  private HtmlScanner scan;
  private String keyword;
  private Cursor oldCursor;
  private PrintWriter fastIndexWriter    = null;
  private BufferedReader fastIndexReader = null;
  private boolean fastIndexAvailable;
  private int numHits = 0;

  // constructors
  /** Create a new DefaultHelpWindow object. */
  public DefaultHelpWindow(ResourceBundle resources, Properties desktop) {
    super(resources,desktop);

    String str;
    boolean b;

    // load the root path of the api reference
    b          = Boolean.valueOf(desktop.getProperty(REF_ABS)).booleanValue();
    apiRefRoot = b ? "" : System.getProperty("java.home");
    str        = desktop.getProperty(REF_PATH);
    if (str == null) {
      str = ResourceUtil.getResourceString(resources,REF_PATH);
      b   = false;
    } // end if
    if (str != null) {
      if (!b) apiRefRoot += File.separatorChar;
      apiRefRoot += str;
    } // end if
    apiRefRoot = apiRefRoot.replace('/',File.separatorChar);
    // load the root path of the application help
    helpRoot = desktop.getProperty(HELP_PATH);
    if (helpRoot == null) {
      helpRoot = System.getProperty("user.dir");
      str = ResourceUtil.getResourceString(resources,HELP_PATH);
      if (str != null) helpRoot += (File.separatorChar + str);
      helpRoot = ResourceUtil.fileToURL(helpRoot);
      desktop.put(HELP_PATH,helpRoot);
    } // end if
    if (!helpRoot.endsWith("/")) helpRoot += "/";
    maxHits = ResourceUtil.getResourceInt(resources,MAX_HITS,10);
    ht = new Hashtable();
  } // end constructor DefaultHelpWindow

  // methods
  /** Show help for a specified keyword. */
  public void showHelpOnKeyword(String keyword) {
    File file = getIndexFile(keyword,INDEX_NAME);
    File fast = getIndexFile(keyword,FAST_INDEX);

    this.keyword = keyword;
    if (!file.canRead()) {
      String text,title;
      try {
        text = resources.getString(FILE_NOT_FOUND);
      } catch (MissingResourceException excep) { text = FILE_NOT_FOUND; }
      try {
        title = resources.getString(TITLE);
      } catch (MissingResourceException excep) { title = TITLE; }
      text = "\"" + file + "\" " + text;
      JOptionPane.showMessageDialog(frame,text,title,JOptionPane.ERROR_MESSAGE);
      return;
    } // end if
    numHits            = 0;
    fastIndexAvailable = fast.canRead() &&
                         (fast.lastModified() >= file.lastModified());
    if (fastIndexAvailable) {
      try {
        fastIndexReader = new BufferedReader(new FileReader(fast));
      } catch (IOException excep) {
        excep.printStackTrace();
        fastIndexReader    = null;
        fastIndexAvailable = false;
      }
    } // end if
    if (!fastIndexAvailable) {
      scan = new HtmlScanner(file);
      try {
        fastIndexWriter = new PrintWriter(new FileWriter(fast));
      } catch (IOException excep) {
        excep.printStackTrace();
        fastIndexWriter = null;
      }
    } // end if
    ht.clear();
    if (listPanel == null) {
      makeList();
    }
    else {
      if (!frame.isVisible()) frame.setVisible(true);
      frame.toFront();
      frame.requestFocus();
    } // end if
    okButton.setEnabled(false);
    list.removeMouseListener(mouseListener);
    listModel.removeAllElements();
    layout.show(panel,"list");
    oldCursor = frame.getCursor();
    list.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
    SwingUtilities.invokeLater(this);
  } // end showHelpOnKeyword

  /** Conpletes an url to be displayable by the help system. */
  protected URL completeURL(String relPath) {
    URL url;

    try {
      if (relPath.startsWith("ref:")) {
        String str = apiRefRoot + File.separator + relPath.substring(4);

        str = ResourceUtil.fileToURL(str);
        url = new URL(str);
      }
      else {
        url = new URL(helpRoot + relPath);
      } // end if
    } catch (MalformedURLException excep) {
      url = null;
      Toolkit.getDefaultToolkit().beep();
    }
    return url;
  } // end completeURL

  /** Shows the content of the specified document. */
  protected void showContent(String relPath) {
    URL url = completeURL(relPath);
    String[] cmdLine;

    if (url == null) return;
    // view with browser
    cmdLine = buildCmdLine(url.toString());
    if (cmdLine != null) {
      executeBrowser(cmdLine);
      return;
    } // end if
    // or view with internal viewer
    if (html == null) {
      makeHtml(url);
    }
    else {
      if (!frame.isVisible()) frame.setVisible(true);
      html.setPage(url);
    } // end if
    layout.show(panel,"html");
  } // end showContent

  /** Execute the browser. */
  protected void executeBrowser(String[] cmdArray) {
    try {
      Process proc = Runtime.getRuntime().exec(cmdArray);
      new OutputRedirector(proc,false);
      new OutputRedirector(proc,true);
    } catch (IOException excep) {
      ioExceptionMessage(excep);
    }
  } // end executeBrowser

  /** Displays a settings dialog. */
  public void settings(Frame parent) {
    SettingsDlg dlg = new SettingsDlg(parent);

    dlg.show();
    if (!dlg.isCanceled()) dlg.storeResults();
  } // end settings

  /**
   * Return an array containing the command line.  Return null
   * if the internal viewer should be executed.
   */
  protected String[] buildCmdLine(String url) {
    String str;
    int selection;
    StringTokenizer tok;
    int i,j;
    String[] result;

    str = getSetting(SELECTED_VIEWER);
    if (str.equals("")) str = "0";
    selection = Integer.parseInt(str);
    if (selection == 0) return null; // if internal viewer
    str    = getSetting(BR_COMMAND) + " ";
    str   += getSetting(BR_ADDPARAM);
    tok    = new StringTokenizer(str," \n\r\t",false);
    i      = tok.countTokens() + 1;
    result = new String[i];
    i      = 0;
    while (tok.hasMoreTokens()) result[i++] = tok.nextToken();
    result[i] = url;
    if (Utils.Assert.ENABLED) {
      str = "";
      for (i=0; i<result.length; i++) str += (result[i] + " ");
      System.out.println("cmdLine: " + str);
    } // end if
    return result;
  } // end buildCmdLine

  /** Called if a help window instance is destroyed. */
  protected void quitInstance() {
    if (frame != null) {
      frame.setVisible(false);
      frame.dispose();
    } // end if
  } // quitInstance

  /** Extracts the description of an anchor (link). */
  private String getAnchorDesc(HtmlScanner scan) {
    String str    = "";
    String curTok = "";

    while (!curTok.equals("/a") && scan.hasMoreTokens()) {
      curTok = scan.nextToken();
      if (!scan.isTag()) str += curTok;
    } // end while
    return str;
  } // end getAnchorDesc

  private void makeHtml(URL url) {
    if (frame == null) makeWindow();
    html = new HtmlPanel(url) {

      protected void reportError(IOException excep) {
        ioExceptionMessage(excep);
      } // end reportError

    };
    html.setDefaultPage(completeURL(getContext("help-on-help")),0);
    html.setBackground(SystemColor.control);
    // install the command table
    commands = new Hashtable();
    Action[] actions = html.getActions();
    for (int i = 0; i < actions.length; i++) {
      Action a = actions[i];
      commands.put(a.getValue(Action.NAME),a);
    } // end for
    htmlPanel = new JPanel(new BorderLayout());
    htmlPanel.add("Center",html);
    toolbarLoader = new SmartToolBarLoader(resources);
    toolbarLoader.setCommands(commands);
    toolbar       = toolbarLoader.createToolBar(TOOLBAR);
    htmlPanel.add("South",toolbar);
    panel.add("html",htmlPanel);
    frame.setVisible(true);
  } // end makeHtml

  private void makeList() {
    String str;

    if (frame == null) makeWindow();
    listModel = new DefaultListModel();
    list.setModel(listModel);
    list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    mouseListener = new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() == 2) {
          actionPerformed(new ActionEvent(e.getSource(),
                            ActionEvent.ACTION_PERFORMED,"",
                            e.getModifiers()));
        } // end if
      } // end mouseClicked
    }; // end anonymous class
    listScroll = new JScrollPane(list);
    listPanel  = new JPanel(new BorderLayout());
    listPanel.add("Center",listScroll);
    JPanel btnPanel = new JPanel();
    btnPanel.setBackground(SystemColor.control);
    str = ResourceUtil.getResStringOrKey(resources,BUTTON_OK);
    okButton = new JButton(str);
    okButton.addActionListener(this);
    btnPanel.add(okButton);
    str = ResourceUtil.getResStringOrKey(resources,BUTTON_CANCEL);
    cancelButton = new JButton(str);
    cancelButton.addActionListener(this);
    btnPanel.add(cancelButton);
    listPanel.add("South",btnPanel);
    panel.add("list",listPanel);
    frame.setVisible(true);
  } // end makeList

  private void makeWindow() {
    frame = new JFrame();
    try {
      frame.setTitle(resources.getString(TITLE));
    } catch (MissingResourceException e) { frame.setTitle(TITLE); }
    frame.getContentPane().setLayout(new BorderLayout());
    layout = new CardLayout();
    panel  = new JPanel(layout);
    frame.getContentPane().add(panel);
    frame.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
    frame.pack();
    frame.setSize(600,400);
    FrameUtil.shuffleLocation(frame);
    try {
      URL u = this.getClass().getResource(resources.getString(ICON_IMAGE));
      if (u != null) frame.setIconImage((new ImageIcon(u)).getImage());
    } catch (MissingResourceException mre) {
      // just don't use an icon
    }
  } // end makeWindow

  /** Return a index file to a given keyword. */
  protected File getIndexFile(String keyword, String resourceName) {
    String str;
    char ch;
    int pos;
    File f;

    if (keyword.length() < 1) return null;
    str = desktop.getProperty(resourceName);
    if (str == null)
      str = ResourceUtil.getResourceString(resources,resourceName);
    // end if
    if (str == null) return null;
    ch = Character.toLowerCase(keyword.charAt(0));
    str = str.replace('?',ch);
    pos = str.indexOf('#');
    if (pos >= 0) {
      int num;

      if (ch == '_') num = 27;
        else num = (int)ch + 1 - (int)'a';
      do {
        if (pos == str.length() - 1)
          str = str.substring(0,pos) + num;
        else
          str = str.substring(0,pos) + num + str.substring(pos + 1);
        // end if
        pos = str.indexOf('#');
      } while (pos >= 0);
    } // end if
    str = str.replace('/',File.separatorChar);
    f = new File(apiRefRoot,str);
    return f;
  } // end getIndexFile

  // ---- ActionListener implementation --------------------------
  /** Invoked when an action occurs. */
  public void actionPerformed(ActionEvent e) {
    String str = (String)list.getSelectedValue();
    URL url;
    String[] cmdLine;

    if (e.getSource() == cancelButton) {
      if (frame != null) frame.setVisible(false);
      return;
    } // end if
    str = (String)ht.get(str);
    str = ResourceUtil.fileToURL(apiRefRoot + File.separator +
                                 str.replace('/',File.separatorChar));
    try {
      url = new URL(str);
      // view with browser
      cmdLine = buildCmdLine(url.toString());
      if (cmdLine != null) {
        executeBrowser(cmdLine);
        return;
      } // end if
      // or view with internal viewer
      if (html == null) {
        makeHtml(url);
      }
      else {
        frame.toFront();
        frame.requestFocus();
        html.setPage(url);
      } // end if
      layout.show(panel,"html");
    } catch (MalformedURLException excep) {
      System.err.println("Malformed URL: " + str + " " + excep);
      Toolkit.getDefaultToolkit().beep();
    }
  } // end actionPerformed

  // ---- Runnable implementation --------------------------------
  /** This run method is called to scan the help topic index. */
  public void run() {
    if (fastIndexAvailable) runFast();
      else runSlow();
  } // end run

  private void runFast() {
    String str,desc,link;
    int index;

    try {
      str = fastIndexReader.readLine();
      if ((str != null) && (str.length() >= keyword.length())) {
        String tmp = str.substring(0,keyword.length()).toLowerCase();
        if (tmp.compareTo(keyword.toLowerCase()) > 0) {
          if (Utils.Assert.ENABLED)
            System.out.println("Stopping search at " + str);
          // end if
          str = null;
        } // end if
      } // end if
    } catch (IOException excep) {
      excep.printStackTrace();
      str = null;
    }
    while (str != null) {
      if (str.startsWith(keyword)) { // Yep, we have it!
        index = str.indexOf((int)'>');
        desc  = str.substring(0,index);
        str   = str.substring(index+1);
        index = str.indexOf((int)'>');
        desc += str.substring(0,index);
        link  = str.substring(index+1);
        ht.put(desc,link);
        listModel.addElement(desc);
        numHits++;
        if ((numHits % maxHits != 0) || queryContinue()) {
          SwingUtilities.invokeLater(this);
          return;
        } // end if
        break;
      } // end if
      try {
        str = fastIndexReader.readLine();
      } catch (IOException excep) {
        excep.printStackTrace();
        str = null;
      }
    } // end while
    checkResult();
    try {
      fastIndexReader.close();
    } catch (IOException excep) {
      excep.printStackTrace();
    }
    // garbage collect
    fastIndexReader = null;
  } // end runSlow

  private void runSlow() {
    String key,desc,link;
    String str    = "";
    String oldTag = "";

    while (scan.hasMoreTokens()) {
      if (scan.isTag()) oldTag = str;
      str = scan.nextToken();
      if (scan.isTag() && oldTag.equals("dt") && str.startsWith("a ")) {
        link   = scan.getQuotedString();
        key    = getAnchorDesc(scan);
        desc   = "";
        oldTag = "/a";
        if (scan.hasMoreTokens()) {
          str = scan.nextToken();
          if (scan.isTag()) oldTag = str;
            else desc = str;
          if (scan.hasMoreTokens()) {
            str = scan.nextToken();
            if (scan.isTag()) {
              oldTag = str;
              if (str.startsWith("a ")) { // if it is an achor
                desc += getAnchorDesc(scan);
              } // end if
            } // end if
          } // end if
        } // end if
        if (fastIndexWriter != null)
          fastIndexWriter.println(key+'>'+desc+'>'+link);
        // end if
        if (key.startsWith(keyword)) { // Yep, we have it!
          ht.put(key+desc,link);
          listModel.addElement(key+desc);
          numHits++;
          SwingUtilities.invokeLater(this);
          return;
        } // end if
      } // end if
    } // end while
    checkResult();
    scan.close();
    if (fastIndexWriter != null) fastIndexWriter.close();
    // garbage collect
    scan            = null;
    fastIndexWriter = null;
  } // end runSlow

  private void checkResult() {
    if (listModel.isEmpty()) {
      String text;
      try {
        text = resources.getString(KEYW_NOT_FOUND);
      } catch (MissingResourceException excep) { text = KEYW_NOT_FOUND; }
      text = text + "\" " + keyword + "\".";
      JOptionPane.showMessageDialog(frame,text,frame.getTitle(),
                                    JOptionPane.INFORMATION_MESSAGE);
    }
    else {
      if (list.isSelectionEmpty()) list.setSelectedIndex(0);
      okButton.setEnabled(true);
      list.addMouseListener(mouseListener);
    } // end if
    list.setCursor(oldCursor);
    oldCursor = null;
  } // end checkResult

  private boolean queryContinue() {
      String str;

      str = ResourceUtil.getResStringOrKey(resources,CONTINUE_SEARCH);
      str = numHits + " " + str;
      switch (JOptionPane.showConfirmDialog(frame,str,frame.getTitle(),
                JOptionPane.YES_NO_OPTION,JOptionPane.QUESTION_MESSAGE)) {
        case JOptionPane.YES_OPTION: return true;
        default:
      } // end switch
      return false;
  } // end queryContinue

  /** Displays a io exception in a messagebox. */
  private void ioExceptionMessage(IOException excep) {
    String text,title;

    text  = ResourceUtil.getResStringOrKey(resources,IO_EXCEPTION);
    title = ResourceUtil.getResStringOrKey(resources,TITLE);
    text += ("\n" + excep);
    JOptionPane.showMessageDialog(frame,text,title,
                                  JOptionPane.ERROR_MESSAGE);
  } // end ioExceptionMessage

  private final String getSetting(String key) {
    String str = desktop.getProperty(key);

    if (str != null) return str;
    str = ResourceUtil.getResourceString(resources,key);
    if (str != null) return str;
    return "";
  } // end getSetting

  // inner classes
  /**
   * A scanner for the index file, assume that it is a html file.
   */
  protected final class HtmlScanner {

    // instance variable declarations
    private Reader in;
    private BufferedReader bin;
    private int iChar;
    private char ch          = ' ';
    private boolean tag      = false;
    private String quotedStr = null;

    // constructors
    HtmlScanner(File f) {
      if (f == null) {
        iChar = -1;
      }
      else {
        try {
          in = new FileReader(f);
          bin = new BufferedReader(in);
          readChar();
        } catch (FileNotFoundException excep) {
          excep.printStackTrace();
          iChar = -1;
        }
      } // end if
    } // end constructor HtmlScanner

    // methods
    /** Reads a character, skip line feed and cr. */
    void readChar() {
      try {
        do {
          iChar = bin.read();
        } while ((iChar == (int)'\n') || (iChar == (int)'\r'));
        if (iChar >= 0) ch = (char)iChar;
      } catch (IOException excep) {
        excep.printStackTrace();
        iChar = -1;
      }
    } // end readChar

    /** Returns true if this stream has more tokens. */
    boolean hasMoreTokens() {
      return (iChar >= 0);
    } // end hasMoreTokens

    /** Return the next token. */
    String nextToken() {
      if (iChar < 0) throw new NoSuchElementException();
      switch (ch) {
        case '<':
          tag = true;
          return scanTag();
        default:
          tag = false;
          return scanText();
      } // end switch
    } // end nextToken

    /** Frees all resources used by html scanner. */
    void close() {
      try { bin.close(); } catch (IOException e) {}
      try { in.close();  } catch (IOException e) {}
      iChar = -1;
    } // end close

    /** Scan a tag. */
    private String scanTag() {
      String str = "";

      quotedStr = null;
      readChar(); // overread the left bracket
      if (iChar < 0) return "";
      do {
        if (ch == '"') {
          str = str + '"' + scanString() + '"';
        }
        else {
          str += Character.toLowerCase(ch);
          readChar();
        } // end if
      } while ((iChar >= 0) && (ch != '>'));
      if (iChar >= 0) readChar(); // if iChar is right bracket
      return str;
    } // end scanTag

    /** Scan a quoted string. */
    private String scanString() {
      quotedStr = "";
      readChar(); // overread the first quote
      if (iChar < 0) return "";
      do {
        quotedStr += ch;
        readChar();
      } while ((iChar >= 0) && (ch != '"'));
      if (iChar >= 0) readChar(); // if iChar is quote character
      return quotedStr;
    } // end scanString

    /** Scan a portion of normal text. */
    private String scanText() {
      String str = "";

      do {
        str += ch;
        readChar();
      } while ((iChar >= 0) && (ch != '<'));
      return str;
    } // end scanText

    /** Returns true if the last token was a tag. */
    boolean isTag() {
      return tag;
    } // end isTag

    /** Return the quoted string in the last tag or null, if there was no. */
    String getQuotedString() {
      return quotedStr;
    } // end getQuotedString

  } // end HtmlScanner

  /** A dialog that changes the settings of this class. */
  class SettingsDlg extends BaseDialog implements ActionListener {

    // instance variable declarations
    protected Choice choice;
    protected ComboTextField brPath;
    protected TextField brParam;
    protected ComboTextField tfHelpPath;
    protected ComboTextField tfRefPath;
    protected ComboTextField tfRefIndex;
    private Frame parent;

    // constructors
    /** Create a new SettingsDlg object. */
    public SettingsDlg(Frame parent) {
      super(parent,true);
      this.parent = parent;

      Panel panel = getContentPanel();
      Label label;
      SeparatorCanvas separator;
      String[] options;
      String str;
      int i;

      setTitle(ResourceUtil.getResStringOrKey(resources,SETTDLG_TITLE));
      createButtons(ResourceUtil.getResStringOrKey(resources,SETTDLG_BUTTONS));

      GridBagLayout gridbag          = new GridBagLayout();
      GridBagConstraints constraints = new GridBagConstraints();
      panel.setLayout(gridbag);
      constraints.insets     = new Insets(10,6,10,6);
      constraints.anchor     = GridBagConstraints.WEST;
      constraints.weightx    = 1.0;
      constraints.weighty    = 1.0;
      constraints.gridwidth  = 1;
      constraints.gridheight = 1;
      label = new Label(getSetting(SETTDLG_VLABEL));
      gridbag.setConstraints(label,constraints);
      panel.add(label);
      constraints.insets     = new Insets(10,0,10,6);
      constraints.gridwidth  = GridBagConstraints.REMAINDER;
      choice  = new Choice();
      options = ResourceUtil.splitToWords(getSetting(SETTDLG_VCHOICE));
      for (i=0; i<options.length; i++) choice.add(options[i]);
      str = getSetting(SELECTED_VIEWER);
      if (str.equals("")) str = "0";
      i = Integer.parseInt(str);
      if (i < choice.getItemCount()) choice.select(i);
      gridbag.setConstraints(choice,constraints);
      panel.add(choice);
      constraints.insets     = new Insets(0,0,10,0);
      constraints.fill       = GridBagConstraints.HORIZONTAL;
      separator = new SeparatorCanvas();
      gridbag.setConstraints(separator,constraints);
      panel.add(separator);
      // HTML browser settings
      constraints.insets     = new Insets(0,6,10,6);
      constraints.fill       = GridBagConstraints.NONE;
      label = new Label(getSetting(SETTDLG_BRTITLE));
      gridbag.setConstraints(label,constraints);
      panel.add(label);
      constraints.insets     = new Insets(0,12,10,6);
      constraints.gridwidth  = 1;
      label = new Label(getSetting(SETTDLG_PATH));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets = new Insets(0,0,10,6);
      brPath = new ComboTextField(getSetting(BR_COMMAND),12);
      brPath.addActionListener(this);
      gridbag.setConstraints(brPath,constraints);
      panel.add(brPath);
      constraints.gridwidth  = GridBagConstraints.RELATIVE;
      label = new Label(getSetting(SETTDLG_ADDPARA));
      gridbag.setConstraints(label,constraints);
      panel.add(label);
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      brParam = new TextField(getSetting(BR_ADDPARAM),15);
      gridbag.setConstraints(brParam,constraints);
      panel.add(brParam);
      constraints.insets     = new Insets(0,0,10,0);
      constraints.fill       = GridBagConstraints.HORIZONTAL;
      separator = new SeparatorCanvas();
      gridbag.setConstraints(separator,constraints);
      panel.add(separator);

      // Help path settings
      constraints.insets     = new Insets(0,6,10,6);
      constraints.fill       = GridBagConstraints.NONE;
      label = new Label(getSetting(SETTDLG_HPTITLE));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets    = new Insets(0,12,10,6);
      constraints.gridwidth = 1;
      label = new Label(getSetting(SETTDLG_HELPPATH));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets    = new Insets(0,0,10,6);
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      tfHelpPath = new ComboTextField(ResourceUtil.urlToFile(helpRoot),12);
      tfHelpPath.setActionCommand(HELP_PATH);
      tfHelpPath.addActionListener(this);
      gridbag.setConstraints(tfHelpPath,constraints);
      panel.add(tfHelpPath);

      constraints.insets     = new Insets(0,12,10,6);
      constraints.gridwidth  = 1;
      constraints.gridheight = GridBagConstraints.RELATIVE;
      label = new Label(getSetting(SETTDLG_REF_PATH));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets    = new Insets(0,0,10,6);
      tfRefPath = new ComboTextField(apiRefRoot,12);
      tfRefPath.setActionCommand(REF_PATH);
      tfRefPath.addActionListener(this);
      gridbag.setConstraints(tfRefPath,constraints);
      panel.add(tfRefPath);

      constraints.insets     = new Insets(0,12,10,6);
      constraints.gridwidth  = GridBagConstraints.RELATIVE;
      label = new Label(getSetting(SETTDLG_REFINDEX));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets    = new Insets(0,0,10,6);
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      tfRefIndex = new ComboTextField(getSetting(INDEX_NAME),12);
//      tfRefIndex.setActionCommand(INDEX_NAME);
      tfRefIndex.setLabel("v");
      options = ResourceUtil.splitToWords(getSetting(SETTDLG_RCHOICE));
      for (i=0; i<options.length; i++) tfRefIndex.add(options[i]);
      gridbag.setConstraints(tfRefIndex,constraints);
      panel.add(tfRefIndex);

      constraints.insets     = new Insets(0,0,10,0);
      constraints.gridheight = GridBagConstraints.REMAINDER;
      constraints.fill       = GridBagConstraints.HORIZONTAL;
      separator = new SeparatorCanvas();
      gridbag.setConstraints(separator,constraints);
      panel.add(separator);
      pack();
      center(parent);
      setResizable(false);
    } // end constructor SettingsDlg

    /** Write the dialog controls content to the desktop properties. */
    public void storeResults() {
      boolean b;
      int i;
      String str;

      desktop.put(SELECTED_VIEWER,"" + choice.getSelectedIndex());
      desktop.put(BR_COMMAND,brPath.getText());
      desktop.put(BR_ADDPARAM,brParam.getText());
      helpRoot = ResourceUtil.fileToURL(tfHelpPath.getText());
      desktop.put(HELP_PATH,helpRoot);
      if (!helpRoot.endsWith("/")) helpRoot = helpRoot + "/";
      apiRefRoot = tfRefPath.getText();
      if (!(new File(apiRefRoot)).isAbsolute())
        apiRefRoot = System.getProperty("java.home") + File.separatorChar + apiRefRoot;
      // end if
      b   = !apiRefRoot.startsWith(System.getProperty("java.home"));
      if (b) {
        str = apiRefRoot.replace(File.separatorChar,'/');
      }
      else {
        str = apiRefRoot.substring(System.getProperty("java.home").length());
        str = str.replace(File.separatorChar,'/');
        if (str.startsWith("/")) str = str.substring(1);
      } // end if
      desktop.put(REF_PATH,str);
      desktop.put(REF_ABS,"" + b);
      if (tfRefIndex.getText().length() > 0) {
        desktop.put(INDEX_NAME,tfRefIndex.getText());
        str = ResourceUtil.getResourceString(resources,INDEX_NAME);
        i   = str == null ? 0 : str.length();
        str = ResourceUtil.getResourceString(resources,FAST_INDEX);
        if (str.length() > i) str = str.substring(i);
          else str = DEF_INDEX_EXT;
        desktop.put(FAST_INDEX,tfRefIndex.getText() + str);
      } // end if
    } // end storeResults

    // --- ActionListener ---------------------------------------------
    /** Invoked when an action occurs. */
    public void actionPerformed(ActionEvent e) {
      File file;
      String name,path;
      FileDialog dlg;
      File result = null;
      ComboTextField tf;
      String cmd = e.getActionCommand();

      if (cmd.equals(HELP_PATH)) {
        tf = tfHelpPath;
      } else if (cmd.equals(REF_PATH)) {
        tf = tfRefPath;
      } else if (cmd.equals(INDEX_NAME)) {
        tf = tfRefIndex;
      } else {
        tf = brPath;
      } // end if
      file = new File(tf.getText());
      dlg  = new FileDialog(parent);
      if (file.isDirectory()) {
        dlg.setDirectory(file.getAbsolutePath());
      }
      else {
        dlg.setDirectory(file.getParent());
        if (file.getName().indexOf("?") >= 0) dlg.setFile("*.html");
          else dlg.setFile(file.getName());
      } // end if
      dlg.show();
      name = dlg.getFile();
      if (name != null) {
        path = dlg.getDirectory();
        if (path == null) result = new File(name);
          else result = new File(path,name);
      } // end if
      if (result == null) return;
      if (tf == brPath) {
        tf.setText(result.getAbsolutePath());
      } else if (tf == tfRefIndex) {
        if (result.isDirectory()) getToolkit().beep();
          else tf.setText(result.getName());
      } else {
        if (result.isDirectory()) {
          tf.setText(result.getAbsolutePath());
        }
        else {
          String str = result.getParent();

          if (str != null) tf.setText(str);
        } // end if
      } // end if
    } // end actionPerformed

  } // end SettingsDlg

} // end DefaultHelpWindow

/* ******************************************************************
   end of file
*********************************************************************/
