/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.6)

   class      : EditorActionsSet
   date       : 18.07.98
   description: A collection of frequently used editor
                file I/O actions.

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

import javax.swing.*;
import javax.swing.text.TextAction;

import Utils.ResourceUtil;
import Dialogs.AutoSaveDlg;
import Dialogs.ErrorDlg;

/* ******************************************************************
   top level class declaration
*********************************************************************/

public class EditorActionsSet {

  // instance variable declarations
  private AbstractEditor editor;
  private FileDialog fileDialog;
  private File currentFile;
  /** Actions defined by the EditorActionsSet class. */
  private OpenAction openAction     = new OpenAction();
  private SaveAction saveAction     = new SaveAction();
  private RevertAction revertAction = new RevertAction();
  private Action[] actions = {
    new NewAction(false),
    new NewAction(true),
    openAction,
    saveAction,
    new SaveAsAction(),
    revertAction,
    new AutoSaveAction()
  };

  // constructors
  public EditorActionsSet(AbstractEditor editor) {
    this.editor   = editor;
  } // end constructor EditorActionsSet

  // methods
  /** Returns a array of file I/O editor actions. */
  public Action[] getActions() {
    return actions;
  } // end getActions

  /**
   * Should be called on closing a window containing an editor.
   *
   * @return True if window can be closed, false otherwise.
   */
  public boolean canClose() {
    return openAction.conditionalSave();
  } // end canClose

  /** Load a file. */
  public void loadFile(File file) {
    if (file == null) return;
    if (!openAction.conditionalSave()) return;
    try {
      if (!editor.load(file)) return;
      currentFile = file;
      revertAction.setRevertable(true);
    } catch (IOException excep) {
        reportError(excep,"Exception on loading file \"" + file.getName() + "\"");
    }
  } // end loadFile

  /** Save the current editor file. */
  public boolean saveFile() {
    return saveAction.save();
  } // end saveFile

  /** Returns the current editor file. */
  public File getCurrentFile() {
    return currentFile;
  } // end getCurrentFile

  /** Returns the frame title without file name. */
  protected String getTitle() {
   String title = editor.getFrame().getTitle();
   int index;

   if (title != null) {
     index = title.lastIndexOf((int)'-');
     if (index >= 0) title = title.substring(index + 1);
     title = title.trim();
   } // end if
   return title;
  } // end getTitle

  protected void reportError(Exception e, String desc) {
    StringWriter sw = new StringWriter();

    e.printStackTrace(new PrintWriter(sw,true));
    new ErrorDlg(editor.getFrame(),getTitle() + " - Error",
                 true,desc + "\n" + sw);
  } // end reportError

  // inner classes
  /**
   * Create a new project.
   */
  class NewAction extends OpenAction {

    // instance variable declarations
    private boolean alternate;

    // constructors
    NewAction(boolean alternate) {
      super(alternate ? Const.ALTNEW_ACTION : Const.NEW_ACTION);
      this.alternate = alternate;
    } // end constructor NewAction

    // methods
    public void actionPerformed(ActionEvent e) {
      if (!conditionalSave()) return;
      if (!editor.create(alternate)) return;
      currentFile = null;
      revertAction.setRevertable(false);
    } // end actionPerformed

  } // end NewAction

  /**
   * Load a project file.
   */
  class OpenAction extends SaveAction {

    // constructors
    OpenAction() {
      super(Const.OPEN_ACTION);
    } // end constructor OpenAction

    OpenAction(String nm) {
      super(nm);
    } // end constructor OpenAction

    // methods
    public void actionPerformed(ActionEvent e) {
      File file    = getLoadFile();

      if (file == null) return;
      if (!conditionalSave()) return;
      try {
        if (!editor.load(file)) return;
        currentFile = file;
        revertAction.setRevertable(true);
      } catch (IOException excep) {
        reportError(excep,"Exception on loading file \"" + file.getName() + "\"");
      }
    } // end actionPerformed

    /** Saves a file, if it had been changed and the user wants to. */
    protected boolean conditionalSave() {
      if (editor.isChanged()) {
        switch(saveOption()) {
          case JOptionPane.YES_OPTION: return save();
          case JOptionPane.NO_OPTION:  return true;
          default:                     return false; /* CANCEL */
        } // end switch
      } // end if
      return true;
    } // end conditionalSave

    /** Return a file name for load operations or null, if canceled. */
    protected File getLoadFile() {
      String f,d;
      File file = editor.getDefaultFile();

      if (fileDialog == null) {
        fileDialog = new FileDialog(editor.getFrame());
        fileDialog.setFilenameFilter(editor);
      } // end if
      fileDialog.setMode(FileDialog.LOAD);
      if (file != null) {
        fileDialog.setFile(file.getName());
        if (file.isAbsolute()) {
          file = new File(file.getParent());
          fileDialog.setDirectory(file.getAbsolutePath());
        } // end if
      } // end if
      fileDialog.show();
      f = fileDialog.getFile();
      if (f == null) return null;
      d = fileDialog.getDirectory();
      return new File(d,f);
    } // end getLoadFile

    /** Asks the user if he wants to save a file before it is discarted */
    protected int saveOption() {
      ResourceBundle res = editor.getResources();
      String text        = ResourceUtil.getResStringOrKey(res,Const.SAVECONFIRM_TEXT);
      String untitled    = ResourceUtil.getResStringOrKey(res,Const.UNTITLED_NAME);
      File file          = currentFile;

      if (file == null) text = "\"" + untitled + "\" " + text;
        else text = "\"" + file.getPath() + "\" " + text;
      return JOptionPane.showConfirmDialog(editor.getFrame(),text,
               getTitle(),JOptionPane.YES_NO_CANCEL_OPTION,
               JOptionPane.WARNING_MESSAGE);
    } // end saveOption

  } // end OpenAction

  /**
   * Save a project file.  If this is the first time, create a file dialog.
   */
  class SaveAction extends SaveAsAction {

    // constructors
    SaveAction() {
      super(Const.SAVE_ACTION);
    } // end constructor SaveAction

    SaveAction(String nm) {
      super(nm);
    } // end constructor SaveAction

    // instance methods
    /** Return a file name for save operations or null, if canceled. */
    protected File getSaveFile() {
      File file = currentFile;

      if (file == null) return super.getSaveFile();
      return file;
    } // end getSaveFile

  } // end SaveAction

  /**
   * Save a project file under a different name.
   */
  class SaveAsAction extends AbstractAction {

    // constructors
    SaveAsAction() {
      this(Const.SAVEAS_ACTION);
    } // end constructor SaveAsAction

    SaveAsAction(String nm) {
      super(nm);
    } // end constructor SaveAsAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      save();
    } // end actionPerformed

    /** Save the document.  Returns false, if the operation has been canceled. */
    protected boolean save() {
      File f         = getSaveFile();
      boolean result = (f != null);

      if (result) try {
        result = editor.save(f);
        if (result) {
          currentFile = f;
          revertAction.setRevertable(true);
        } // end if
      } catch (IOException excep) {
        reportError(excep,"Exception on saving file \"" + f.getName() + "\"");
        result = false;
      }
      return result;
    } // end save

    /** Return a file name for save operations or null, if canceled. */
    protected File getSaveFile() {
      String f,d;
      File file = editor.getDefaultFile();

      if (fileDialog == null) {
        fileDialog = new FileDialog(editor.getFrame());
        fileDialog.setFilenameFilter(editor);
      } // end if
      fileDialog.setMode(FileDialog.SAVE);
      if (file != null) {
        fileDialog.setFile(file.getName());
        if (file.isAbsolute()) {
          file = new File(file.getParent());
          fileDialog.setDirectory(file.getAbsolutePath());
        } // end if
      } // end if
      fileDialog.show();
      f = fileDialog.getFile();
      if (f == null) return null;
      d = fileDialog.getDirectory();
      if (d == null) return new File(f);
        else return new File(d,f);
    } // end getSaveFile

  } // end SaveAsAction

  /**
   * Revert to last saved or loaded project.
   */
  class RevertAction extends OpenAction {

    // instance variable declarations
    private boolean revertable = false;
    private boolean enabled    = true;

    // constructors
    RevertAction() {
      super(Const.REVERT_ACTION);
      super.setEnabled(enabled && revertable);
    } // end constructor RevertAction

    // methods
    /** Sets if revert is possible or not. */
    public void setRevertable(boolean b) {
      revertable = b;
      super.setEnabled(enabled && revertable);
    } // end setRevertable

    /** Overwritten for revert purposes */
    public synchronized void setEnabled(boolean newValue) {
      enabled = newValue;
      super.setEnabled(enabled && revertable);
    } // end setEnabled

    /** Return a file name for load operations or null, if canceled. */
    protected File getLoadFile() {
      return currentFile;
    } // end getLoadFile

    /** Overwritten for revert purposes */
    protected boolean conditionalSave() {
      if (!editor.isChanged()) return true;
      if (revertOption() == JOptionPane.OK_OPTION) return true;
      return false;
    } // end conditionalSave

    /** Asks the user if he wants to revert a file to its last load/save. */
    protected int revertOption() {
      ResourceBundle res = editor.getResources();
      String text1       = ResourceUtil.getResStringOrKey(res,Const.REVERTCONFIRM_TEXT1);
      String text2       = ResourceUtil.getResStringOrKey(res,Const.REVERTCONFIRM_TEXT2);
      File file          = currentFile;

      text1 += " \"" + file + "\" " + text2;
      return JOptionPane.showConfirmDialog(editor.getFrame(),text1,getTitle(),
               JOptionPane.OK_CANCEL_OPTION,
               JOptionPane.WARNING_MESSAGE);
    } // end revertOption

  } // end RevertAction

  class AutoSaveAction extends SaveAction {

    // instance variable declarations
    private boolean initOk = false;
    private long millis    = System.currentTimeMillis();
    private long distance;
    private boolean enabled;
    private boolean disabled = false;
    private Timer timer;
    private WindowStateTracker wndState;

    // constructors
    AutoSaveAction() {
      super(Const.AUTOSAVE_ACTION);
    wndState = new WindowStateTracker();
    timer    = new Timer(3000,this);  // action event in three seconds
    timer.setRepeats(false);          // only one event
    timer.start();
    } // end constructor AutoSaveAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      if (initOk) {
        if (e.getSource() instanceof Timer) timerEvent(e);
          else settingsEvent(e);
      }
      else {
        initEvent(e);
        initOk = true;
      } // end if
      timer.restart();
    } // end actionPerformed

    /** called if the action event was posted from a timer */
    protected void timerEvent(ActionEvent e) {
      if (!editor.isChanged()) {
        millis   = System.currentTimeMillis();
        disabled = false;
        return;
      } // end if
      if (enabled && !disabled && wndState.isActive() &&
          (System.currentTimeMillis() >= (millis + distance))) {
        switch (autoSaveOption()) {
          case JOptionPane.NO_OPTION:
            millis = System.currentTimeMillis();
          break;
          case JOptionPane.YES_OPTION:
            if (save()) {
              millis = System.currentTimeMillis();
              break;
            } // end if
          default:
            disabled = true;
        } // end switch
      } // end if
    } // end timerEvent

    /** called if the action event was posted from a component */
    protected void settingsEvent(ActionEvent e) {
      ResourceBundle res = editor.getResources();
      AutoSaveDlg autoSaveDlg;
      boolean enabledNow;

      autoSaveDlg = new AutoSaveDlg(editor.getFrame(),(int)(distance / 60000),enabled,
                                    res,Const.AUTOSAVE_OPTIONS);
      autoSaveDlg.setVisible(true);
      if (autoSaveDlg.isCanceled()) return;
      enabled = autoSaveDlg.getEnabled();
      editor.getDesktopProperties().put(Const.AUTOSAVE_ENBL,"" + enabled);
      distance = autoSaveDlg.getMinutes() * 60000;
      editor.getDesktopProperties().put(Const.AUTOSAVE_TIME,"" + distance);
      disabled = false;
    } // end settingsEvent

    /** called on  the first actionPerformed call */
    protected void initEvent(ActionEvent e) {
      String str;

      str = editor.getDesktopProperties().getProperty(Const.AUTOSAVE_TIME);
      editor.getFrame().addWindowListener(wndState);
      if (str == null) distance = 10 * 60000; // ten minutes default
        else distance = Long.valueOf(str).longValue();
      if (distance < 60000) {
        System.err.println("EditorActionsSet.AutoSaveAction: Auto save time" +
                           "out of range");
        distance = 10 * 60000; // ten minutes default
      } // end if
      str     = editor.getDesktopProperties().getProperty(Const.AUTOSAVE_ENBL);
      enabled = (str == null) || Boolean.valueOf(str).booleanValue();
    } // end initEvent

    /** Asks the user if he wants to save a file */
    protected int autoSaveOption() {
      ResourceBundle res = editor.getResources();
      String text1,text2,name;
      File file    = currentFile;
      long minutes = (System.currentTimeMillis() - millis) / 60000;

      text1 = ResourceUtil.getResStringOrKey(res,Const.AUTOSAVE_TEXT1);
      text2 = ResourceUtil.getResStringOrKey(res,Const.AUTOSAVE_TEXT2);
      name  = file == null ?
               ResourceUtil.getResStringOrKey(res,Const.UNTITLED_NAME) :
               file.getPath();
      name  = "\"" + name + "\" ";
      text1 = name + text1 + " " + minutes + " " + text2;
      return JOptionPane.showConfirmDialog(editor.getFrame(),text1,
               getTitle(),JOptionPane.YES_NO_CANCEL_OPTION,
               JOptionPane.QUESTION_MESSAGE);
    } // end saveOption

  } // end AutoSaveAction

} // end EditorActionsSet

/* ******************************************************************
   end of file
*********************************************************************/
