/* ******************************************************************

   class    : InsFileAction
   date     : 03.05.99
   compiler : javac (jdk 1.2)

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.ActionEvent;
import java.io.*;
import java.util.ResourceBundle;
import java.util.Properties;

import javax.swing.*;
import javax.swing.text.JTextComponent;

import Utils.ResourceUtil;
import Utils.FrameUtil;
import Streams.TabReader;
import Dialogs.ErrorDlg;

/**
 * Inserts a file into the editor contained in frame.
 *
 * @author puehringer edgar
 */
class InsFileAction extends AbstractAction {

  // class variable declarations
  private static final String INSFILE_ACTION = "insert-file";
  private static final String TITLE          = "InsFileDlg.title";
  /** Key to the current to insert file resource. */
  private final static String CUR_FILE = "InsFileDlg.currentFile";
  /** Key to the macroprocessing enabled flag resource. */
  private final static String MACRO_PROC_ENBL = "InsFileDlg.macroProcEnbl";

  // instance variable declarations
  protected Frame frame;
  protected ResourceBundle resources;
  protected Properties project;

  // constructors
  InsFileAction(Frame frame, ResourceBundle resources, Properties project) {
    super(INSFILE_ACTION);
    this.frame     = frame;
    this.resources = resources;
    this.project   = project;
  } // end constructor InsFileAction

  // methods
  /**
   * Returns an array containing all extensions of source files.
   */
  protected String[] getExtensions() {
    String str;

    str = project.getProperty(Const.SOURCE_EXT);
    Utils.Assert.assert(Utils.Assert.ENABLED &&
      (str != null),"Extension for source files not found");
    return ResourceUtil.splitToWords(str);
  } // end getExtensions

  /** Returns directory where the files to insert should be. */
  protected String getDirectory() {
    String path = project.getProperty(Const.INSFILE_DIR);

    if (path != null) return ResourceUtil.urlToFile(path);
    return System.getProperty("user.dir");
  } // end getDirectory

  /** Return a file name to load the editor content to insert from. */
  private final File getFile(Frame frame) {
    String str = project.getProperty(CUR_FILE);
    String name,path;
    File file;
    FileDialog dlg;

    if (str == null) {
      name = "*" + getExtensions()[0];
      path = getDirectory();
    }
    else {
      file = new File(ResourceUtil.urlToFile(str));
      name = file.getName();
      path = file.getParent();
    } // end if
    str = ResourceUtil.getResStringOrKey(resources,TITLE);
    dlg = new FileDialog(frame,str,FileDialog.LOAD);
    dlg.setDirectory(path);
    dlg.setFile(name);
    dlg.show();
    name = dlg.getFile();
    if (name != null) {
      path = dlg.getDirectory();
      if (path == null) return new File(name);
        else return new File(path,name);
    } // end if
    return null;
  } // end getFile

  /** Load the document from the spezified file. */
  public String load(File file, AbstractMacroprocessor mp)
    throws IOException {
    Reader in          = new FileReader(file);
    BufferedReader bin = new BufferedReader(in);
    int fmt            = ResourceUtil.getResourceInt(resources,Const.LOAD_FORMAT,
                                                     TabReader.NO_CHANGE);
    Reader tab;
    StringWriter sw;
    int ts = 8;
    Object[] aObj;
    int i;
    String str;
    boolean onlyWhitespaces = true;
    boolean canceled        = false;

    try {
      str = project.getProperty(Const.TAB_SIZE,"8");
      ts  = Integer.parseInt(str);
    } catch (NumberFormatException excep) {
      ts = 8;
    }
    tab = new TabReader(bin,ts,fmt);
    sw  = new StringWriter();
    str = "";
    while ((i = tab.read()) != -1) {
      if (onlyWhitespaces) {
        str += (char)i;
        if (!Character.isWhitespace((char)i)) {
          if ((mp != null) && mp.isPreprocEsc((char)i)) {
            tab = mp.getTemplate(frame,file);
            if (tab == null) {
              canceled = true;
              break;
            } // end if
          }
          else {
            sw.write(str);
          } // end if
          onlyWhitespaces = false;
        } // end if
      }
      else {
        sw.write(i);
      } // end if
    } // end while
    if (tab != null) tab.close();
    bin.close();
    in.close();
    sw.close();
    if (canceled) return null;
    return sw.toString();
  } // end load

  protected void reportError(Exception e, String desc) {
    String title    = ResourceUtil.getResStringOrKey(resources,TITLE);
    StringWriter sw = new StringWriter();

    e.printStackTrace(new PrintWriter(sw,true));
    new ErrorDlg(frame,title + " - Error",
                 true,desc + "\n" + sw);
  } // end reportError

  public void actionPerformed(ActionEvent e) {
    CodepadPanel cpp;
    JTextComponent tk;
    String str;
    File file;
    AbstractMacroprocessor mp;

    cpp = (CodepadPanel)FrameUtil.contains(frame,CodepadPanel.class);
    if (cpp == null) {
      System.err.println("InsFileAction: CodepadPanel not found");
      return;
    } // end if
    if (ResourceUtil.getResourceBool(resources,MACRO_PROC_ENBL)) {
      mp = ProjectPanel.getPreprocessor();
    }
    else {
      mp = null;
    } // end if
    tk   = cpp.getEditor();
    file = getFile(frame);
    if (file == null) return;
    try {
      str = load(file,mp);
      if (str == null) return;
      tk.replaceSelection(str);
      // store the current inserted file name
      str = ResourceUtil.fileToURL(file.getAbsolutePath());
      if (!str.equals(project.getProperty(CUR_FILE))) {
        project.put(CUR_FILE,str);
        project.put(Const.PROJECT_CHANGED,"" + true);
      } // end if
    } catch (IOException excep) {
      reportError(excep,"Exception on loading file \"" + file.getName() + "\"");
    }
  } // end actionPerformed

} // end InsFileAction

/* ******************************************************************
   end of file
*********************************************************************/
