/* ******************************************************************

   class    : JarTool
   date     : Okt 05 1998
   compiler : javac (1.1.6)

******************************************************************* */

/* ******************************************************************
   import statements
******************************************************************* */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

import Utils.ResourceUtil;
import Utils.ZipUtil;
import Utils.StringUtil;
import CustCtrl.BaseDialog;
import CustCtrl.SeparatorCanvas;
import Messages.*;

/**
 * This class exports the project to a jar file.
 *
 * @author puehringer edgar
 */
public class JarTool extends AbstractTool
  implements Observer, Runnable {

  // class variable declarations
  /** Key to the properties file extensions. */
  protected static final String PROPERTY = "propertiesExtension";
  /** Key to the image file extensions. */
  protected static final String IMAGE = "imageExtension";
  /** Key to the html file extensions. */
  protected static final String HTML = "htmlExtension";
  /** Key to the progress bar compress label. */
  protected static final String COMPRESS_LABEL = "compressLabel";

  // Keys to the jar tool dialog
  private final static String JARDLG_TITLE      = "jarToolDlg.title";
  private final static String JARDLG_BUTTONS    = "jarToolDlg.buttons";
  private final static String JARDLG_SELBUTTON  = "jarToolDlg.selButton";
  private final static String JARDLG_CBCOMPRESS = "jarToolDlg.cbCompress";
  private final static String JARDLG_CBRECURSE  = "jarToolDlg.cbRecurse";
  private final static String JARDLG_INCLABEL1  = "jarToolDlg.incLabel1";
  private final static String JARDLG_FCHOICE    = "jarToolDlg.fileChoice";
  private final static String JARDLG_INCLABEL2  = "jarToolDlg.incLabel2";
  private final static String JARDLG_CBPROJECT  = "jarToolDlg.cbProjectFiles";
  private final static String JARDLG_CBSOURCE   = "jarToolDlg.cbSourceFiles";
  private final static String JARDLG_CBEXEC     = "jarToolDlg.cbExecutableFiles";
  private final static String JARDLG_CBPROP     = "jarToolDlg.cbPropertyFiles";
  private final static String JARDLG_CBIMAGE    = "jarToolDlg.cbImageFiles";
  private final static String JARDLG_CBHTML     = "jarToolDlg.cbHtmlFiles";

  // Keys to jar related project settings
  private final static String COMPRESS     = "compress";
  private final static String RECURSE      = "recurseSubDir";
  private final static String INC_ALL      = "includeAll";
  private final static String INC_PROJECT  = "includeProjectFiles";
  private final static String INC_SOURCE   = "includeSourceFiles";
  private final static String INC_EXEC     = "includeExecutableFiles";
  private final static String INC_PROP     = "includePropertyFiles";
  private final static String INC_IMAGE    = "includeImageFiles";
  private final static String INC_HTML     = "includeHTMLFiles";
  // instance variable declarations
  String[] cmds = {
    "export-to-jar"
  };

  /** True if archive should be compressed. */
  protected boolean compress;
  /** True if subdirectories should be recursed. */
  protected boolean recurse;
  /** True if all files should be processed. */
  protected boolean includeAll;
  /** True if the project files should be processed. */
  protected boolean incProjFiles;
  /** True if all source files should be included. */
  protected boolean incSourceFiles;
  /** True if all executable files should be included. */
  protected boolean incExecFiles;
  /** True if all property files should be included. */
  protected boolean incPropFiles;
  /** True if all image files should be included. */
  protected boolean incImgFiles;
  /** True if all html files should be included. */
  protected boolean incHtmlFiles;
  /** Class handling the util.zip package. */
  protected ZipUtil zip;
  /** Number of current compressed file. */
  protected int nrOfCurFile;
  /** The archive file to write to. */
  protected File archiveFile;

  // constructors
  /** Create a new JarTool object. */
  public JarTool(ResourceBundle resources, Properties project,
                 String keyPrefix) {
    super(resources,project,keyPrefix);
  } // end constructor JarTool

  // class methods
  /* DEBUG ON */
  public static void main(String[] args) {
    try {
      ResourceBundle res = ResourceBundle.getBundle("_Main",Locale.getDefault());
      Properties proj = new Properties();
      proj.load(new FileInputStream("Diplom.prj"));
      JarTool jt      = new JarTool(res,proj,"ext1.");
      Frame frame     = new Frame();

      JarToolDlg dlg = jt.new JarToolDlg(frame);
      dlg.show();
      frame.dispose();
    } catch (Exception excep) {
      excep.printStackTrace();
    }
    System.exit(0);
  } // end main
  /* DEBUG OFF */

  // methods
  /** Returns an array containing all action commands of this tool. */
  protected String[] getActionCommands() {
    return cmds;
  } // end getActionCommands

  private String getSetting(String key) {
    return ResourceUtil.getResStringOrKey(resources,
             keyPrefix + key);
  } // end getSetting

  private boolean getBool(String key) {
    String str = getToolProperty(key);

    if (str == null)
      str = ResourceUtil.getResourceString(resources,keyPrefix + key);
    // end if
    return Boolean.valueOf(str).booleanValue();
  } // end getBool

  //---- Runnable --------------------------------------------------
  public void run() {
    FileOutputStream fos     = null;
    BufferedOutputStream bos = null;

    try {
      fos     = new FileOutputStream(archiveFile);
      bos = new BufferedOutputStream(fos);
      zip.dirToZip(bos,new File(getProjectDirectory()),
                   new JarFilter(),recurse,compress);
    } catch (IOException excep) {
      excep.printStackTrace();
    } finally {
      if (bos != null) try { bos.close(); } catch (IOException excep) {}
      if (fos != null) try { fos.close(); } catch (IOException excep) {}
    }
    zip         = null; // allows garbage collection
    archiveFile = null;
    setEnabled(cmds[0],true);
  } // end run

  // --- Observer ---------------------------------------------------
  /**
   * This method is called whenever the observed object is changed.
   * An application calls an observable object's notifyObservers
   * method to have all the object's observers notified of the change.
   */
  public void update(Observable o, Object arg) {
    if (arg instanceof File) {
      String path = ((File)arg).getAbsolutePath();

      path = StringUtil.relativePath(path,getProjectDirectory());
      postMessage(new ProgressMessage(getSetting(COMPRESS_LABEL),
                  new File(path),nrOfCurFile++,zip.nrOfFiles,Message.LOW));
    } else if (arg instanceof String) {
      postMessage(new ErrorMessage(ErrorMessage.ERROR,(String)arg));
    } else { // if arg is null
      postMessage(new ClearProgressMessage(Message.LOW));
    } // end if
  } // end update

  // --- ActionListener ---------------------------------------------
  /** Invoked when an action occurs. */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals(keyPrefix + cmds[0])) {
      compress       = getBool(COMPRESS);
      recurse        = getBool(RECURSE);
      includeAll     = getBool(INC_ALL);
      incProjFiles   = getBool(INC_PROJECT);
      incSourceFiles = getBool(INC_SOURCE);
      incExecFiles   = getBool(INC_EXEC);
      incPropFiles   = getBool(INC_PROP);
      incImgFiles    = getBool(INC_IMAGE);
      incHtmlFiles   = getBool(INC_HTML);

      JarToolDlg dlg = new JarToolDlg(parent);

      dlg.show();
      if (dlg.isCanceled()) return;
      setEnabled(cmds[0],false);
      zip = new ZipUtil();
      zip.addObserver(this);
      nrOfCurFile = 0;
      archiveFile = dlg.getArchiveFile();
      (new Thread(this)).start();
      putToolProperty(COMPRESS,"" + compress);
      putToolProperty(RECURSE,"" + recurse);
      putToolProperty(INC_ALL,"" + includeAll);
      putToolProperty(INC_PROJECT,"" + incProjFiles);
      putToolProperty(INC_SOURCE,"" + incSourceFiles);
      putToolProperty(INC_EXEC,"" + incExecFiles);
      putToolProperty(INC_PROP,"" + incPropFiles);
      putToolProperty(INC_IMAGE,"" + incImgFiles);
      putToolProperty(INC_HTML,"" + incHtmlFiles);
    } // end if
  } // end actionPerformed

  // inner classes
  class JarToolDlg extends BaseDialog
    implements ActionListener, ItemListener {

    // instance variable declarations
    private Frame parent;
    private TextField tfFile;
    private Button btnSelect;
    private Checkbox cbCompress;
    private Checkbox cbRecurse;
    private Choice choice;
    private Checkbox cbProjFiles;
    private Checkbox cbSourceFiles;
    private Checkbox cbExecFiles;
    private Checkbox cbPropFiles;
    private Checkbox cbImgFiles;
    private Checkbox cbHtmlFiles;

    // constructors
    /**
     * Create a dialog that lets you specify archive options.
     *
     * @param parent The parent frame.
     */
    public JarToolDlg(Frame parent) {
      super(parent,true);
      this.parent = parent;

      Panel panel = getContentPanel();
      SeparatorCanvas separator;
      Label label;
      String[] options;
      String str;

      setTitle(getSetting(JARDLG_TITLE));
      createButtons(getSetting(JARDLG_BUTTONS));

      GridBagLayout gridbag          = new GridBagLayout();
      GridBagConstraints constraints = new GridBagConstraints();
      panel.setLayout(gridbag);

      constraints.insets     = new Insets(10,6,10,0);
      constraints.fill       = GridBagConstraints.NONE;
      constraints.anchor     = GridBagConstraints.WEST;
      constraints.weightx    = 1.0;
      constraints.weighty    = 1.0;
      constraints.gridwidth  = GridBagConstraints.RELATIVE;
      constraints.gridheight = 1;
      tfFile = new TextField(15);
      tfFile.setEditable(false);
      gridbag.setConstraints(tfFile,constraints);
      panel.add(tfFile);

      constraints.insets     = new Insets(10,0,10,6);
      constraints.gridwidth  = GridBagConstraints.REMAINDER;
      btnSelect = new Button(getSetting(JARDLG_SELBUTTON));
      btnSelect.addActionListener(this);
      gridbag.setConstraints(btnSelect,constraints);
      panel.add(btnSelect);

      constraints.insets    = new Insets(0,6,10,6);
      constraints.gridwidth = GridBagConstraints.RELATIVE;
      cbCompress = new Checkbox(getSetting(JARDLG_CBCOMPRESS),compress);
      gridbag.setConstraints(cbCompress,constraints);
      panel.add(cbCompress);

      constraints.insets    = new Insets(0,0,10,6);
      constraints.gridwidth = GridBagConstraints.REMAINDER;
      cbRecurse = new Checkbox(getSetting(JARDLG_CBRECURSE),recurse);
      gridbag.setConstraints(cbRecurse,constraints);
      panel.add(cbRecurse);

      constraints.insets = new Insets(0,0,10,0);
      constraints.fill   = GridBagConstraints.HORIZONTAL;
      separator = new SeparatorCanvas();
      gridbag.setConstraints(separator,constraints);
      panel.add(separator);

      constraints.insets    = new Insets(0,6,10,6);
      constraints.fill      = GridBagConstraints.NONE;
      constraints.gridwidth = 1;
      label = new Label(getSetting(JARDLG_INCLABEL1));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets    = new Insets(0,0,10,6);
      constraints.gridwidth = GridBagConstraints.RELATIVE;
      choice  = new Choice();
      options = ResourceUtil.splitToWords(getSetting(JARDLG_FCHOICE));
      for (int i=0; i<options.length; i++) choice.add(options[i]);
      if ((choice.getItemCount() > 1) && includeAll) choice.select(1);
        else choice.select(0);
      choice.addItemListener(this);
      gridbag.setConstraints(choice,constraints);
      panel.add(choice);

      constraints.gridwidth = GridBagConstraints.REMAINDER;
      label = new Label(getSetting(JARDLG_INCLABEL2));
      gridbag.setConstraints(label,constraints);
      panel.add(label);

      constraints.insets = new Insets(0,6,10,6);
      cbProjFiles = new Checkbox(getSetting(JARDLG_CBPROJECT),incProjFiles);
      gridbag.setConstraints(cbProjFiles,constraints);
      panel.add(cbProjFiles);

      options = getExtensions(SOURCE);
      str     = " (" + options[0];
      for (int i=1; i<(options.length / 2); i++) str += (", " + options[i]);
      str += ")";
      cbSourceFiles = new Checkbox(getSetting(JARDLG_CBSOURCE) + str,
                                   incSourceFiles);
      gridbag.setConstraints(cbSourceFiles,constraints);
      panel.add(cbSourceFiles);

      options = getExtensions(EXECUTABLE);
      str     = " (" + options[0];
      for (int i=1; i<(options.length / 2); i++) str += (", " + options[i]);
      str += ")";
      cbExecFiles = new Checkbox(getSetting(JARDLG_CBEXEC) + str,
                                 incExecFiles);
      gridbag.setConstraints(cbExecFiles,constraints);
      panel.add(cbExecFiles);

      options = ResourceUtil.splitToWords(getSetting(PROPERTY));
      str     = " (" + options[0];
      for (int i=1; i< options.length; i++) str += (", " + options[i]);
      str += ")";
      cbPropFiles = new Checkbox(getSetting(JARDLG_CBPROP) + str,
                                 incPropFiles);
      gridbag.setConstraints(cbPropFiles,constraints);
      panel.add(cbPropFiles);

      constraints.gridheight = GridBagConstraints.RELATIVE;
      options = ResourceUtil.splitToWords(getSetting(IMAGE));
      str     = " (" + options[0];
      for (int i=1; i< options.length; i++) str += (", " + options[i]);
      str += ")";
      cbImgFiles = new Checkbox(getSetting(JARDLG_CBIMAGE) + str,
                                  incImgFiles);
      gridbag.setConstraints(cbImgFiles,constraints);
      panel.add(cbImgFiles);

      constraints.gridheight = GridBagConstraints.REMAINDER;
      options = ResourceUtil.splitToWords(getSetting(HTML));
      str     = " (" + options[0];
      for (int i=1; i< options.length; i++) str += (", " + options[i]);
      str += ")";
      cbHtmlFiles = new Checkbox(getSetting(JARDLG_CBHTML) + str,
                                   incHtmlFiles);
      gridbag.setConstraints(cbHtmlFiles,constraints);
      panel.add(cbHtmlFiles);
      // Set the enabled state of some components
      itemStateChanged(null);

      pack();
      center(parent);
      setResizable(false);
      setButtonEnabled(0,false);
    } // end constructor JarToolDlg

    // methods
    public File getArchiveFile() {
      return new File(tfFile.getText());
    } // end getArchiveFile

    /** Called to close this dialog. */
    protected void closeDialog(AWTEvent event) {
      super.closeDialog(event);
      if (isCanceled()) return;
      compress       = cbCompress.getState();
      recurse        = cbRecurse.getState();
      includeAll     = (choice.getSelectedIndex() > 0);
      incProjFiles   = cbProjFiles.getState();
      incSourceFiles = cbSourceFiles.getState();
      incExecFiles   = cbExecFiles.getState();
      incPropFiles   = cbPropFiles.getState();
      incImgFiles    = cbImgFiles.getState();
      incHtmlFiles   = cbHtmlFiles.getState();
    } // end closeDialog

    // --- ActionListener ---------------------------------------------
    /** Invoked when an action occurs. */
    public void actionPerformed(ActionEvent e) {
      File file = tfFile.getText().equals("") ?
                  new File(getProjectDirectory()) :
                  new File(tfFile.getText());
      String name,path;
      FileDialog dlg;
      File result = null;

      if (file.isDirectory()) {
        name = "*.jar";
        path = file.getAbsolutePath();
      }
      else {
        name = file.getName();
        path = file.getParent();
      } // end if
      dlg = new FileDialog(parent,getSetting(JARDLG_TITLE),FileDialog.SAVE);
      dlg.setDirectory(path);
      dlg.setFile(name);
      dlg.show();
      name = dlg.getFile();
      if (name != null) {
        path = dlg.getDirectory();
        if (path == null) result = new File(name);
          else result = new File(path,name);
      } // end if
      if (result == null) return;
      path = result.getAbsolutePath();
      tfFile.setText(path);
      setButtonEnabled(0,result.canWrite() || !result.exists());
    } // end actionPerformed

    // --- ItemListener -----------------------------------------------
    /** Invoked when an item's state has been changed. */
    public void itemStateChanged(ItemEvent e) {
      boolean select = (choice.getSelectedIndex() == 0);

      cbProjFiles.setEnabled(select);
      cbSourceFiles.setEnabled(select);
      cbExecFiles.setEnabled(select);
      cbPropFiles.setEnabled(select);
      cbImgFiles.setEnabled(select);
      cbHtmlFiles.setEnabled(select);
    } // end itemStateChanged

  } // end JarToolDlg

  class JarFilter implements FilenameFilter {

    // instance variable declarations
    boolean acceptAll;
    String[] extensions;
    String[] projFiles;

    // constructors
    JarFilter() {
      acceptAll = includeAll;
      if (acceptAll) return;

      Hashtable ht = new Hashtable();
      String[] aStr;
      int i;
      Enumeration enum;

      if (incSourceFiles) {
        aStr = getExtensions(SOURCE);
        for (i=0; i<aStr.length; i++) ht.put(aStr[i],new Object());
      } // end if
      if (incExecFiles) {
        aStr = getExtensions(EXECUTABLE);
        for (i=0; i<aStr.length; i++) ht.put(aStr[i],new Object());
      } // end if
      if (incPropFiles) {
        aStr = ResourceUtil.splitToWords(getSetting(PROPERTY));
        for (i=0; i<aStr.length; i++) ht.put(aStr[i],new Object());
      } // end if
      if (incImgFiles) {
        aStr = ResourceUtil.splitToWords(getSetting(IMAGE));
        for (i=0; i<aStr.length; i++) ht.put(aStr[i],new Object());
      } // end if
      if (incHtmlFiles) {
        aStr = ResourceUtil.splitToWords(getSetting(HTML));
        for (i=0; i<aStr.length; i++) ht.put(aStr[i],new Object());
      } // end if
      extensions = new String[ht.size()];
      enum       = ht.keys();
      i          = 0;
      while (enum.hasMoreElements())
        extensions[i++] = enum.nextElement().toString();
      // end while
      if (incProjFiles) {
        File[] files    = getProjectFiles();
        String projPath = getProjectProperty(Const.PROJECTFILE_PATH);

        if (projPath == null) {
          projFiles = new String[files.length];
        }
        else {
          projFiles = new String[files.length + 1];
          projPath  = ResourceUtil.urlToFile(projPath);
          projFiles[files.length] = projPath;
        } // end if
        for (i=0; i<files.length; i++)
          projFiles[i] = files[i].getAbsolutePath();
        // end for
      }
      else {
        projFiles = new String[0];
      } // end if
    } // end constructor JarFilter

    // methods
    // --- FilenameFilter ---------------------------------------------
    /** Tests if a specified file should be included in a file list. */
    public boolean accept(File dir, String name) {
      String path = (new File(dir,name)).getAbsolutePath();

      if (path.equals(archiveFile.getAbsolutePath())) return false;
      if (acceptAll) return true;
      for (int i=0; i<extensions.length; i++)
        if (name.endsWith(extensions[i])) return true;
      // end for
      for (int i=0; i<projFiles.length; i++)
        if (path.equals(projFiles[i])) return true;
      // end for
      return false;
    } // end accept

  } // end JarFilter

} // end JarTool

/* ******************************************************************
   end of file
******************************************************************* */
