/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.6)

   class      : Main
   date       : 13.07.98
   description: this class contains the main methode for the project.
   modified   : 18.12.98 - swing installation check added.

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.*;
import java.net.URL;
import java.util.*;

import javax.swing.*;

import Utils.ResourceUtil;
import Utils.FrameUtil;
import CustCtrl.ImageCanvas;
import Dialogs.ErrorDlg;

/* ******************************************************************
   top level class declaration
*********************************************************************/

/**
 * Creates one instance of the ProjectPanel.
 *
 * @author  Puehringer Edgar
 * @version 1.0 13/07/98
 */
class Main {

  // class variable declarations
  /** Name of the default resource bundle */
  private static final String BUNDLE_NAME = "_Main";
  /** The key to the look and feel specification */
  private static final String LOOK_AND_FEEL = "useSystemLookAndFeel";
  /** The key to the frame title resource */
  private static final String FRAME_TITLE = "pTitle";
  /** The key to the frame icon resource */
  private static final String FRAME_ICON = "pIconImage";
  /** The key to the desktop default settings */
  private static final String DEF_SETTINGS = "desktopDefSettings";
  /** The title picture relative url. */
  private static final String TITLE_IMAGE = "images/title.gif";

  private static ResourceBundle resources;

  // class methods
  public static void main(String[] args) {
    try {
      // manipulate the java.home system property to be a canonical path
      try {
        String s = System.getProperty("java.home");

        s = (new File(s)).getCanonicalPath();
        System.getProperties().put("java.home",s);
      } catch (IOException e) {}
      // set the security manager
      SecurityManager sm = new SwitchableSecurityManager();
      System.setSecurityManager(sm);
      // load the resource bundle
      String bundleName    = args.length > 0 ? args[0] : BUNDLE_NAME;
      Frame titleFrame     = new Frame();
      Panel p              = new Panel(new GridLayout(1,1));
      URL url              = sm.getClass().getResource(TITLE_IMAGE);
      Image img            = Toolkit.getDefaultToolkit().getImage(url);
      long timeStamp;

      p.setBackground(SystemColor.control);
      p.add(new ImageCanvas(img));
      titleFrame.add(p);
      titleFrame.pack();
      titleFrame.setResizable(false);
      FrameUtil.center(titleFrame,null);
      titleFrame.setVisible(true);
      timeStamp = System.currentTimeMillis();
      try {
        resources = ResourceBundle.getBundle(bundleName,Locale.getDefault());
      } catch (MissingResourceException mre) {
        System.err.println("Main: " + bundleName + ".properties not found");
        titleFrame.setVisible(false);
        titleFrame.dispose();
        System.exit(1);
       }
      // Load the desktop settings
      Properties desktop = new Properties();
      String name = ResourceUtil.getResourceString(resources,Const.DESKTOP_NAME);
      File file;
      String key    = ResourceUtil.getResourceString(resources,DEF_SETTINGS);
      String[] keys = key == null ? new String[0] : ResourceUtil.splitToWords(key);

      for (int i=0; i<keys.length; i++) {
        String str = ResourceUtil.getResourceString(resources,keys[i]);
        if (str == null)
          System.err.println("Main: missing resource: " + keys[i]);
        else
          desktop.put(keys[i],str);
        // end if
      } // end for
      try {
        file = new File(name);
      } catch (NullPointerException npe) {
        System.err.println("Main: " + bundleName + ".properties: " +
                           Const.DESKTOP_NAME + " not found");
        titleFrame.setVisible(false);
        titleFrame.dispose();
        System.exit(2);
        file = null; // to avoid not initialized error
      }
      if (file.exists()) {
        FileInputStream fis    = new FileInputStream(file);
        BufferedInputStream in = new BufferedInputStream(fis);

        desktop.load(in);
        in.close();
        fis.close();
      } // end if
      // Force the application to come up in the specified L&F
      String str    = ResourceUtil.getResourceString(resources,LOOK_AND_FEEL);
      boolean sysLF = (str != null) && Boolean.valueOf(str).booleanValue();

      try {
        if (sysLF)
          UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        else
          UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
        // end if
      } catch (Exception excep) {
        System.err.println("Main: Error loading L&F: " + excep);
      } catch (NoClassDefFoundError error) {
      }
      // create a frame for the project window
      FrameManager fm  = new FrameManager();
      String title     = ResourceUtil.getResStringOrKey(resources,FRAME_TITLE);
      Frame f          = null;

      try {
        f = new JFrame(title);
        if (false) throw new ClassNotFoundException();
      } catch (NoClassDefFoundError error) {
        StringWriter sw = new StringWriter();

        error.printStackTrace(new PrintWriter(sw,true));
        new ErrorDlg(titleFrame,"Error",
                     true,"Swing class(es) not found.\n" +
                     "Check your swing installation and try again.\n\n" + sw);
        titleFrame.setVisible(false);
        titleFrame.dispose();
        System.exit(3);
      }

      JFrame frame     = (JFrame)f;
      Container cp     = frame.getContentPane();
      Action[] actions = new Action[1];

      actions[0] = new ExitAction(fm);
      // create a project panel
      ProjectPanel project = new ProjectPanel(frame,resources,desktop,fm,
                                              actions);

      frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
      cp.setLayout(new BorderLayout());
      cp.add("Center",project);
      frame.pack();
      frame.setSize(600,400);
      FrameUtil.shuffleLocation(frame);
      try {
        url = ResourceUtil.getResource(project,resources,FRAME_ICON);
        if (url != null) frame.setIconImage((new ImageIcon(url)).getImage());
      } catch (MissingResourceException mre) {
        // just don't use an icon
      }
      fm.register(frame,project);
      while (System.currentTimeMillis() < timeStamp + 3000) try {
        Thread.sleep(100);
      } catch (InterruptedException excep) {}
      frame.show();
      titleFrame.setVisible(false);
      titleFrame.dispose();
    } catch (Throwable t) {
      System.err.println("Main: Uncaught exception: "+t);
      t.printStackTrace();
    }
  } // end main

} // end Main

/* ******************************************************************
   end of file
*********************************************************************/
