/* ******************************************************************

   class      : MenuDocExtract
   date       : Aug 13 1999

******************************************************************* */

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import javax.swing.event.*;
import javax.swing.*;

/**
 * Builds up a menu by reading the settings from a resource file.
 *
 * @author  Puehringer Edgar
 */
class MenuDocExtract {

  // class variable declarations
  /**
   * Suffix applied to the key used in resource file
   * lookups for an image.
   */
  public static final String IMAGE_SUFFIX = "Image";

  /**
   * Suffix applied to the key used in resource file
   * lookups for a tooltip.
   */
  public static final String TOOLTIP_SUFFIX = "Tooltip";

  /**
   * Suffix applied to the key used in resource file
   * lookups for a label.
   */
  public static final String LABEL_SUFFIX = "Label";

  /**
   * Suffix applied to the key used in resource file
   * lookups for a accelerator.
   */
  public static final String ACCEL_SUFFIX = "Accel";

  /**
   * Suffix applied to the key used in resource file
   * lookups for an action.
   */
  public static final String ACTION_SUFFIX = "Action";

  /**
   * Suffix applied to the key used in resource file
   * lookups for the type of a menu item.
   */
  public static final String TYPE_SUFFIX = "Type";

  /** Type identifier of a menu item. */
  public static final String TYPE_ITEM = "menuitem";

  /** Type identifier of a submenu. */
  public static final String TYPE_MENU = "menu";

  /** Type identifier of a checkbox menu item. */
  public static final String TYPE_CHECKBOX = "checkbox";

  /** The escape character to identify a mnemonic in a label. */
  public static final char MNEMONIC_ESCAPE = '&';

 /** The char that forms the mnemonic in the label string. */
  private static Class eventClass;
  private static Class keyEventClass;
  private static File sourceDir;
  private static File targetDir;

  // instance variable declarations
  private ResourceBundle resources;
  private Hashtable commands;
  private ActionListener defaultAction;
  private Vector backLinkTitle;
  private Vector backLinkURL;

  // static initializers
  static {
    try {
      keyEventClass = Class.forName("java.awt.event.KeyEvent");
    } catch (ClassNotFoundException e) { e.printStackTrace(); }
    try {
      eventClass = Class.forName("java.awt.Event");
    } catch (ClassNotFoundException e) { e.printStackTrace(); }
  } // end static initializer

  // constructors
  MenuDocExtract(ResourceBundle resources) {
    this.resources = resources;
    commands       = null;
    defaultAction  = null;
    backLinkTitle  = new Vector();
    backLinkURL    = new Vector();
  } // end constructor MenuDocExtract

  // class methods
  public static void main(String[] args) {
    if (args.length != 4) {
      System.err.println("Illegal number of parameters");
      usage();
    } // end if

    ResourceBundle res = null;
    MenuDocExtract md;

    try {
      res = ResourceBundle.getBundle(args[0],Locale.getDefault());
    } catch (MissingResourceException mre) {
      System.err.println("Resource bundle not found");
      usage();
    }
    sourceDir = new File(args[2]);
    if (!sourceDir.isDirectory()) {
      System.err.println("Source directory parameter doesn't refer a directory");
      usage();
    } // end if
    targetDir = new File(args[3]);
    if (!targetDir.isDirectory()) {
      System.err.println("Target directory parameter doesn't refer a directory");
      usage();
    } // end if
    md = new MenuDocExtract(res);
    md.createMenuBar(args[1]);
    System.exit(0);
  } // end main

  /**
   * Returns an input stream for a html file that describes the
   * given menu item.
   */
  protected static InputStream getInputStreamForKey(String key) {
    File file = new File(sourceDir,key + ".html");

    try {
      return new FileInputStream(file);
    } catch (IOException excep) {
      excep.printStackTrace();
      System.exit(2);
    }
    return null;
  } // end getInputStreamForKey

  /**
   * Returns an output stream for the texct file to write that
   * describes the given menu item.
   */
  protected static OutputStream getOutputStreamForKey(String key) {
    File file = new File(targetDir,key + ".txt");

    try {
      return new FileOutputStream(file);
    } catch (IOException excep) {
      excep.printStackTrace();
      System.exit(3);
    }
    return null;
  } // end getOutputStreamForKey

  private static void usage() {
    System.err.println("Usage: java MenuDocExtract <resourcebundle> <key> <source directory> <target directory>");
    System.exit(1);
  } // end usage

  /** Filters the mnemonic identifiers out of a label string. */
  private static String filterMnemonic(String label, char[] ch) {
    int index = label.indexOf(MNEMONIC_ESCAPE);

    if (index < 0) return label;
    if ((index + 1) < label.length()) {
      char c = label.charAt(index + 1);

      if (c != MNEMONIC_ESCAPE) ch[0] = c;
      return label.substring(0,index) + c +
        filterMnemonic(label.substring(index + 2),ch);
    }
    else {
      return label.substring(0,index);
    } // end if
  } // end filterMnemonic

  // methods
  /** Return a hashtable containing the menu commands. */
  public Hashtable getCommands() {
    return commands;
  } // end getCommands

  /**
   * Set the hashtable containing the menu commands.
   * If you use it, this must be called before createMenuBar.
   *
   * @see #createMenuBar
   */
  public void setCommands(Hashtable commands) {
    this.commands = commands;
  } // end setCommands

  /**
   * Return the default action listener.
   */
  public ActionListener getDefaultActionListener() {
    return defaultAction;
  } // end getDefaultActionListener

  /**
   * Set the default action listener.
   * If you use it, this must be called before createMenuBar.
   *
   * @see #createMenuBar
   */
  public void setDefaultActionListener(ActionListener al) {
    this.defaultAction = al;
  } // end setDefaultActionListener

  /** This is the hook through which all menu items are created. */
  public String createMenuItem(String prefix, String key) {
    String url   = getResourceString(key + IMAGE_SUFFIX);
    String label = getResourceString(key + LABEL_SUFFIX);
    String type  = getResourceString(key + TYPE_SUFFIX);
    String astr  = getResourceString(key + ACTION_SUFFIX);
    String acc   = getResourceString(key + ACCEL_SUFFIX);
    String tip   = getResourceString(key + TOOLTIP_SUFFIX);
    char[] mnemonic   = new char[1];
    OutputStream os   = getOutputStreamForKey(prefix + key);
    PrintWriter pw    = new PrintWriter(os);
    InputStream is    = getInputStreamForKey(prefix + key);
    BufferedReader br = new BufferedReader(new InputStreamReader(is));

    // prepare the menu item label
    if (label == null) {
      reportError("Label for menu item key \"" + key + "\" not found");
      return null;
    }
    else {
      label = filterMnemonic(label,mnemonic);
    } // end if
    // create a menu item matching the spezified type
    if (type != null) type = type.toLowerCase();
    if (astr == null) astr = key;
    if ((type == null) || type.equals(TYPE_ITEM)) {
    } else if (type.equals(TYPE_CHECKBOX)) {
    } else if (type.equals(TYPE_MENU)) {
      return createMenu(prefix,key);
    } else {
      reportError("Unknown menu type: " + type);
      return null;
    } // end if
    String line;

    try {
      line = br.readLine();
      while ((line != null) && (line.indexOf("</H3>") < 0)) line = br.readLine();
      if (line != null) {                // found!
        line = br.readLine();
        if ((line != null) && line.trim().equals("")) line = br.readLine();
        if ((line != null) && (line.indexOf("<H4>") >= 0)) {
          line = br.readLine();
          if ((line != null) && (line.indexOf("</H4>") >= 0)) line = br.readLine();
        } // end if
        while ((line != null) && (line.indexOf("</BODY>") < 0)) {
          pw.println(line);
          line = br.readLine();
        } // end while
      } // end if
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
    pw.close();
    try { os.close(); } catch (IOException e) {}
    try { br.close(); } catch (IOException e) {}
    try { is.close(); } catch (IOException e) {}
    return label;
  } // end createMenuItem

  /**
   * Return the Action stored in the command hashtable.  If it doesn't
   * exist, return the default action.
   */
  protected ActionListener getAction(String cmd) {
    ActionListener a;
    if (commands == null) a = null;
      else a = (ActionListener)commands.get(cmd);
    if (a == null) return defaultAction;
      else return a;
  } // end getAction

  /**
   * Read a string from the resource file containing the menu
   * definition.
   */
  protected String getResourceString(String key) {
    String str;

    try {
      str = resources.getString(key);
    } catch (MissingResourceException mre) {
      str = null;
    }
    return str;
  } // end getResourceString

  protected URL getResource(String key) {
    String name = getResourceString(key);
    URL url;

    if(name == null) return null;
    url = this.getClass().getResource(name);
    return url;
  } // end getResource

  /**
   * Take the given string and chop it up into a series
   * of strings on whitespace boundries.  This is useful
   * for trying to get an array of strings out of the
   * resource file.
   */
  protected String[] tokenize(String input) {
    Vector v          = new Vector();
    StringTokenizer t = new StringTokenizer(input);
    String cmd[];

    while (t.hasMoreTokens()) v.addElement(t.nextToken());
    cmd = new String[v.size()];
    for (int i=0; i<cmd.length; i++) cmd[i] = (String)v.elementAt(i);
    return cmd;
  } // end tokenize

  /**
   * Create the menubar for the app. By default this pulls the
   * definition of the menu from the associated resource file.
   */
  protected void createMenuBar(String key) {
    String mbDesc = getResourceString(key);
    String[] menuKeys;

    if (mbDesc == null) {
      reportError("MenuBar \"" + key + "\" not found or empty");
      return;
    } // end if
    backLinkTitle.addElement("Index");
    backLinkURL.addElement("index.html");
    menuKeys = tokenize(mbDesc);
    for (int i=0; i<menuKeys.length; i++) {
      String m = createMenu(key,menuKeys[i]);

      System.out.println(m);
    } // end for
    backLinkURL.removeElementAt(backLinkURL.size() - 1);
    backLinkTitle.removeElementAt(backLinkTitle.size() - 1);
  } // end createMenuBar

  /**
   * Create a menu.  By default this pulls the definition
   * of the menu from the associated resource file.
   */
  protected String createMenu(String prefix, String key) {
    String label    = getResourceString(key + LABEL_SUFFIX);
    String menuDesc = getResourceString(key);
    String[] itemKeys;
    char[] mnemonic  = new char[1];

    if (menuDesc == null) {
      reportError("Menu \"" + key + "\" not found or empty");
      return null;
    } // end if
    itemKeys = tokenize(menuDesc);
    // prepare the menu label
    if (label == null) {
      reportError("Label for menu key \"" + key + "\" not found");
      return null;
    } // end if
    label = filterMnemonic(label,mnemonic);
    if (menuDesc == null) {
      reportError("Menu \"" + key + "\" not found or empty");
      return null;
    } // end if
    backLinkTitle.addElement(label + " Menu");
    backLinkURL.addElement(prefix + key + ".html");
    for (int i = 0; i < itemKeys.length; i++) {
      if (itemKeys[i].equals("-")) {
        // skip separators
      }
      else {
        String mi = createMenuItem(prefix,itemKeys[i]);

      } // end if
    } // end for
    backLinkURL.removeElementAt(backLinkURL.size() - 1);
    backLinkTitle.removeElementAt(backLinkTitle.size() - 1);
    return label;
  } // end createMenu

  /** Hook through which ImageIcons are loaded. */
  protected ImageIcon loadImageIcon(URL url) {
    return new ImageIcon(url);
  } // end loadImageIcon

  protected void reportError(String text) {
    System.err.println("MenuDocExtract: " + text + ".");
  } // end reportError

  // inner classes
  private class ActionChangedListener implements PropertyChangeListener {

    // class variable declarations
    JMenuItem menuItem;

    // constructors
    ActionChangedListener(JMenuItem mi) {
      super();
      menuItem = mi;
    } // end constructor ActionChangedListener

    // methods
    public void propertyChange(PropertyChangeEvent e) {
      String propertyName = e.getPropertyName();

      if (e.getPropertyName().equals(Action.NAME)) {
        String text = (String)e.getNewValue();

        menuItem.setText(text);
      } else if (propertyName.equals("enabled")) {
        Boolean enabledState = (Boolean)e.getNewValue();

        menuItem.setEnabled(enabledState.booleanValue());
      } else if (propertyName.equals("checked") &&
                 (menuItem instanceof JCheckBoxMenuItem)) {
        boolean checkedState   = ((Boolean)e.getNewValue()).booleanValue();
        JCheckBoxMenuItem cbmi = (JCheckBoxMenuItem)menuItem;

        if (cbmi.getState() != checkedState)
          cbmi.setState(checkedState);
        // end if
      } // end if
    } // end propertyCange

  } // end ActionChangedListener

} // end MenuDocExtract

/* ******************************************************************
   end of file
*********************************************************************/
