/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.6)

   class      : MenuLoader
   date       : 27.09.98
   description: A class that builds up a menu by reading the
                settings from a resource file.

*********************************************************************/

package MenuPreview;

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.io.InputStream;
import java.io.IOException;
import java.util.*;

/* ******************************************************************
   top level class declaration
*********************************************************************/

/**
 * A class that builds up a menu by reading the settings from a
 * resource file.
 *
 * @author  Puehringer Edgar
 */
class MenuLoader {

  // class variable declarations
 /** The char that forms the mnemonic in the label string. */
  private static char mnemonicChar;
  private static Class eventClass;
  private static Class keyEventClass;

  // instance variable declarations
  private Properties resources;
  private ActionListener defaultAction;

  // static initializers
  static {
    try {
      keyEventClass = Class.forName("java.awt.event.KeyEvent");
    } catch (ClassNotFoundException e) { e.printStackTrace(); }
    try {
      eventClass = Class.forName("java.awt.Event");
    } catch (ClassNotFoundException e) { e.printStackTrace(); }
  } // end static initializer

  // constructors
  MenuLoader(InputStream is) throws IOException {
    resources = new Properties();
    resources.load(is);
    defaultAction = null;
  } // end constructor MenuLoader

  /** Create a new MenuLoader object. */
  MenuLoader(Properties resources) {
    this.resources = resources;
    defaultAction  = null;
  } // end constructor MenuLoader

  // class methods
  /**
   * Filters the mnemonic identifiers out of a label string.
   * Writes the mnemonic position into mnemonicPos.
   */
  private static String filterMnemonic(String label) {
    int l              = label.length();
    StringBuffer sb    = new StringBuffer(l);
    boolean lastWasEsc = false;
    char ch;

    mnemonicChar = (char)0;
    for (int i=0; i<l; i++) {
      ch = label.charAt(i);
      if (ch == Const.MNEMONIC_ESCAPE) {
        if (lastWasEsc) sb.append(ch);
        lastWasEsc = !lastWasEsc;
      }
      else {
        sb.append(ch);
        if (lastWasEsc) mnemonicChar = Character.toLowerCase(ch);
        lastWasEsc = false;
      } // end if
    } // end for
    return sb.toString();
  } // end filterMnemonic

  // methods
  /** Return the default action listener. */
  public ActionListener getDefaultActionListener() {
    return defaultAction;
  } // end getDefaultActionListener

  /**
   * Set the default action listener.
   * If you use it, this must be called before createMenuBar.
   *
   * @see #createMenuBar
   */
  public void setDefaultActionListener(ActionListener al) {
    this.defaultAction = al;
  } // end setDefaultActionListener

  /** This is the hook through which all menu items are created. */
  public MenuItem createMenuItem(String key) {
    String label = getResourceString(key + Const.LABEL_SUFFIX);
    String type  = getResourceString(key + Const.TYPE_SUFFIX);
    String astr  = getResourceString(key + Const.ACTION_SUFFIX);
    String acc   = getResourceString(key + Const.ACCEL_SUFFIX);
    MenuItem mi;
    ActionListener a;
    boolean isSubmenu;
    ActionAdapter adapter;

    // prepare the menu item label
    if (label == null) {
      reportError("Label for menu item key \"" + key + "\" not found");
      return null;
    }
    else {
      label = filterMnemonic(label);
    } // end if
    // create a menu item matching the spezified type
    if (type != null) type = type.toLowerCase();
    if (astr == null) astr = key;
    if ((type == null) || type.equals(Const.TYPE_ITEM)) {
      mi = new MenuItem(label);
      mi.setActionCommand(astr);
      a = getAction(astr);
      isSubmenu = false;
      adapter   = null;
    } else if (type.equals(Const.TYPE_CHECKBOX)) {
      CheckboxMenuItem cbmi = new CheckboxMenuItem(label);

      cbmi.setActionCommand(astr);
      a = getAction(astr);
      isSubmenu = false;
      adapter   = new ActionAdapter();
      cbmi.addItemListener(adapter);
      mi = cbmi;
    } else if (type.equals(Const.TYPE_MENU)) {
      mi        = createMenu(key);
      a         = null;
      isSubmenu = true;
      adapter   = null;
    } else {
      reportError("Unknown menu type: " + type);
      return null;
    } // end if
    // connect an action to the menu command
    if (a != null) {
      if (adapter == null) mi.addActionListener(a);
        else adapter.addActionListener(a);
    } else if (!isSubmenu) {
      mi.setEnabled(false);
    } // end if
    // set the accelerator
    if (acc != null) {
      try {
        String[] arr = tokenize(acc);
        int value    = keyEventClass.getField(arr[0]).getInt(null);
        int mod      = Event.CTRL_MASK;

        for (int i=1; i<arr.length; i++) {
          mod ^= eventClass.getField(arr[i]).getInt(null);
        } // end for

        mi.setShortcut(new MenuShortcut(value,
          ((mod & Event.SHIFT_MASK) != 0)));
      } catch (NoSuchFieldException e1) {
        reportError("Unknown keycode: " + acc);
      } catch (SecurityException e2) {
        e2.printStackTrace();
      } catch (IllegalAccessException e3) {
        e3.printStackTrace();
      }
    } // end if
    return mi;
  } // end createMenuItem

  /** Return the default action. */
  protected ActionListener getAction(String cmd) {
    return defaultAction;
  } // end getAction

  /**
   * Read a string from the resource properties containing the menu
   * definition.
   */
  protected String getResourceString(String key) {
    return resources.getProperty(key);
  } // end getResourceString

  /**
   * Take the given string and chop it up into a series
   * of strings on whitespace boundries.  This is useful
   * for trying to get an array of strings out of the
   * resource file.
   */
  protected String[] tokenize(String input) {
    Vector v          = new Vector();
    StringTokenizer t = new StringTokenizer(input);
    String cmd[];

    while (t.hasMoreTokens()) v.addElement(t.nextToken());
    cmd = new String[v.size()];
    for (int i=0; i<cmd.length; i++) cmd[i] = (String)v.elementAt(i);
    return cmd;
  } // end tokenize

  /**
   * Create the menubar for the app.  By default this pulls the
   * definition of the menu from the associated resource file.
   */
  protected MenuBar createMenuBar(String key) {
    MenuBar mb   = new MenuBar();
    String mbDesc = getResourceString(key);
    String[] menuKeys;

    if (mbDesc == null) {
      reportError("MenuBar \"" + key + "\" not found or empty");
      return null;
    } // end if
    menuKeys = tokenize(mbDesc);
    for (int i=0; i<menuKeys.length; i++) {
      Menu m = createMenu(menuKeys[i]);

      if (m != null) mb.add(m);
    } // end for
    return mb;
  } // end createMenuBar

  /**
   * Create a menu.  By default this pulls the definition
   * of the menu from the associated resource file.
   */
  protected Menu createMenu(String key) {
    String label    = getResourceString(key + Const.LABEL_SUFFIX);
    String menuDesc = getResourceString(key);
    String[] itemKeys;
    Menu menu;

    if (menuDesc == null) {
      reportError("Menu \"" + key + "\" not found or empty");
      return null;
    } // end if
    itemKeys = tokenize(menuDesc);
    // prepare the menu label
    if (label == null) {
      reportError("Label for menu key \"" + key + "\" not found");
      return null;
    } // end if
    label = filterMnemonic(label);
    if (menuDesc == null) {
      reportError("Menu \"" + key + "\" not found or empty");
      return null;
    } // end if
    menu = new Menu(label);
    for (int i = 0; i < itemKeys.length; i++) {
      if (itemKeys[i].equals("-")) {
        menu.addSeparator();
      }
      else {
        MenuItem mi = createMenuItem(itemKeys[i]);

        if (mi != null) menu.add(mi);
      } // end if
    } // end for
    return menu;
  } // end createMenu

  protected void reportError(String text) {
    System.err.println("MenuLoader: " + text + ".");
  } // end reportError

  // inner classes
  /**
   * Converts an item event to an action event.  This is useful
   * when dealing around with CheckboxMenuItems.
   */
  class ActionAdapter implements ItemListener {

    // instance variable declarations
    Hashtable listeners;

    // constructors
    ActionAdapter() {
      listeners = new Hashtable(1);
    } // end constructor ActionAdapter

    // methods
    /**
     * Adds the specified action listener to receive action events.
     *
     * @param l The action listener.
     * @see removeActionListener
     */
    public synchronized void addActionListener(ActionListener l) {
      listeners.put(l,new Object());
    } // end addActionListener

    /**
     * Removes the specified action listener so that it no longer
     * receives action events.
     *
     * @param l The action listener.
     * @see addActionListener
     */
    public synchronized void removeActionListener(ActionListener l) {
      listeners.remove(l);
    } // end removeActionListener

    // --- ItemListener -----------------------------------------------
    /** Invoked when an item's state has been changed. */
    public void itemStateChanged(ItemEvent e) {
      Enumeration enum = listeners.keys();
      Object obj       = e.getSource();
      String cmd       = obj instanceof CheckboxMenuItem ?
                         ((CheckboxMenuItem)obj).getActionCommand() :
                         "";

      while (enum.hasMoreElements()) {
        ActionListener al = (ActionListener)enum.nextElement();
        ActionEvent ae;

        ae = new ActionEvent(obj,ActionEvent.ACTION_PERFORMED,cmd);
        al.actionPerformed(ae);
      } // end while
    } // end itemStateChanged

  } // end ActionAdapter

} // end MenuLoader

/* ******************************************************************
   end of file
*********************************************************************/
