/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.6)

   class      : ProjectFilePanel
   date       : 15.07.98
   description: a panel that displays a project file list

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.net.URL;

import javax.swing.*;
import javax.swing.text.TextAction;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

import Utils.ResourceUtil;
import Utils.FrameUtil;

/* ******************************************************************
   top level class declaration
*********************************************************************/

class ProjectFilePanel extends JPanel
  implements FilenameFilter, ListSelectionListener,Runnable {

  // class variable declarations
  /** Key to the project file panel label text */
  public static final String PROJFILE_LABEL = "pfLabelText";
  /** Identifier of the remove action */
  public static final String REMOVE_ACTION = "remove";

  // instance variable declarations
  /** The parent frame of this panel. */
  private Frame frame;
  private ResourceBundle resources;
  private Properties project;
  /** The Action to perform on a mouse double click. */
  private Action action;
  private Action[] externalActions;
  private String labelText;
  private JLabel label;
  JScrollPane scroller;
  private DefaultListModel model    = new DefaultListModel();
  private DefaultListModel tmpModel = new DefaultListModel();
  private JList list                = new JList(model);
  private boolean somethingSelected = false;
  private File path;
  private String[] sourceExt;
  private Cursor oldCursor;
  /** The number of source code documents in this list. */
  private int numDocs = 0;
  /** Actions defined by the ProjectFilePanel class. */
  private RemoveAction removeAction = new RemoveAction();
  private Action[] defaultActions = {
    removeAction
  };

  // constructors
  ProjectFilePanel(Frame frame, ResourceBundle resources,
                   Properties project, Action action) {
    this(frame,resources,project,action,null);
  } // end constructor ProjectFilePanel

  ProjectFilePanel(Frame frame, ResourceBundle resources,
                   Properties project, Action action,
                   Action[] externalActions) {
    this.frame           = frame;
    this.resources       = resources;
    this.project         = project;
    this.action          = action;
    this.externalActions = externalActions;

    String str = this.project.getProperty(Const.PROJECT_DIR);

    if (str.startsWith("file:/")) str = ResourceUtil.urlToFile(str);
    path      = new File(str);
    str       = this.project.getProperty(Const.SOURCE_EXT);
    sourceExt = ResourceUtil.splitToWords(str);
    setLayout(new BorderLayout());
    labelText  = ResourceUtil.getResStringOrKey(resources,PROJFILE_LABEL);
    labelText += " ";
    label = new JLabel(labelText + path);
    label.setToolTipText(path.toString());
    add("North",label);
    scroller = new JScrollPane(list);

    URL url = ResourceUtil.getResource(this,this.resources,Const.DOC_IMAGE);
    ImageIcon docImage = url == null ? null : new ImageIcon(url);

    list.setCellRenderer(new FileCellRenderer(docImage,this));
    list.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    MouseListener mouseListener = new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        if ((e.getClickCount() == 2) && !list.isSelectionEmpty()) {
          File file = (File)list.getSelectedValue();

          startAction(file);
        } // end if
      } // end mouseClicked
    }; // end anonymous class
    list.addMouseListener(mouseListener);
    list.addListSelectionListener(this);
    updateList();
    add("Center",scroller);
  } // end constructor ProjectFilePanel

  // class methods
  /** Forces update of the enclosing scrollers on list model change. */
  private static void setModel(JList l,DefaultListModel m) {
    int size = m.getSize();

    if (size > 0) {
      Object obj = m.lastElement();

      m.removeElementAt(size - 1);
      l.setModel(m);
      m.add(size - 1,obj);
    }
    else {
      l.setModel(m);
    } // end if
  } // end setModel

  // methods
  /**
   * Adds a file to the project list.
   *
   * @return True if file was added, false if it was allready in the list.
   */
  public boolean addFile(File f) {
    File[] oneFile = new File[1];

    oneFile[0] = f;
    return addFiles(oneFile);
  } // end addFile

  /**
   * Adds an array of files to the project list and the project properties.
   *
   * @return True if all files were added, false if one or more
   *         were allready in the list.
   */
  public boolean addFiles(File[] f) {
    boolean result    = true;
    boolean swapModel = (f.length > 1);
    int oldNumFiles   = Integer.parseInt(project.getProperty(Const.NUM_FILES));
    int numFiles      = oldNumFiles;
    String str;

    if (swapModel) setModel(list,tmpModel);
    for (int i=0; i<f.length; i++) {
      if (model.contains(f[i])) {
        result = false;
      }
      else {
        if (accept(new File(f[i].getParent()),f[i].getName())) {
          model.insertElementAt(f[i],numDocs);
          numDocs++;
        }
        else {
          model.addElement(f[i]);
        } // end if
        str = ResourceUtil.fileToURL(f[i].getAbsolutePath());
        project.put(Const.FILE_PREFIX + numFiles,str);
        numFiles++;
      } // end if
    } // end for
    if (swapModel) setModel(list,model);
    if (numFiles != oldNumFiles) {
      project.put(Const.PROJECT_CHANGED,"" + true);
      project.put(Const.NUM_FILES,"" + numFiles);
    } // end if
    return result;
  } // end addFiles

  /**
   * Removes afile from the project list and the project properties.
   *
   * @return True if the file was removed, false otherwise.
   */
  public boolean removeFile(File f) {
    int index = model.indexOf(f);

    if (index < 0) return false;
    list.clearSelection();
    list.setSelectedIndex(index);
    removeAction.actionPerformed(null);
    return true;
  } // end removeFile

  /**
   * Set the project home directory and update the list.
   */
  public void setHomeDirectory(File f) {
    if (f.equals(path)) return;
    path = f;
    project.put(Const.PROJECT_CHANGED,""+true);
    project.put(Const.PROJECT_DIR,
                ResourceUtil.fileToURL(path.getAbsolutePath()));
    label.setText(labelText + path);
    label.setToolTipText(path.toString());
    list.repaint();
  } // end setHomeDirectory

  /** Reload the list from the project properties. */
  public void updateList() {
    File file;
    String str;
    int numFiles = Integer.parseInt(project.getProperty(Const.NUM_FILES));

    numDocs  = 0;
    // reload the path
    str = project.getProperty(Const.PROJECT_DIR);
    if (str.startsWith("file:/")) str = ResourceUtil.urlToFile(str);
    file = new File(str);
    if (!path.equals(file)) {
      path = file;
      label.setText(labelText + path);
      label.setToolTipText(path.toString());
    } // end if
    // reload the source extensions
    str       = project.getProperty(Const.SOURCE_EXT);
    sourceExt = ResourceUtil.splitToWords(str);
    // creating a new model is fast than changing the current one !
    setModel(list,tmpModel);
    model.removeAllElements();
    for (int i=0; i<numFiles; i++) {
      String name = project.getProperty(Const.FILE_PREFIX + i);
      File f;

      if (name == null) {
        System.err.println("ProjectFilePanel: Project file number wrong.");
        break;
      } // end if
      name = ResourceUtil.urlToFile(name);
      f = new File(name);
      if (accept(new File(f.getParent()),f.getName())) {
        model.insertElementAt(f,numDocs);
        numDocs++;
      }
      else {
        model.addElement(f);
      } // end if
    } // end for
    setModel(list,model);
    list.repaint();
    scroller.validate();
  } // end updateList;

  /**
   * Fetch the list of actions supported by this panel.
   */
  public Action[] getActions() {
    if (externalActions == null)
      return defaultActions;
    else
      return TextAction.augmentList(defaultActions,externalActions);
  } // end getActions

  /** Returns an array containing the selected list items. */
  public Object[] getSelectedValues() {
    return list.getSelectedValues();
  } // end getSelectedValues

  /** Start the action which should be invoked after double clicks. */
  protected void startAction(File file) {
    if ((action != null) && action.isEnabled()) {
      oldCursor = frame.getCursor();
      frame.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      list.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      frame.getToolkit().sync();
      SwingUtilities.invokeLater(this);
    }
    else {
      Toolkit.getDefaultToolkit().beep();
    } // end if
  } // end startAction

  //---- Runnable --------------------------------------------------
  public void run() {
    File file = (File)list.getSelectedValue();
    ActionEvent ae = new ActionEvent(this,ActionEvent.ACTION_PERFORMED,
                                     file.getAbsolutePath());

    action.actionPerformed(ae);
    frame.setCursor(oldCursor);
    list.setCursor(oldCursor);
    frame.getToolkit().sync();
  } // end run

  //--------- FilenameFilter ----------------------------------------
  public boolean accept(File dir, String name) {
    for (int i=0; i<sourceExt.length; i++)
      if (name.endsWith(sourceExt[i])) return true;
    // end for
    return false;
  } // end accept

  //--------- ListSelectionListener ---------------------------------
  public void valueChanged(ListSelectionEvent e) {
    int i;

    if (somethingSelected != list.isSelectionEmpty()) return;
    somethingSelected = !list.isSelectionEmpty();
    removeAction.setEnabled(somethingSelected);
    if (externalActions != null) for (i=0; i<externalActions.length; i++) {
      if (externalActions[i] instanceof CompileAction)
        ((CompileAction)externalActions[i]).setActive(somethingSelected);
    } // end for
  } // end valueChanged

  // inner classes
  /** Sets the current displayed directory to the parent directory. */
  class RemoveAction extends AbstractAction {

    // constructors
    RemoveAction() {
      super(REMOVE_ACTION);
      setEnabled(false);
    } // end constructor RemoveAction

    // instance methods
    public void actionPerformed(ActionEvent e) {
      int[] indices     = list.getSelectedIndices();
      String[]removed   = new String[indices.length];
      boolean swapModel = (indices.length > 1);

      if (swapModel) setModel(list,tmpModel);
      for (int i=indices.length-1; i>=0; i--) {
        removed[i] = ((File)model.getElementAt(indices[i])).getAbsolutePath();
        model.removeElementAt(indices[i]);
        if (i < numDocs) numDocs--;
      } // end for
      if (swapModel) setModel(list,model);
      if (removed.length > 0) {
        int numFiles = Integer.parseInt(project.getProperty(Const.NUM_FILES));
        Vector v     = new Vector(numFiles);

        // add all project file names to a vector
        for (int i=0; i<numFiles; i++) {
          String name = project.getProperty(Const.FILE_PREFIX + i);

          if (name == null) {
            System.err.println("ProjectFilePanel: Project file number wrong.");
            break;
          } // end if
          name = ResourceUtil.urlToFile(name);
          v.addElement(name);
        } // end for
        // remove all removed project file names from the vector
        for (int i=0; i<removed.length; i++) {
          int index = v.indexOf(removed[i]);

          if (index >= 0) v.removeElementAt(index);
            else System.err.println("ProjectFilePanel: Wrong element in list.");
        } // end for
        // write project files to project properties
        project.put(Const.PROJECT_CHANGED,"" + true);
        for (int i=0; i<v.size(); i++)
          project.put(Const.FILE_PREFIX + i,
                      ResourceUtil.fileToURL(v.elementAt(i).toString()));
        project.put(Const.NUM_FILES,"" + v.size());
        // erase old file references
        int i = v.size();

        while (project.containsKey(Const.FILE_PREFIX + i)) {
          project.remove(Const.FILE_PREFIX + i);
          i++;
        } // end while
      } // end if
    } // end actionPerformed

  } // end RemoveAction

  class FileCellRenderer extends JLabel implements ListCellRenderer {

    // instance variable declarations
    private boolean focused = false;
    private Icon dirIcon;
    private Icon docIcon;
    private FilenameFilter filter;

    // constructors
    public FileCellRenderer (Icon docIcon, FilenameFilter filter) {
      this.docIcon = docIcon;
      this.filter  = filter;
      setOpaque(true);
    } // end constructor FileCellRenderer

    public Component getListCellRendererComponent(
        JList list, Object value, int index,
        boolean isSelected, boolean cellHasFocus) {
      File f = (File)value;
      if (value == null) {
        setText("");
        setIcon(null);
      }
      else {
        String name    = f.getPath();
        String absPath = path.getPath() + File.separator;

        if (name.startsWith(absPath)) name = name.substring(absPath.length());
        setText(name);
        if (filter.accept(new File(f.getParent()),f.getName())) {
          setIcon(docIcon);
        }
        else {
          setIcon(null);
        } // end if
      } // end if
      setBackground (isSelected ?
        SystemColor.textHighlight :
        SystemColor.window);
      setForeground (isSelected ?
        SystemColor.textHighlightText :
        SystemColor.windowText);
      return this;
    } // end getListCellRendererComponent

    public Dimension getPreferredSize() {
      Dimension dim = super.getPreferredSize();
      dim.width += 15;  // widen
      return dim;
    } // end getPreferredSize

  } // end FileCellRenderer

} // end ProjectFilePanel

/* ******************************************************************
   end of file
*********************************************************************/
