/* ******************************************************************

   class    : PreprocMainFilter
   date     : 05.08.98
   compiler : javac (jdk 1.1.6)

*********************************************************************/

package Streams;

/* ******************************************************************
   import statements
*********************************************************************/

import java.io.*;
import java.util.StringTokenizer;
import Utils.QuotedStringTokenizer;

/**
 * A reader that implements a part of a standard preprocessors
 * for ANSI C.
 *
 * @author puehringer edgar *
 */
public abstract class PreprocMainFilter extends LineNumberReader {

  // class vaiable declarations
  /** The default block comment characters. */
  protected final static String DEF_BLOCK_COMMENT = "/**/";
  /** The default eol comment characters. */
  protected final static String DEF_EOL_COMMENT = "//";
  /** The default escape character for preprocessor directives. */
  protected final static char DEF_PREPROC_ESC = '#';
  /** The default quote character for string literals. */
  protected final static char DEF_QUOTECHAR = '"';
  /** The default extra characters in identifiers. */
  protected final static String DEF_EXTRACHAR = "_";
  /** The default delimiter characters. */
  protected final static String DEF_DELIMITERS = "+*-/%=()[]{}\"'\\.;:,#<>";

  private static String lineSepar = System.getProperty("line.separator");

  // instance variable declarations
  private char[] cbuf = new char[0];
  private int offs    = 0;
  private StringBuffer stringBuffer;
  private QuotedStringTokenizer tokenizer;
  /** The nested comment level */
  private int commentLevel = 0;
  /** True if nested block comments are recognised. */
  protected boolean nestedComEnabled = false;
  protected String blockComment      = DEF_BLOCK_COMMENT;
  protected String blockCommentStart = DEF_BLOCK_COMMENT.substring(0,2);
  protected String blockCommentEnd   = DEF_BLOCK_COMMENT.substring(2);
  protected String eolComment        = DEF_EOL_COMMENT;
  protected String extraChar         = DEF_EXTRACHAR;
  protected String delimiters        = DEF_DELIMITERS + "\r\n\t ";
  /** The escape character for preprocessor directives. */
  protected char preprocEsc = DEF_PREPROC_ESC;
  /** The quote character for string literals. */
  protected char quoteCh = DEF_QUOTECHAR;
  /** True if filter output is enabled. */
  private boolean outputEnbl = true;
  /** True if the last filtered line was a preprocessor line */
  private boolean isPreprocLine;

  // constructors
  /**
   * Constructs a new PreprocMainFilter initialized with the
   * specified reader.
   *
   * @param in The reader.
   */
  public PreprocMainFilter(Reader in) {
    super(in);
  } // end constructor PreprocMainFilter

  /**
   * Constructs a new PreprocMainFilter initialized with the
   * specified reader that uses an input buffer of the specified size.
   *
   * @param in The reader.
   * @param sz Input buffer size.
   */
  public PreprocMainFilter(Reader in, int sz) {
    super(in,sz);
  } // end constructor PreprocMainFilter

  // class methods
  /**
   * Returns true if the given character equals the escape character for
   * preprocessor directives.
   */
  public static boolean isPreprocEsc(char ch) {
    return ch == DEF_PREPROC_ESC;
  } // end isPreprocEsc

  // methods
  /**
   * Reads a character.  The method will block if no input is available.
   *
   * @return  the character read, or -1 if the end of the stream is reached.
   * @exception IOException If an I/O error has occurred.
   */
  public int read() throws IOException {
    if (cbuf == null) return -1; // end of file
    if (offs >= cbuf.length) { // read new line
      String str = readLine();

      offs = 0;
      if (str == null) {
        cbuf = null;
        return -1;
      } // end if
      str += lineSepar;
      cbuf = str.toCharArray();
    } // end if
    return cbuf[offs++];
  } // end read

  /**
   * Reads into an array of characters.
   * Blocks until some input is available.
   *
   * @param cbuf the buffer into which the data is read
   * @param off the start offset of the data
   * @param len the maximum number of characters read
   * @return the actual number of characters read, -1 is
   *         returned when the end of the stream is reached.
   * @exception IOException If an I/O error has occurred.
   */
  public int read (char cbuf[], int off, int len) throws IOException {
    for (int i=off; i<(off + len); i++) {
      int c = read();

      if (c == -1) {
        if (i == 0) return -1;
          else return i - off;
      } // end if
      cbuf[i] = (char)c;
    } // end for
    return len;
  } // end read

  /**
   * Skips bytes of input.
   * @param n         bytes to be skipped
   * @return  actual number of bytes skipped
   * @exception IOException If an I/O error has occurred.
   */
  public long skip(long n) throws IOException {
    // Can't just read n bytes from 'in' and throw them
    // away, because n bytes from 'in' doesn't necessarily
    // correspond to n bytes from 'this'.
    for (int i=1; i <= n; i++) {
      int c = read();

      if (c == -1) return i - 1;
    } // end for
    return n;
  } // end skip

  /**
   * Returns the number of bytes that can be read without blocking.
   * @return the number of available bytes
   * @exception IOException If an I/O error has occurred.
   */
  public int available() throws IOException {
    if (cbuf == null) return 0;
    return (cbuf.length - offs);
  } // end available

  /** Returns a line of the input stream or null if eof is reached */
  public String readLine() throws IOException {
    String str;

    do {
      str = super.readLine();
      if (str == null) return str;
      tokenizer     = new QuotedStringTokenizer(str,delimiters,true,quoteCh);
      stringBuffer  = new StringBuffer();
      isPreprocLine = false;
      readUnknown();
    } while ((stringBuffer.length() == 0) &&
             (!outputEnbl || isPreprocLine));
    return stringBuffer.toString();
  } // end readLine

  /**
   * Should be overwritten; Funktionallity: Changes identifiers to prev.
   * stored define values.
   */
  protected abstract String evalIdentifier(String ident);

  /** Parse a preprocessor command line. */
  protected abstract void processCmdLine(String str);

  /** Enables or disables the output of this filter. */
  protected void enableOutput(boolean enbl) {
    outputEnbl = enbl;
  } // end enableOutput

  /** Returns true if the output of this filter is enabled. */
  protected boolean isOutputEnabled() {
    return outputEnbl;
  } // end isOutputEnabled

  /**
   * Read from tokenizer without knowing if this line is a preprozessor line.
   */
  protected void readUnknown() {
    String tok;
    char ch;
    int pos;
    int bComMatch = 0;
    int eComMatch = 0;

    if (!tokenizer.hasMoreTokens()) return;
    // if some comment is open, then read it
    if (commentLevel > 0) readComment();
    tokenizer.delimiterChars(delimiters);
    tokenizer.quoteChar(quoteCh);
    if (!tokenizer.hasMoreTokens()) return;
    // remember this position to discard all scanned characters later
    pos = stringBuffer.length();
    // clip leading whitespaces
    tok = tokenizer.previewToken();
    ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;
    while (tokenizer.hasMoreTokens() && Character.isWhitespace(ch)) {
      if (outputEnbl) stringBuffer.append(tokenizer.nextToken());
        else tokenizer.nextToken();
      tok = tokenizer.previewToken();
      ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;
    } // end while
    // the main loop
    while (tokenizer.hasMoreTokens()) {
      tok = tokenizer.previewToken();
      ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;

      if (blockCommentStart.charAt(bComMatch) == ch) bComMatch++;
        else bComMatch = 0;
      if (eolComment.charAt(eComMatch) == ch) eComMatch++;
        else eComMatch = 0;
      if (bComMatch == blockCommentStart.length()) { // block comment
        bComMatch = 0;
        eComMatch = 0;
        commentLevel++;
        Utils.Assert.assert(Utils.Assert.ENABLED &&
          (commentLevel == 1));
        if (outputEnbl) stringBuffer.append(tokenizer.nextToken());
          else tokenizer.nextToken();
        readComment();
        tokenizer.delimiterChars(delimiters);
        tokenizer.quoteChar(quoteCh);
      } else if (eComMatch == eolComment.length()) { // eol comment
        bComMatch = 0;
        eComMatch = 0;
        if (outputEnbl) stringBuffer.append(tokenizer.nextToken());
          else tokenizer.nextToken();
        readEolComment();
        Utils.Assert.assert(Utils.Assert.ENABLED &&
          !tokenizer.hasMoreTokens());
      } else if (ch == preprocEsc) { // preprocessor statement
        tokenizer.nextToken();
        stringBuffer.setLength(pos);
        readMacroprocessor();
        Utils.Assert.assert(Utils.Assert.ENABLED &&
          !tokenizer.hasMoreTokens());
      } else if (Character.isWhitespace(ch) || // whitechar or
                 (bComMatch != 0) ||           // start of block comment or
                 (eComMatch != 0)) {           // start of eol comment
          if (outputEnbl) stringBuffer.append(tokenizer.nextToken());
            else tokenizer.nextToken();
      } else {                         // other token
        readSource();
        Utils.Assert.assert(Utils.Assert.ENABLED &&
          !tokenizer.hasMoreTokens());
      } // end if
    } // end while
  } // end readUnknown

  /**
   * Read the rest of a block comment.
   */
  protected void readComment() {
    String tok;
    char ch;
    int startMatch = 0;
    int endMatch   = 0;

    tokenizer.delimiterChars(blockComment);
    tokenizer.quoteChar(-1);
    while (tokenizer.hasMoreTokens() && (commentLevel > 0)) {
      tok = tokenizer.nextToken();
      ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;

      if (outputEnbl) stringBuffer.append(tok);
      if (blockCommentStart.charAt(startMatch) == ch) startMatch++;
          else startMatch = 0;
      if (blockCommentEnd.charAt(endMatch) == ch) endMatch++;
          else endMatch = 0;
      if (endMatch == blockCommentEnd.length()) {
        startMatch = 0;
        endMatch   = 0;
        commentLevel--;
      } else if (startMatch == blockCommentStart.length()) {
        startMatch = 0;
        endMatch   = 0;
        commentLevel++;
      } // end if
    } // end while
  } // end readComment

  /**
   * Read a comment that reaches til eol.
   */
  protected void readEolComment() {
    String tok;
    char ch;
    boolean done = false;

    tokenizer.delimiterChars("\r\n");
    tokenizer.quoteChar(-1);
    while (tokenizer.hasMoreTokens() && !done) {
      tok = tokenizer.nextToken();
      ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;

      if (outputEnbl) stringBuffer.append(tok);
      switch (ch) {
        case '\r':
          if (tokenizer.hasMoreTokens() &&
              tokenizer.previewToken().equals("\n")) {
            if (outputEnbl) stringBuffer.append(tokenizer.nextToken());
              else tokenizer.nextToken();
          } // end if
        case '\n': done = true;
        default:
      } // end switch
    } // end while
  } // end readEolComment

  /**
   * Read from the tokenizer assuming that it contains source code.
   */
  protected void readSource() {
    String tok;
    char ch;
    int bComMatch = 0;
    int eComMatch = 0;

    tokenizer.delimiterChars(delimiters);
    tokenizer.quoteChar(quoteCh);
    while (tokenizer.hasMoreTokens()) {
      tok = tokenizer.nextToken();
      ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;

      if (blockCommentStart.charAt(bComMatch) == ch) bComMatch++;
        else bComMatch = 0;
      if (eolComment.charAt(eComMatch) == ch) eComMatch++;
        else eComMatch = 0;
      if (bComMatch == blockCommentStart.length()) { // block comment
        bComMatch = 0;
        eComMatch = 0;
        commentLevel++;
        Utils.Assert.assert(Utils.Assert.ENABLED &&
          (commentLevel == 1));
        if (outputEnbl) stringBuffer.append(tok);
        readComment();
        tokenizer.delimiterChars(delimiters);
        tokenizer.quoteChar(quoteCh);
      } else if (eComMatch == eolComment.length()) { // eol comment
        bComMatch = 0;
        eComMatch = 0;
        if (outputEnbl) stringBuffer.append(tok);
        readEolComment();
        tokenizer.delimiterChars(delimiters);
        tokenizer.quoteChar(quoteCh);
      } else {                                       // ident or other token
        if (isIdentifier(tok)) tok = evalIdentifier(tok);
        if (outputEnbl) stringBuffer.append(tok);
      } // end if
    } // end while
  } // end readSource

  /**
   * Read from the tokenizer assuming that it contains preprocessor code.
   */
  protected void readMacroprocessor() {
    StringBuffer cmdLine = new StringBuffer();
    String tok;
    char ch;
    int pos;
    int bComMatch = 0;
    int eComMatch = 0;

    isPreprocLine = true;
    tokenizer.delimiterChars(delimiters);
    tokenizer.quoteChar(quoteCh);
    while (tokenizer.hasMoreTokens()) {
      tok = tokenizer.nextToken();
      ch  = tok.length() == 1 ? tok.charAt(0) : (char)0;

      if (blockCommentStart.charAt(bComMatch) == ch) bComMatch++;
        else bComMatch = 0;
      if (eolComment.charAt(eComMatch) == ch) eComMatch++;
        else eComMatch = 0;
      if (bComMatch == blockCommentStart.length()) { // block comment
        bComMatch = 0;
        eComMatch = 0;
        commentLevel++;
        Utils.Assert.assert(Utils.Assert.ENABLED &&
          (commentLevel == 1));
        cmdLine.append(tok);
        cmdLine.setLength(cmdLine.length() - blockCommentStart.length());
        cmdLine.append(" ");
        pos = stringBuffer.length();
        if (outputEnbl) stringBuffer.append(blockCommentStart);
        readComment();
        if (commentLevel == 0) stringBuffer.setLength(pos);
        tokenizer.delimiterChars(delimiters);
        tokenizer.quoteChar(quoteCh);
      } else if (eComMatch == eolComment.length()) { // eol comment
        bComMatch = 0;
        eComMatch = 0;
        cmdLine.append(tok);
        cmdLine.setLength(cmdLine.length() - eolComment.length());
        cmdLine.append(" ");
        pos = stringBuffer.length();
        readEolComment();
        stringBuffer.setLength(pos);
        tokenizer.delimiterChars(delimiters);
        tokenizer.quoteChar(quoteCh);
      } else {                                       // ident or other token
        // if (isIdentifier(tok)) tok = evalIdentifier(tok);
        cmdLine.append(tok);
      } // end if
    } // end while
    processCmdLine(cmdLine.toString());
  } // end readMacroprocessor

  protected boolean isIdentifier(String str) {
    int len = str.length();
    char ch;

    if (len == 0) return false;
    ch = str.charAt(0);
    if (!Character.isLetter(ch) &&
        (extraChar.indexOf(ch) < 0)) return false;
    for (int i=1; i<len; i++) {
      ch = str.charAt(0);
      if (!Character.isLetter(ch) &&
          !Character.isDigit(ch) &&
          (extraChar.indexOf(ch) < 0)) return false;
    } // end for
    return true;
  } // end isIdentifier

  /** Replaces identifiers in a token sequence. */
  protected String processTokenSeq(String str) {
    QuotedStringTokenizer tok;
    StringBuffer buf = new StringBuffer();
    String s;

    tok = new QuotedStringTokenizer(str,delimiters,true,quoteCh);
    while (tok.hasMoreTokens()) {
      s = tok.nextToken();
      if (isIdentifier(s)) s = evalIdentifier(s);
      buf.append(s);
    } // end while
    return buf.toString();
  } // end processTokenSeq

} // end PreprocMainFilter

/* ******************************************************************
   end of file
*********************************************************************/
